/* Parser global settings.
 *
 * PegSoft configuration parser library (c) 2004 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software (the COPYING file); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 */

/** 
 * \file conf/global.h
 * \brief Configuration parser global settings header file.
 */

#ifndef CONF_GLOBAL_H
#define CONF_GLOBAL_H

/*************************************************************************/

#include <list>
#include <map>
#include <string>

#include <misc/string.h>

namespace conf
{

/*************************************************************************/

class Conf;
class ConfElse;
class ConfIfnset;
class ConfIfset;
class ConfInclude;
class ConfPath;
class ConfVarSet;
class ConfUnset;

/*************************************************************************/

/// Global settings for configuration parser.

/** This class is used by the Conf class to retrieve special global directives,
 *  retrieve and set variables and look for include paths.
 *
 *  The include paths are used by Conf::parse_file() as potential paths to
 *  look into when opening a configuration file with a non-absolute (i.e. no
 *  leading /) name.
 *
 *  Variables are used by the configuration parser to do substitution in quoted
 *  strings (i.e. in "$(somename)", $(somename) will be replaced by the value
 *  of the variable named somename).
 *
 *  Finally, global directives are used by the parser as an additional list
 *  of directives to look into when resolving a directive name. The local
 *  directives are checked before the global ones.
 */

class ConfGlobal
{
	public:
		/// Type of the path list.
		typedef std::list<std::string> PathList;
		/// Type of defines list.
		typedef std::map<std::string, std::string, misc::insensitive_less>
				VarList;
	
		friend class Conf;
	
		/// Constructor setting the global directives.
		/** This constructor will set the global directives list. Global
		  * directives are stored in a simple Conf object, so you can
		  * add and remove directives to it easily. The object is NOT
		  * copied, so it must exist until the ConfGlobal object is
		  * destructed.
		  *
		  * You do not need to call Conf::begin() or Conf::end() on
		  * the Conf object, this is done automatically by the
		  * Conf::begin() and Conf::end() functions of the object
		  * that owns this ConfGlobal object.
		  *
		  * When the ConfDir::set() function is called from the parser
		  * for a global directive, the directive's owning Conf object
		  * will be temporarily set to the calling Conf object, before
		  * being reset to its previous value. This in order to have
		  * a meaningful context... =) (Actually it makes it even 
		  * more practical to write global directives -- they're just
		  * like standard ones.)
		  *
		  * This constructor will add some default global directives
		  * to the list automatically.
		  *
		  * \param conf The Conf object containing the global directives. */
		explicit ConfGlobal(Conf &conf) : m_refcount(0), m_directives(&conf),
				m_paths(), m_vars(), m_else(0), m_ifnset(0), m_ifset(0),
				m_include(0), m_path(0), m_set(0), m_unset(0)
				{ add_default(); }
		/// Destructor
		~ConfGlobal();
		
		/// Returns the global directives.
		/** Returns the global directives list. They are stored in a
		  * simple object of the Conf class, so you may add and remove
		  * directives to it easily.
		  * \return The global directives list.
		  * \sa ConfGlobal(Conf &) */
		Conf &directives() const { return *m_directives; }
		
		/// Returns the ConfElse object.
		/** Returns the ConfElse object that has been added to the
		  * global directives list as the "else" global directive. This
		  * is necessary for ConfIf (and derived) classes to call the
		  * ConfElse::can_process() function accordingly.
		  * \return The ConfElse object. */
		ConfElse &else_directive() const { return *m_else; }
		
		/// Returns the current include paths.
		/** Returns a container of the current include paths. The include
		  * paths are used by Conf::parse_file() as potential paths to look
		  * into when opening a configuration file with a non-absolute
		  * (i.e. no leading /) name.
		  * \return The current include paths. */
		PathList const &incl_paths() const { return m_paths; }
		/// Adds an include path.
		/** Adds an include path. Existence of the path is checked before
		  * being added to the list. If the specified path is not absolute,
		  * then it must be relative to either the process current working
		  * directory or to a path already in the list, and it is
		  * automatically converted to an absolute path.
		  * \param[in] path The new include path.
		  * \param[in] pi Optional pointer to the iterator that will receive
		  * a copy of the iterator to the newly-added path in the list (only
		  * when the function returns true).
		  * \return true if the path exists and was added, false otherwise.
		  * \sa incl_paths() */
		bool incl_paths_add(std::string const &path, 
				PathList::const_iterator *pi = 0);
		/// Removes an include path.
		/** Removes an include path.
		  * \param[in] path The include path to be removed.
		  * \return true if the path was found and deleted,
		  *         false otherwise.
		  * \sa incl_paths() */
		bool incl_paths_remove(std::string const &path);
		
		/// Retrieves the value of a variable.
		/** Retrieves the value of a given variable (if it exists).
		  * \param[in] name The variable name.
		  * \param[out] value A pointer to the string that will have
		  * been set to the variable value when function returns. If
		  * this parameter is NULL then the function just checks
		  * whether the variable exists or not.
		  * \return true if the variable exists, false otherwise. */
		bool var_get(std::string const &name, std::string *value = NULL) const;
		/// Sets a variable.
		/** Sets a variable to a given value. If a variable with this
		  * name already exists its value is overriden.
		  * \param[in] name The variable name.
		  * \param[in] value The new value of the variable.
		  * \return true if the variable has been set, false otherwise
		  * (because the name is invalid).
		  */
		bool var_set(std::string const &name, std::string const &value);
		/// Sets a variable from an assignment string.
		/** Sets a variable to a given value, using the given assignment.
		  * If a variable with this name already exists its value is
		  * overriden.
		  * \param[in] assignment A string in the form xxxx=yyyy where
		  * xxxx is the variable name and yyyy is the value.
		  * \return true if the variable has been set, false otherwise
		  * (because either the assignment or the name is invalid).
		  */
		bool var_set(std::string const &assignment);
		/// Unsets a variable.
		/** Unsets the given variable, if it exists.
		  * \param[in] name The variable name.
		  * \return true if the variable existed and has been unset,
		  * false otherwise.
		  */
		bool var_unset(std::string const &name);
	private:
		int m_refcount;
		
		Conf *m_directives;
		PathList m_paths;
		VarList m_vars;
		
		ConfElse *m_else;
		ConfIfnset *m_ifnset;
		ConfIfset *m_ifset;
		ConfInclude *m_include;
		ConfPath *m_path;
		ConfVarSet *m_set;
		ConfUnset *m_unset;
		
		/* Don't let anyone use these two. */
		ConfGlobal(ConfGlobal const &);
		ConfGlobal &operator=(ConfGlobal const &right);
		
		void add_default();
};

/*************************************************************************/

} /* namespace conf */

/*************************************************************************/

#endif /* CONF_GLOBAL_H */
