/* Quick configuration parser setup.
 *
 * PegSoft configuration parser library (c) 2004 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software (the COPYING file); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 */

/** 
 * \file conf/quick.h
 * \brief Classes for quick configuration parser setup and use.
 */

#ifndef CONF_QUICK_H
#define CONF_QUICK_H

/*************************************************************************/

#include <iosfwd>
#include <string>

#include <misc/misc.h>
#include <conf/parser.h>
#include <conf/global.h>

namespace conf
{

/*************************************************************************/

/// Class for easy use of the configuration global settings.

/** ConfGlobalQ is a class that encapsulates all configuration global
  * settings stuff, allowing you to focus on the essential, i.e. adding
  * global directives, variables and include paths.
  *
  * The setup() function must be called before using an object of
  * this class. */

class ConfGlobalQ
{
	public:
		/// Constructor.
		/** Constructs the ConfGlobalQ object.
		  *
		  * \param[in] logger Logger used to construct the Conf
		  * object used to store global directives. */
		ConfGlobalQ(ConfLogger &logger) : m_gconf(logger),
				m_global(m_gconf), m_setup(false) { }
		/// Destructor.
		virtual ~ConfGlobalQ() { }
		
		/// Adds global directives, variables and include paths.
		/** Override this function to add global directives to the
		  * m_conf object, along with variables and include paths
		  * to the m_global object.
		  *
		  * This function must be called before using an object
		  * of this class.
		  *
		  * \return true if setup completed successfully, false
		  * otherwise. */
		virtual bool setup() { return (m_setup = true); }
		
		/// Returns the Conf object managing global directives.
		Conf &gconf() { return m_gconf; }
		/// Returns the ConfGlobal object managing variables and include paths.
		ConfGlobal &global() { return m_global; }
		
		/// Has the setup() function been called?
		bool is_setup() const { return m_setup; }
	protected:
		/// Add global directives to this object.
		Conf m_gconf;
		/// Add variables and include paths to this object.
		ConfGlobal m_global;
	private:
		bool m_setup;
		
		ConfGlobalQ(ConfGlobalQ const &);
		ConfGlobalQ &operator=(ConfGlobalQ const &);
};

/*************************************************************************/

/// Class for easy use of the configuration parser.

/** ConfQ is a class that encapsulates all configuration parser stuff,
  * allowing you to focus on the essential, i.e. adding directives.
  *
  * The setup() function must be called prior to using the parse(),
  * parse_file() and parse_string() functions. */

class ConfQ
{
	public:
		/// Constructor.
		/** Constructs the ConfQ object.
		  *
		  * \param[in] logger Logger used by the parser to
		  * report errors and debug messages (see Conf::Conf()). */
		ConfQ(ConfLogger &logger) : m_conf(logger), m_setup(false) { }
		/// Copy constructor.
		ConfQ(ConfQ const &cq) : m_conf(cq.m_conf), m_setup(cq.m_setup) { }
		/// Assignment operator.
		virtual ConfQ &operator=(ConfQ const &right);
		/// Destructor.
		virtual ~ConfQ() { }
		
		/// Adds directives.
		/** Override this function to add directives to the
		  * parser.
		  *
		  * This function is called once per instance by parse(),
		  * parse_file() and parse_string().
		  *
		  * \return true if setup completed successfully, false
		  * otherwise.
		  *
		  * \sa parse(), parse_file(), parse_string() */
		virtual bool setup() { return (m_setup = true); }
		
		/// Parses a stream.
		/** Parses a stream. This function calls the Conf::begin(),
		  * Conf::parse() and Conf::end() functions for the parser.
		  *
		  * \param[in] cgq Sets the ConfGlobalQ object used to
		  * initialize the parser.
		  * \param[in] reload Specifies whether the configuration
		  * is being reloaded (true) or not (false).
		  * \param[in] is The input stream to parse.
		  * \param[in] context An arbitrary string giving indication
		  * on the parsing location.
		  *
		  * \return true if everything succeeded, false otherwise
		  * (appropriate error should have been logged by the parser
		  * in the latter case).
		  *
		  * \sa Conf::begin(), Conf::parse(), Conf::end()
		  */
		bool parse(ConfGlobalQ &cgq, bool reload, std::istream &is,
				std::string const &context);
		/// Parses a file.
		/** Parses a file. This function calls the Conf::begin(),
		  * Conf::parse_file() and Conf::end() functions for the parser.
		  *
		  * \param[in] cgq Sets the ConfGlobalQ object used to
		  * initialize the parser.
		  * \param[in] reload Specifies whether the configuration
		  * is being reloaded (true) or not (false).
		  * \param[in] filename The file to parse.
		  *
		  * \return true if everything succeeded, false otherwise
		  * (appropriate error should have been logged by the parser
		  * in the latter case).
		  *
		  * \sa Conf::begin(), Conf::parse_file(), Conf::end()
		  */
		bool parse_file(ConfGlobalQ &cgq, bool reload,
				std::string const &filename);
		/// Parses a string.
		/** Parses a string. This function calls the Conf::begin(),
		  * Conf::parse_string() and Conf::end() functions for the parser.
		  *
		  * \param[in] cgq Sets the ConfGlobalQ object used to
		  * initialize the parser.
		  * \param[in] reload Specifies whether the configuration
		  * is being reloaded (true) or not (false).
		  * \param[in] str The string to parse.
		  * \param[in] context An arbitrary string giving indication
		  * on the parsing location.
		  *
		  * \return true if everything succeeded, false otherwise
		  * (appropriate error should have been logged by the parser
		  * in the latter case).
		  *
		  * \sa Conf::begin(), Conf::parse_string(), Conf::end()
		  */
		bool parse_string(ConfGlobalQ &cgq, bool reload,
				std::string const &str, std::string const &context);
		
		/// Returns the configuration parser object.
		Conf &conf() { return m_conf; }
		/// Has the setup() function been called?
		bool is_setup() const { return m_setup; }
	protected:
		/// Add directives to this object.
		Conf m_conf;
	private:
		bool m_setup;
};

/*************************************************************************/

/// Base class for directives encapsulation.
/** This class can be used as a base class for directives encapsulation.
  *
  * It creates a configuration parser and allows derived classes to
  * encapsulate directives seamlessly and add them to the parser as needed.
  *
  * The add() function must be overridden in derived classes. */

class DirectivesQ
{
	public:
		/* Classes */
		
		/// Custom configuration parser class.
		/** This configuration parser class will call the
		  * DirectivesQ::add() function before parsing the
		  * configuration for the first time.
		  *
		  * This class is meant to be instantiated by its
		  * enclosing DirectivesQ object only, and so it
		  * won't be throughly documented. */
		class DConfQ : public ConfQ
		{
			public:
				/// Constructor.
				DConfQ(ConfLogger &logger, DirectivesQ &directives) :
						ConfQ(logger), m_directives(&directives) { }
				/// Copy constructor.
				DConfQ(DConfQ const &dcq) : ConfQ(dcq),
						m_directives(dcq.m_directives) { }
				/// Assignment operator.
				DConfQ &operator=(DConfQ const &right);
				
				/// Calls DirectivesQ::add().
				virtual bool setup();
			protected:
				DirectivesQ *m_directives;
		};
		
		/* Functions */
		
		/// Constructor.
		/** Constructor.
		  * \param[in] logger The configuration logger used to create
		  * the parser. */
		DirectivesQ(ConfLogger &logger) : m_parser(logger, *this) { }
		/// Copy constructor.
		DirectivesQ(DirectivesQ const &d) : m_parser(d.m_parser) { }
		/// Assignment operator.
		DirectivesQ &operator=(DirectivesQ const &right);
		/// Destructor.
		virtual ~DirectivesQ() { }
		
		/// Adds directives to the parser.
		/** This function must be overriden in derived classes
		  * to add directives to the given configuration parser.
		  *
		  * \param[in] conf The configuration parser.
		  * \return true if successful, false otherwise. */
		virtual bool add(Conf &conf) = 0;
		
		/// Returns the configuration parser.
		DConfQ &parser() { return m_parser; }
	protected:
		/// The configuration parser.
		DConfQ m_parser;
};

/*************************************************************************/

} /* namespace conf */

/*************************************************************************/

#endif /* CONF_QUICK_H */
