/* Vector SP.
 *
 * Containers collection (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** \file containers/vectorsp.h
  * \brief Mini vector container header file. */

#ifndef CONTAINERS_VECTORSP_H
#define CONTAINERS_VECTORSP_H

/*************************************************************************/

#include <iterator>

/*************************************************************************/

namespace containers
{

/*************************************************************************/

/// A small-size pointer-focused vector.

/** This class is a vector-like designed with small size, pointer
  * values and efficiency in mind.
  *
  * \param T The type of elements in the mini vector (must be a
  * pointer).
  * \param SizeT An unsigned integer type used for the size, which
  * determines how much elements can be inserted in the vector. */

template <class T, class SizeT>
class VectorSP
{
	public:
		/// Value type.
		typedef T value_type;
		/// Reference type
		typedef T & reference;
		/// Const reference type
		typedef T const & const_reference;
		/// Pointer type
		typedef T * pointer;
		/// Const pointer type
		typedef T const * const_pointer;
		/// Size type.
		typedef SizeT size_type;
		/// Difference type.
		typedef ptrdiff_t difference_type;
	
		/// Normal iterator type.
		typedef pointer iterator;
		/// Const iterator type.
		typedef const_pointer const_iterator;
		/// Reverse iterator
		typedef std::reverse_iterator<iterator> reverse_iterator;
		/// Const reverse iterator
		typedef std::reverse_iterator<const_iterator> const_reverse_iterator;
		
		/// Constructor.
		VectorSP() : m_array(0), m_size(0), m_capacity(0) { }
		/// Copy constructor.
		VectorSP(VectorSP const &v);
		/// Assignment operator.
		VectorSP &operator=(VectorSP const &right);
		/// Destructor.
		~VectorSP();
		
		/// Clears the vector.
		/** Clears the vector.
		  *
		  * \param[in] free true if operator delete should be called on
		  * non-NULL elements in the vector, false otherwise. */
		void clear(bool free = false);
		
		/// Erases an element of the vector.
		/** Erases an element of the vector.
		  *
		  * \param[in] it An iterator pointing to the element to erase.
		  * \param[in] free true if operator delete should be called
		  * on the removed element (if non-NULL), false otherwise. */
		void erase(iterator it, bool free = false);
		
		/// Erases a range of elements of the vector.
		/** Erases a range of elements of the vector.
		  *
		  * \param[in] first The first iterator of the range.
		  * \param[in] last The last iterator of the range.
		  * \param[in] free true if operator delete should be called
		  * on removed, non-NULL pointers, false otherwise. */
		void erase(iterator first, iterator last, bool free = false);
		
		/// Inserts an element at a specified position.
		/** Inserts an element at a specified position.
		  *
		  * \param[in] pos An iterator pointing to the element 
		  * immediately after where the new element will be added.
		  * \param[in] val The value of the new element.
		  * \return An iterator to the new element, or end() if
		  * no new element could be added. */
		iterator insert(iterator pos, value_type const &val);
		
		/// Removes the last element.
		void pop_back(bool free = false) { erase(end() - 1, free); }
		/// Inserts an element at the end.
		void push_back(value_type const &val) { insert(end(), val); }
		
		/// Sets the minimum capacity of the vector.
		/** Sets the minimum capacity of the vector. If the given
		  * capacity is inferior to the current one, this
		  * function is a no-op.
		  *
		  * \param[in] capacity The new capacity of the vector. If this
		  * is superior to max_size(), it is truncated to
		  * max_size(). */
		void reserve(size_type capacity);
		/// Resizes the vector.
		/** Resizes the vector. If the new size is inferior to
		  * the current one, then additional elements are removed,
		  * with the delete pointer possibly being called on non-NULL
		  * pointers in the vector (the capacity is left untouched though).
		  * Else, new elements are added at the end of the vector.
		  *
		  * \param[in] size The new size of the vector. If this
		  * is superior to max_size(), it is truncated to
		  * max_size().
		  * \param[in] val The value of the potential new elements.
		  * \param[in] free true if operator delete should be called
		  * on removed, non-NULL pointers, false otherwise. */
		void resize(size_type size, value_type const &val = NULL,
				bool free = false);
		
		/// Returns an iterator to the first element of the vector.
		iterator begin() { return m_array; }
		/// Returns an iterator beyond the last element of the vector.
		iterator end() { return &m_array[m_size]; }
		/// Returns a const iterator to the first element of the vector.
		const_iterator begin() const 
				{ return m_array; }
		/// Returns a const iterator beyond the last element of the vector.
		const_iterator end() const 
				{ return &m_array[m_size]; }
		
		/// Returns a reverse iterator to the beginning of the vector.
		reverse_iterator rbegin() { return reverse_iterator(end()); }
		/// Returns a reverse iterator to the end of the vector.
		reverse_iterator rend() { return reverse_iterator(begin()); }
		/// Returns a const reverse iterator to the beginning of the vector.
		const_reverse_iterator rbegin() const { return const_reverse_iterator(end()); }
		/// Returns a const reverse iterator to the end of the vector.
		const_reverse_iterator rend() const { return const_reverse_iterator(begin()); }
		
		/// Operator to access an arbitrary element of the vector.
		reference operator[](size_type idx) { return m_array[idx]; }
		/// Operator to access an arbitrary element of the vector (const version).
		const_reference operator[](size_type idx) const { return m_array[idx]; }
		
		/// Returns the first element of the vector.
		reference front() { return m_array[0]; }
		/// Returns the first element of the vector (const version).
		const_reference front() const { return m_array[0]; }
		/// Returns the last element of the vector.
		reference back() { return m_array[m_size - 1]; }
		/// Returns the last element of the vector (const version).
		const_reference back() const { return m_array[m_size - 1]; }
		
		/// Returns the data in the container.
		const_pointer data() const { return m_array; }
		
		/// Returns the actual number of elements in the vector.
		size_type size() const { return m_size; }
		/// Returns whether the vector is empty.
		bool empty() const { return (m_size == 0); }
		/// Returns the capacity of the vector.
		size_type capacity() const { return m_capacity; }
		/// Returns the maximum size of the vector.
		size_type max_size() const;
	private:
		T *m_array;
		
		size_type m_size;
		size_type m_capacity;
		
		void copy_array(T const *array, size_type size, size_type capacity);
		void free_array();
};

/*************************************************************************/

template <class T, class SizeT>
VectorSP<T, SizeT>::VectorSP(VectorSP const &v) : m_array(0), m_size(0),
		m_capacity(0)
{
	copy_array(v.m_array, v.m_size, v.m_size /* Don't use the same capacity. */);
}

/*************************************************************************/

template <class T, class SizeT>
VectorSP<T, SizeT> &VectorSP<T, SizeT>::operator=(VectorSP const &right)
{
	if (this != &right)
	{
		free_array();
		copy_array(right.m_array, right.m_size, right.m_size);
	}
	
	return *this;
}

/*************************************************************************/

template <class T, class SizeT>
VectorSP<T, SizeT>::~VectorSP()
{
	free_array();
}

/*************************************************************************/

template <class T, class SizeT>
void VectorSP<T, SizeT>::clear(bool free)
{
	if (free)
	{
		iterator it, iend;
		
		for (it = begin(), iend = end(); it != iend; ++it)
			if (*it)
				delete (*it);
	}
	
	free_array();
}

/*************************************************************************/

template <class T, class SizeT>
void VectorSP<T, SizeT>::copy_array(T const *array, size_type size,
		size_type capacity)
{
	if (!capacity)
		return; /* Nothing to do. */
	
	m_capacity = capacity;
	m_array = new value_type[m_capacity];
	
	if (!array || !size)
		return;
	
	m_size = size;
	memcpy(m_array, array, (m_size * sizeof(value_type)));
}

/*************************************************************************/

template <class T, class SizeT>
void VectorSP<T, SizeT>::erase(iterator it, bool free)
{
	/* Sanity check -- I know there shouldn't be any in
	 * containers, but well -- I'm a bit worried on this one. :P */
	if (!m_size)
		return;
	
	if (free && (*it))
		delete (*it);
	
	if (it != (end() - 1))
		memmove(it, it + 1, (sizeof(value_type) * (end() - it - 1)));
	--m_size;
}

/*************************************************************************/

template <class T, class SizeT>
void VectorSP<T, SizeT>::erase(iterator first, iterator last, bool free)
{
	/* Sanity check -- I know there shouldn't be any in
	 * containers, but well -- I'm a bit worried on this one (too). :P */
	if ((last - first) > m_size)
		return;
	
	if (free)
	{
		for (iterator it = first; it != last; ++it)
			if (*it)
				delete (*it);
	}
	
	if (last != end())
		memmove(first, last, (sizeof(value_type) * (end() - last)));
	m_size -= (last - first);
}

/*************************************************************************/

template <class T, class SizeT>
void VectorSP<T, SizeT>::free_array()
{
	if (m_array)
		delete [] m_array;
	m_array = 0;
	m_capacity = m_size = 0;
}

/*************************************************************************/

template <class T, class SizeT>
typename VectorSP<T, SizeT>::iterator VectorSP<T, SizeT>::insert(iterator pos,
		value_type const &val)
{
	if (m_size == max_size())
		return end();
	
	size_type idx = (m_size - (end() - pos));
	
	reserve(++m_size);
	
	iterator newpos = &m_array[idx];
	
	if (newpos != end())
		memmove(newpos + 1, newpos, (sizeof(value_type) * (end() - newpos - 1)));
	*newpos = val;
	return newpos;
}

/*************************************************************************/

template <class T, class SizeT>
typename VectorSP<T, SizeT>::size_type VectorSP<T, SizeT>::max_size() const
{
	size_type result = 0;
	
	if (sizeof(T *) <= sizeof(size_type))
	{
		for (unsigned int i = 0; i < (sizeof(T *) * 8); ++i)
		{
			result <<= 1;
			result |= 1;
		}
		
		result /= sizeof(T *);
	} else {
		for (unsigned int i = 0; i < (sizeof(size_type) * 8); ++i)
		{
			result <<= 1;
			result |= 1;
		}
	}
	
	return result;
}

/*************************************************************************/

template <class T, class SizeT>
void VectorSP<T, SizeT>::reserve(size_type capacity)
{
	if (capacity > max_size())
		capacity = max_size();
	
	if (capacity <= m_capacity)
		return;
	
	T *array = m_array;
	
	copy_array(array, m_size, capacity);
	delete [] array;
}

/*************************************************************************/

template <class T, class SizeT>
void VectorSP<T, SizeT>::resize(size_type size,
		value_type const &val, bool free)
{
	if (size > max_size())
		size = max_size();
	
	if (size == m_size)
		return; /* No-op */
	
	if (size < m_size)
	{
		if (free)
		{
			for (size_type i = size; i < m_size; ++i)
				if (m_array[i])
					delete m_array[i];
		}
		
		m_size = size;
		/* No need to reallocate anything! */
	} else {
		if (size > m_capacity)
			reserve(size);
		for (size_type i = m_size; i < size; ++i)
			m_array[i] = val;
		m_size = size;
	}
}

/*************************************************************************/

} /* namespace containers */

/*************************************************************************/

#endif /* CONTAINERS_VECTORSP_H */
