/* Data collection.
 *
 * PegSoft databases library (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** \file databases/collection.h
  * \brief Database collection classes header file. */

#ifndef DATABASES_COLLECTION_H
#define DATABASES_COLLECTION_H

/*************************************************************************/

#include <list>

#include <containers/hashlist.h>
#include <containers/vectorsp.h>
#include <databases/record.h>

/*************************************************************************/

#ifndef ID_HASH_SIZE
# define ID_HASH_SIZE		4096
#endif

#ifndef NAME_HASH_SIZE
# define NAME_HASH_SIZE		1024
#endif

/*************************************************************************/

/// Namespace for databases-related declarations.

namespace databases
{

/*************************************************************************/

class Database;
class Field;

/*************************************************************************/

/// Database records collection class.

/** This is the class for database records collections, which keeps
  * a list of all records in a database, and allows new records to be
  * created, added or deleted.
  *
  * Adding a record to a collection will insert it in the database
  * if it wasn't done already. Removing a record from a collection
  * will erase it from the database if it was in it. Please note that
  * the collection may still be used without having an associated
  * database; that allows you to have collections that optionally get
  * stored in a database.
  *
  * The create() must be overriden in derived classes, the add(), add_list(),
  * add_db(), remove(), remove_list(), remove_db(), clear(), clear_list(),
  * clear_db(), begin_initial(), end_initial(), begin_import(), end_import()
  * and memory_usage() functions may be overriden too. */

class Collection
{
	protected:
		/// Size of the ID hash list
		static uint32_t const id_hashsize = ID_HASH_SIZE;
		/// Type of the ID hash list
		typedef containers::HashList<Record *, Record::identifier_type,
				id_hashsize> id_hashlist;
		
		/// Type of the owned collections list
		typedef std::list<Collection *> owned_list;
	public:
		/* Typedefs */
		
		/// Normal iterator type.
		typedef id_hashlist::iterator iterator;
		/// Const iterator type.
		typedef id_hashlist::const_iterator const_iterator;
		/// Reverse iterator
		typedef id_hashlist::reverse_iterator reverse_iterator;
		/// Const reverse iterator
		typedef id_hashlist::const_reverse_iterator const_reverse_iterator;
		
		/// Error constants
		enum
		{
			/// Unknown error 
			ceUnknown = 1,
			/// No more IDs available (if that happens... everything can happen).
			ceID = 2,
			/// List insertion/deletion error
			ceList = 3,
			/// Database error
			ceDatabase = 4,
			/// Custom errors start with this numbers (for overriders)
			ceCustom = 128
		};
		
		/* Functions */
	
		/// Constructor.
		Collection();
		/// Destructor.
		/** Removes the Record::dbsMember status flag from each record
		  * in the collection, then calls operator delete on them. */ 
		virtual ~Collection();
		
		/// Begins initial import.
		/** Begins initial import. This function must be called by
		  * database modules before they begin to create and add
		  * records during the initial import phase, even if there are
		  * no records to import.
		  *
		  * db().version(Database::vtImport) must already return the
		  * correct version for the database being loaded.
		  *
		  * The default implementation just sets the csInitial flag
		  * to the collection status and returns true.
		  *
		  * \return true if initial loading should continue, false
		  * otherwise. */
		virtual bool begin_initial() { status_add(csInitial); return true; }
		
		/// Ends initial import.
		/** Ends initial import. This function must be called by
		  * database modules after they finish to create and add records
		  * during the initial import phase, and must have been preceded
		  * by a call to begin_initial().
		  *
		  * db().version(Database::vtImport) must still return the correct
		  * version for the database that has been loaded until this 
		  * call is done.
		  *
		  * The default implementation just removes the csInitial flag from
		  * the collection status and returns true.
		  *
		  * \return true if successful, false otherwise. */
		virtual bool end_initial() { status_remove(csInitial); return true; }
		
		/// Begins import.
		/** Begins import. This function must be called by database modules
		  * before they begin to add, update or remove records after the
		  * initial import took place.
		  *
		  * \param[in] type The type of import that will take place, which
		  * must be one of the following:
		  *    - csCreation: creation of records
		  *    - csUpdate: update of records
		  *    - csDeletion deletion of records
		  *
		  * The default implementation just sets the csImport and type flags
		  * to the collection status and returns true.
		  *
		  * \return true if import should continue, false otherwise. */
		virtual bool begin_import(uint16_t type) 
				{ status_add(csImport | type); return true; }
		
		/// Ends import.
		/** Ends import. This function must be called by database modules
		  * after they finish to add new records after the initial import
		  * took place, and must have been preceded by a call to end_import().
		  *
		  * \param[in] type The type of import that took place, which
		  * must be one of the following:
		  *    - csCreation: creation of records
		  *    - csUpdate: update of records
		  *    - csDeletion deletion of records
		  *
		  * The default implementation just removes the csImport flag from
		  * the collection status and returns true.
		  *
		  * \return true if successful, false otherwise. */
		virtual bool end_import(uint16_t type)
				{ status_remove(csImport | type); return true; }
		
		/// Asks the collection to update new fields.
		/** Asks the collection to update any new fields that would need
		  * to be. This is called by the Database interface after
		  * it is ready to take updates for new fields, if the
		  * Database::dpExport permission is set and
		  * Database::writable() yields true.
		  *
		  * The default implementation just returns true.
		  * 
		  * \return true if successful, false otherwise. */
		virtual bool update_new() { return true; }
		
		/// Creates a new record suitable for this collection.
		/** Creates a new record suitable for this collection.
		  *
		  * \param[in] id The identifier of the new record. Use
		  * the value 0 if you want the identifier to be computed
		  * automatically when the record is added to the collection.
		  * \return A newly-allocated record, or 0 if the record
		  * couldn't be created. This record should be freed using
		  * operator delete when it is no longer needed. */
		virtual Record *create(Record::identifier_type id = 0) const = 0;
		
		/// Adds a record to the collection.
		/** Adds a record to the collection. No copy of the record is
		  * made, and so it must remain valid until it is removed
		  * from the collection. The unique identifier of the record
		  * is set by this function if it's not been set already.
		  *
		  * The Record::dbsMember and Record::dbsInsert flags will be
		  * added to the record status, and the csChanges flag to the
		  * collection status.
		  *
		  * \param[in] record The record to add.
		  * \param[in] err If non-NULL and the function returns false,
		  * this variable will contain the error number (which can be
		  * ceUnknown, ceID, ceList, ceDatabase or other numbers defined by
		  * derived classes).
		  * \return true if the record has been added to the list, false
		  * otherwise. */
		virtual bool add(Record *record, int *err = 0);
		
		/// Changes a string value of a record.
		/** Changes a string value of a record. This will call
		  * Database::update_s() provided that there is a database set
		  * to the collection, that the record has already been added
		  * to the collection, that the record is not being updated
		  * by a database module and that the field is listed as exportable
		  * by the database module. If these conditions are not met,
		  * the function just returns true.
		  *
		  * \param[in] record The record that changes.
		  * \param[in] field The record's field that changes.
		  * \param[in] value The new value of for the field of the record.
		  * \return true if the preconditions are not met or the
		  * database has been updated successfully, false otherwise. */
		bool change_s(Record *record, Field const *field, char const *value);
		
		/// Changes an integer value of a record.
		/** Changes an integer value of a record. This will call
		  * Database::update_i() provided that there is a database set
		  * to the collection, that the record has already been added
		  * to the collection, that the record is not being updated
		  * by a database module and that the field is listed as exportable
		  * by the database module. If these conditions are not met,
		  * the function just returns true.
		  *
		  * \param[in] record The record that changes.
		  * \param[in] field The record's field that changes.
		  * \param[in] value The new value of for the field of the record.
		  * \return true if the preconditions are not met or the
		  * database has been updated successfully, false otherwise. */
		bool change_i(Record *record, Field const *field, int32_t value);
		
		/// Changes an unsigned integer value of a record.
		/** Changes an unsigned integer value of a record. This will call
		  * Database::update_ui() provided that there is a database set
		  * to the collection, that the record has already been added
		  * to the collection, that the record is not being updated
		  * by a database module and that the field is listed as exportable
		  * by the database module. If these conditions are not met,
		  * the function just returns true.
		  *
		  * \param[in] record The record that changes.
		  * \param[in] field The record's field that changes.
		  * \param[in] value The new value of for the field of the record.
		  * \return true if the preconditions are not met or the
		  * database has been updated successfully, false otherwise. */
		bool change_ui(Record *record, Field const *field, uint32_t value);
		
		/// Changes a Unix timestamp value of a record.
		/** Changes a Unix timestamp value of a record. This will call
		  * Database::update_ut() provided that there is a database set
		  * to the collection, that the record has already been added
		  * to the collection, that the record is not being updated
		  * by a database module and that the field is listed as exportable
		  * by the database module. If these conditions are not met,
		  * the function just returns true.
		  *
		  * \param[in] record The record that changes.
		  * \param[in] field The record's field that changes.
		  * \param[in] value The new value of for the field of the record.
		  * \return true if the preconditions are not met or the
		  * database has been updated successfully, false otherwise. */
		bool change_ut(Record *record, Field const *field, time_t value);
		
		/// Changes a record identifier value of a record.
		/** Changes a record identifier value of a record. This will call
		  * Database::update_ri() provided that there is a database set
		  * to the collection, that the record has already been added
		  * to the collection, that the record is not being updated
		  * by a database module and that the field is listed as exportable
		  * by the database module. If these conditions are not met,
		  * the function just returns true.
		  *
		  * \param[in] record The record that changes.
		  * \param[in] field The record's field that changes.
		  * \param[in] value The new value of for the field of the record.
		  * \return true if the preconditions are not met or the
		  * database has been updated successfully, false otherwise. */
		bool change_ri(Record *record, Field const *field,
				Record::identifier_type value);
		
		/// Changes a multiple values of a record.
		/** Changes a multiple values of a record. This will call
		  * Database::update_multi() provided that there is a database set
		  * to the collection, that the record has already been added
		  * to the collection and that the record is not being updated
		  * by a database module. If these conditions are not met,
		  * the function just returns true.
		  *
		  * \param[in] record The record that changes.
		  * \param[in] valcount The number of values that need to be
		  * changed.
		  * \param[in] ... For each value that needs to be changed,
		  * a Field pointer and a value of the given field type
		  * (char * for Field::ftString, int32_t for Field::ftInteger,
		  * uint32_t for Field::ftUnsigned, time_t for Field::ftUnixTime
		  * and Record::identifier_type for Field::ftIdentifier).
		  * \return true if the preconditions are not met or the
		  * database has been updated successfully, false otherwise. */
		bool change_multi(Record *record, unsigned int valcount, ...);
		
		/// Changes owned records.
		/** Changes record identifiers values of all records in the
		  * collection whose current value match a given ID. This
		  * will call Database::update_owned() provided that there is a
		  * database set to the collection and that the field is listed
		  * as exportable by the database module. If these conditions
		  * are not met, the function just returns true.
		  *
		  * \param[in] field The owner field.
		  * \param[in] id The record identifier of the owner.
		  * \param[in] new_id The new record identifier of the owner.
		  * \return true if the preconditions are not met or the
		  * database has been updated successfully, false otherwise. */
		bool change_owned(Field const *field, Record::identifier_type id,
				Record::identifier_type new_id);
		
		/// Removes a record from the collection.
		/** Removes a record from the collection. The record will
		  * be freed by the function with operator delete if the
		  * removal is successful.
		  *
		  * The csChanges flag will be added to the collection status.
		  *
		  * \param[in] record The record to remove.
		  * \param[in] err If non-NULL and the function returns false,
		  * this variable will contain the error number (which can be
		  * ceUnknown, ceList, ceDatabase or other numbers defined by
		  * derived classes).
		  * \return true if the record has been removed from the list, false
		  * otherwise. */
		virtual bool remove(Record *record, int *err = 0);
		
		/// Removes owned records from the collection.
		/** Removes owned records from the collection. This function
		  * is called by the remove() function when a record of the
		  * owner collection is being removed.
		  *
		  * The default implementation calls remove_owned_db(), then
		  * if it is successful, calls remove_owned_list() and returns
		  * its return value.
		  *
		  * \param[in] coll The owner collection.
		  * \param[in] record The record of the owner collection being
		  * removed.
		  * \param[in] err If non-NULL and the function returns false,
		  * this variable will contain the error number (which can be
		  * ceUnknown, ceList, ceDatabase or other numbers defined by
		  * derived classes).
		  * \return true if successful, false otherwise. */
		virtual bool remove_owned(Collection const *coll,
				Record const *record, int *err = 0);
		
		/// Clears the collection.
		/** Clears the collection. The function will call operator
		  * delete on each record in the collection.
		  *
		  * \param[in] err If non-NULL and the function returns false,
		  * this variable will contain the error number (which can be
		  * ceUnknown, ceList, ceDatabase or other numbers defined by
		  * derived classes).
		  * \return true if successful, false otherwise. */
		virtual bool clear(int *err = 0);
		
		/// Finds a record in the collection.
		/** Finds a record in the collection.
		  *
		  * \param[in] id The unique identifier of the record to find.
		  * \return A pointer to the found record, or 0 if it was
		  * not found. */
		Record *find(Record::identifier_type id);
		/// Const version of find.
		Record const *find(Record::identifier_type id) const;
		
		/// Sets an owned collection.
		/** Sets an owned collection, i.e. a collection whose remove_owned()
		  * and clear_owned() functions shall be called respectively when
		  * a record is removed from this collection and when this collection
		  * is cleared. 
		  *
		  * \param[in] coll The owned collection. No copy is done and so it
		  * must remain valid until either unset_owned is called or this
		  * collection is destroyed. */
		void set_owned(Collection &coll) { m_owned.push_front(&coll); }
		
		/// Unsets an owned collection.
		/** Unsets an owned collection, that has been previously set through
		  * set_owned().
		  *
		  * \param[in] coll The collection. */
		void unset_owned(Collection &coll) { m_owned.remove(&coll); }
		
		/// Returns an iterator for the beginning of the collection.
		iterator begin() { return m_members.begin(); }
		/// Returns an iterator to the end of the collection.
		iterator end() { return m_members.end(); }
		/// Returns a const iterator for the beginning of the collection.
		const_iterator begin() const { return m_members.begin(); }
		/// Returns a const iterator to the end of the collection.
		const_iterator end() const { return m_members.end(); }
		
		/// Returns a reverse iterator to the beginning of the collection.
		reverse_iterator rbegin() { return m_members.rbegin(); }
		/// Returns a reverse iterator to the end of the collection.
		reverse_iterator rend() { return m_members.rend(); }
		/// Returns a const reverse iterator to the beginning of the collection.
		const_reverse_iterator rbegin() const { return m_members.rbegin(); }
		/// Returns a const reverse iterator to the end of the collection.
		const_reverse_iterator rend() const { return m_members.rend(); }
		
		/// Determines whether the collection is empty.
		bool empty() const { return m_members.empty(); }
		/// Returns the number of records in the collection.
		id_hashlist::size_type size() const { return m_members.size(); }
		
		/// Returns the memory usage of the collection.
		/** Returns the memory usage of the collection. The 
		  * Record::memory_usage() function of each record in the collection
		  * is called to form the final result.
		  *
		  * \return The memory usage of the collection, in bytes. */
		virtual size_t memory_usage() const;
		
		/// Possible status flags.
		enum
		{
			/// There are changes in the collection (since the flag was last cleared).
			csChanges  = 0x0001,
			/// Initial import in progress
			csInitial  = 0x0002,
			/// Import in progress
			csImport   = 0x0004,
			/// Creation of records in progress (with csImport set)
			csCreation = 0x0008,
			/// Update of records in progress (with csImport set)
			csUpdate   = 0x0010,
			/// Deletion of records in progress (with csImport set)
			csDeletion = 0x0020
		};
		
		/// Returns the database associated to the collection (if any).
		Database *db() { return m_db; }
		/// Sets the database associated to the collection.
		/** Sets the database associated to the collection. The collection
		  * will then insert, update and delete records through it when the
		  * corresponding functions are called.
		  *
		  * \param[in] db The database. This may be NULL if you don't want
		  * to store the collection records in a database.
		  * \return The database associated with the collection. */
		Database *db(Database *db) { return (m_db = db); }
		
		/// Returns whether the collection is readable.
		/** Returns whether the collection is readable. This just
		  * returns db()->readable(), unless no database has been set,
		  * in which case it always returns true.
		  *
		  * \return true if the collection is readable, false otherwise. */
		bool readable() const;
		
		/// Returns whether the collection is writable.
		/** Returns whether the collection is writable. This just
		  * returns db()->writable(), unless no database has been set,
		  * in which case it always returns true.
		  *
		  * \return true if the collection is writable, false otherwise. */
		bool writable() const;
				
		/// Returns the owner field corresponding to a owner collection.
		/** Returns the owner field of the records of this collection
		  * corresponding to the given owner collection.
		  *
		  * The default implementation returns 0, and this must be
		  * changed if this collection is owned by another collection.
		  *
		  * \param[in] coll The owner collection.
		  * \return A pointer to the owner field, or 0 if there is no
		  * owner field for this collection (which theorically NEVER
		  * happens when the Collection::remove_owned_db() function calls
		  * this method.)
		  * \sa remove_owned(). */
		virtual Field const *owner_field(Collection const *coll) const
				{ return 0; }
		
		/// Returns the collection status.
		uint16_t status() const { return m_status; }
		/// Adds flags to the status of the collection.
		/** Adds flags to the status of the collection.
		  *
		  * \param[in] flags The flags to add.
		  * \return The new collection status. */ 
		uint16_t status_add(uint16_t flags) { return (m_status |= flags); }
		/// Removes flags from the status of the collection.
		/** Removes flags from the status of the collection.
		  *
		  * \param[in] flags The flags to remove.
		  * \return The new collection status. */ 
		uint16_t status_remove(uint16_t flags) { return (m_status &= ~flags); }
		
		/// Returns the last assigned unique identifier of the collection.
		Record::identifier_type last_id() const { return m_last_id; }
	protected:
		/// The collection members, searchable by ID.
		id_hashlist m_members;
		/// The database associated with the collection.
		Database *m_db;
		/// The collection status.
		uint16_t m_status;
		/// The last assigned identifier.
		Record::identifier_type m_last_id;
		
		/// The owned collections.
		owned_list m_owned;
		
		/// Adds the record to the list(s).
		/** Adds the record to the list(s). This is called by the add()
		  * function, and only when Record::dbsMember is not set for
		  * the record.
		  *
		  * \param[in] record The record to add.
		  * \return true if successful, false otherwise. */
		virtual bool add_list(Record *record);
		
		/// Adds the record to the database.
		/** Adds the record to the database. This is called by the add()
		  * function, and only when Record::dbsInsert is not set for
		  * the record.
		  *
		  * \param[in] record The record to add.
		  * \return true if successful, false otherwise. */
		virtual bool add_db(Record *record);
		
		/// Removes owned records from the database.
		/** Removes owned records from the database. This function
		  * is called by the remove_owner() function.
		  *
		  * The default implementation calls owner_field() to
		  * know which field corresponds to the owner collection, then
		  * calls db()->delete_owned() to remove records owned by
		  * the deleted record.
		  *
		  * \param[in] coll The owner collection.
		  * \param[in] record The record of the owner collection being
		  * removed.
		  * \param[in] err If non-NULL and the function returns false,
		  * this variable will contain the error number (which can be
		  * ceUnknown, ceList, ceDatabase or other numbers defined by
		  * derived classes).
		  * \return true if successful, false otherwise. */
		virtual bool remove_owned_db(Collection const *coll,
				Record const *record, int *err = 0);
		
		/// Removes owned records from the lists.
		/** Removes owned records from the lists. This function
		  * is called by the remove_owned() function.
		  *
		  * The default implementation calls remove() on each record of
		  * the collection that are determined to be owned by the
		  * deleted record through a call to Record::owned() (the
		  * Record::dbsInsert flag is removed from the record status
		  * before the remove() function is called).
		  *
		  * \param[in] coll The owner collection.
		  * \param[in] record The record of the owner collection being
		  * removed.
		  * \param[in] err If non-NULL and the function returns false,
		  * this variable will contain the error number (which can be
		  * ceUnknown, ceList, ceDatabase or other numbers defined by
		  * derived classes).
		  * \return true if successful, false otherwise. */
		virtual bool remove_owned_list(Collection const *coll,
				Record const *record, int *err = 0);
		
		/// Removes the record from the list(s).
		/** Removes the record from the list(s). This is called by the
		  * remove() function, and only when Record::dbsMember is set for
		  * the record.
		  *
		  * This function can't fail to do its job!
		  *
		  * \param[in] record The record to add. */
		virtual void remove_list(Record *record);
		
		/// Removes the record from the database.
		/** Removes the record from the database. This is called by the
		  * remove() function, and only when Record::dbsInsert is set for
		  * the record.
		  *
		  * \param[in] record The record to add.
		  * \return true if successful, false otherwise. */
		virtual bool remove_db(Record *record);
		
		/// Clears all owned records from the collection.
		/** Clears all owned records from the collection. This function
		  * is called by the clear() function when the owner collection
		  * is being cleared.
		  *
		  * The default implementation just calls clear().
		  *
		  * \param[in] coll The owner collection.
		  * \param[in] err If non-NULL and the function returns false,
		  * this variable will contain the error number (which can be
		  * ceUnknown, ceList, ceDatabase or other numbers defined by
		  * derived classes).
		  * \return true if successful, false otherwise. */
		virtual bool clear_owned(Collection const *coll, int *err = 0);
		
		/// Removes all records from the list(s).
		/** Removes all records from the list(s). This is called by the
		  * clear() function.
		  *
		  * This function can't fail to do its job! */
		virtual void clear_list();
		
		/// Removes all records from the database.
		/** Removes all records from the database. This is called by the
		  * clear() function.
		  *
		  * \return true if successful, false otherwise. */
		virtual bool clear_db();
	private:
		Collection(Collection const &);
		Collection &operator=(Collection const &);
};

/*************************************************************************/

/// A collection for named records.

/** This collection adds methods to search records by names. It works on
  * records derived from NameRecord. */

class NameCollection : public Collection
{
	protected:
		/// Size of the name hash list
		static uint32_t const name_hashsize = NAME_HASH_SIZE;
		/// Type of the name hash list
		typedef containers::HashList<NameRecord *, char const *,
				name_hashsize> name_hashlist;
	public:
		/* Typedefs */
		
		/// Type of the name hash.
		typedef containers::Hash<NameRecord *, char const *> name_hash;
		
		/// Normal name iterator type.
		typedef name_hashlist::iterator niterator;
		/// Const name iterator type.
		typedef name_hashlist::const_iterator const_niterator;
		/// Reverse name iterator type.
		typedef name_hashlist::reverse_iterator reverse_niterator;
		/// Const reverse name iterator type.
		typedef name_hashlist::const_reverse_iterator const_reverse_niterator;
		
		/* Functions */
		
		/// Constructor.
		/** Constructor.
		  *
		  * \param[in] hash The hash to use with the name hashlist. */
		explicit NameCollection(name_hash &hash) : Collection(),
				m_nmembers(hash) { }
		
		/// Notifies the collection that a record's name has changed.
		/** Notifies the collection that a record's name has changed. The
		  * collection will then update itself so that it can find the
		  * record through its new name.
		  *
		  * \param[in] record The record, whose name() function should
		  * return its old name. If (record == newname) yields true, then
		  * the function won't do anything.
		  * \param[in] newname The new name of the record.
		  * \return true if successful, false otherwise. In the latter
		  * case, the caller should cancel the name change as the record
		  * will STILL be in the collection under its old name. */
		virtual bool change_name(NameRecord *record, char const *newname);
		
		/// Returns an iterator for the beginning of the name collection.
		niterator nbegin() { return m_nmembers.begin(); }
		/// Returns an iterator to the end of the name collection.
		niterator nend() { return m_nmembers.end(); }
		/// Returns a const iterator for the beginning of the name collection.
		const_niterator nbegin() const { return m_nmembers.begin(); }
		/// Returns a const iterator to the end of the name collection.
		const_niterator nend() const { return m_nmembers.end(); }
		
		/// Returns a reverse iterator to the beginning of the name collection.
		reverse_niterator nrbegin() { return m_nmembers.rbegin(); }
		/// Returns a reverse iterator to the end of the name collection.
		reverse_niterator nrend() { return m_nmembers.rend(); }
		/// Returns a const reverse iterator to the beginning of the name collection.
		const_reverse_niterator nrbegin() const { return m_nmembers.rbegin(); }
		/// Returns a const reverse iterator to the end of the name collection.
		const_reverse_niterator nrend() const { return m_nmembers.rend(); }
		
		/// Same as Collection::memory_usage().
		virtual size_t memory_usage() const;
		
		/// Finds a record in the collection, by name.
		/** Finds a record in the collection, by name.
		  *
		  * \param[in] name The name of the record to find.
		  * \return A pointer to the found record, or 0 if it was
		  * not found. */
		NameRecord *nfind(char const *name);
		/// Const version of nfind().
		NameRecord const *nfind(char const *name) const;
	protected:
		/// The collection members, searchable by names.
		name_hashlist m_nmembers;
		
		/// Same as Collection::add_list().
		virtual bool add_list(Record *record);
		
		/// Same as Collection::remove_list().
		virtual void remove_list(Record *record);
		
		/// Same as Collection::clear_list().
		virtual void clear_list();
};

/*************************************************************************/

/// A collection that allows access to its members through a VectorSP.

/** This collection will add its records to a VectorSP object, that
  * can be used to iterate through the collection. */

class ArrayCollection : public Collection
{
	public:
		/* Typedefs */
		
		/// Type of the array.
		typedef containers::VectorSP<Record *, uint16_t> array_type;
		
		/* Constants */
		
		/// More error constants
		enum
		{
			/// A record with the same name already exists in the collection.
			ceName = 64
		};
		
		/* Functions */
		
		/// Constructor.
		/** Constructor.
		  *
		  * \param[in] named Determines whether the records in the collection
		  * are derived from NameRecord. */
		explicit ArrayCollection(bool named) : Collection(), m_array(),
				m_named(named) { }
		
		/// Overrides Collection::add().
		/** This function does the same thing as Collection::add(), but,
		  * if the collection contains records derived from NameRecord (as
		  * specified in the constructor), *err is set to ceName and the
		  * function fails.
		  *
		  * \param[in] record The record to add.
		  * \param[in] err If non-NULL and the function returns false,
		  * this variable will contain the error number (which can be
		  * ceUnknown, ceID, ceList, ceDatabase, ceName or other numbers
		  * defined by derived classes).
		  * \return true if the record has been added to the list, false
		  * otherwise. */
		virtual bool add(Record *record, int *err = 0);
		
		/// Same as Collection::memory_usage().
		virtual size_t memory_usage() const;
		
		/// Returns the array.
		/** Returns the array filled with the records of the collection.
		  * Note that this array may contain "holes", i.e. NULL pointers,
		  * for efficiency reasons.
		  *
		  * You should not add or remove records to/from the array, even
		  * though you may change records in the array, of course.
		  *
		  * \return The array. */
		array_type &array() { return m_array; }
		
		/// Returns whether the collection contains records derived from NameRecord.
	protected:
		/// An array containing the records, with possible holes (i.e. NULL pointers).
		array_type m_array;
		/// Whether the collection contains records derived from NameRecord.
		bool m_named;
		
		/// Same as Collection::add_list().
		virtual bool add_list(Record *record);
		
		/// Same as Collection::remove_list().
		virtual void remove_list(Record *record);
		
		/// Same as Collection::clear_list().
		virtual void clear_list();
};

/*************************************************************************/

} /* namespace databases */

/*************************************************************************/

#endif /* DATABASES_COLLECTION_H */
