/* Data records.
 *
 * PegSoft databases library (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** \file databases/record.h
  * \brief Database record classes header file. */

#ifndef DATABASES_RECORD_H
#define DATABASES_RECORD_H

/*************************************************************************/
	
namespace databases
{

/*************************************************************************/

class Collection;
class Field;

/*************************************************************************/

/// Database record base class.

/** This is a base class for all database records.
  *
  * Each record gets an unique numeric identifier that may be used to
  * refer to it easily.
  *
  * The begin_update(), get_s(), set_s(), get_i(), set_i(), get_ui(), set_ui(),
  * get_ut(), set_ut(), get_ri(), set_ri(), end_update(), owned()
  * and memory_usage() functions may be overriden in derived classes. */

class Record
{
	public:
		/// Type of the record identifier.
		typedef uint32_t identifier_type;
		
		/// The maximum value for an identifier.
		static uint32_t const max_id = 0xFFFFFFFF;
		
		/// Converts string record identifier to its integer representation.
		static inline identifier_type strtori(char const *nptr, char **endptr,
				int base) { return strtoul(nptr, endptr, base); }

		/// The format to use to print identifier in printf-like functions.
#       define RECORD_FORMAT "%u"
		
		/// Default constructor.
		Record() : m_id(0), m_db_status(0) { }
		/// Constructor that sets the record identifier.
		explicit Record(identifier_type id) : m_id(id), m_db_status(0) { }
		/// Copy constructor.
		Record(Record const &r) : m_id(0), m_db_status(r.m_db_status) { }
		/// Assignment operator.
		Record &operator=(Record const &right);
		/// Destructor.
		virtual ~Record() { }
		
		/// Checks whether the given identifier is the same as the record identifier.
		bool operator==(identifier_type id) { return (m_id == id); }
		
		/// Begins record update.
		/** Begins record update. This function must be called by
		  * database modules before they start updating the record.
		  *
		  * Once this function has been called successfully, all
		  * fields in the record for which a value is available are
		  * expected to be updated, in ANY order (and not always in
		  * the order they were added to the database), and
		  * end_update() MUST be called even if there was an error
		  * in the update phase.
		  *
		  * The default implementation adds the flags dbsUpdating and
		  * dbsInsert (since it is _known_ that the record is indeed
		  * in the database) to the record status, and always returns
		  * true.
		  *
		  * \return true if the database module is allowed to
		  * continue its update, false if it must delete the record. */
		virtual void begin_update();
		
		/// Gets a string field.
		/** Gets a string field.
		  *
		  * The default implementation just returns some undefined value.
		  *
		  * \param[in] field The field to get.
		  * \return The value of the field. */
		virtual char const *get_s(Field const *field) const { return NULL; }
		
		/// Sets a string field.
		/** Sets a string field. This function must have been
		  * preceded by a call to begin_update().
		  *
		  * The default implementation does nothing else than returning
		  * the given parameter.
		  *
		  * \param[in] field The field to set.
		  * \param[in] value The new value for this field.
		  * \return The value of the field, which may be different than
		  * the given value; in this case, the database driver must update
		  * its stored data to reflect the real value of the field. */
		virtual char const *set_s(Field const *field, char const *value)
				{ return value; }
		
		/// Gets an integer field.
		/** Gets an integer field.
		  *
		  * The default implementation just returns some undefined value.
		  *
		  * \param[in] field The field to get.
		  * \return The value of the field. */
		virtual int32_t get_i(Field const *field) const { return 0; }
		
		/// Sets an integer field.
		/** Sets an integer field. This function must have been
		  * preceded by a call to begin_update().
		  *
		  * The default implementation does nothing else than returning
		  * the given parameter.
		  *
		  * \param[in] field The field to update.
		  * \param[in] value The new value for this field.
		  * \return The value of the field, which may be different than
		  * the given value; in this case, the database driver must update
		  * its stored data to reflect the real value of the field. */
		virtual int32_t set_i(Field const *field, int32_t value)
				{ return value; }
		
		/// Gets an unsigned integer field.
		/** Gets an unsigned integer field.
		  *
		  * The default implementation just returns some undefined value.
		  *
		  * \param[in] field The field to get.
		  * \return The value of the field. */
		virtual uint32_t get_ui(Field const *field) const { return 0; }
		
		/// Sets an unsigned integer field.
		/** Sets an unsigned integer field. This function must have been
		  * preceded by a call to begin_update().
		  *
		  * The default implementation does nothing else than returning
		  * the given parameter.
		  *
		  * \param[in] field The field to update.
		  * \param[in] value The new value for this field.
		  * \return The value of the field, which may be different than
		  * the given value; in this case, the database driver must update
		  * its stored data to reflect the real value of the field. */
		virtual uint32_t set_ui(Field const *field, uint32_t value)
				{ return value; }
		
		/// Gets a Unix timestamp field.
		/** Gets a Unix timestamp field.
		  *
		  * The default implementation just returns some undefined value.
		  *
		  * \param[in] field The field to get.
		  * \return The value of the field. */
		virtual time_t get_ut(Field const *field) const { return 0; }
		
		/// Sets a Unix timestamp field.
		/** Sets a Unix timestamp field. This function must have been
		  * preceded by a call to begin_update().
		  *
		  * The default implementation does nothing else than returning
		  * the given parameter.
		  *
		  * \param[in] field The field to update.
		  * \param[in] value The new value for this field.
		  * \return The value of the field, which may be different than
		  * the given value; in this case, the database driver must update
		  * its stored data to reflect the real value of the field. */
		virtual time_t set_ut(Field const *field, time_t value)
				{ return value; }
		
		/// Gets a record identifier field.
		/** Gets a record identifier field.
		  *
		  * The default implementation just returns some undefined value.
		  *
		  * \param[in] field The field to get.
		  * \return The value of the field. */
		virtual identifier_type get_ri(Field const *field) const { return 0; }
		
		/// Sets a record identifier field.
		/** Sets a record identifier field. This function must have been
		  * preceded by a call to begin_update().
		  *
		  * The default implementation does nothing else than returning
		  * the given parameter.
		  *
		  * \param[in] field The field to update.
		  * \param[in] value The new value for this field.
		  * \return The value of the field, which may be different than
		  * the given value; in this case, the database driver must update
		  * its stored data to reflect the real value of the field. */
		virtual identifier_type set_ri(Field const *field,
				identifier_type value) { return value; }
		
		/// Ends record update.
		/** Ends record update. This function must be called by
		  * database modules after they finish updating the record
		  * (begin_update() is therefore expected to have been called
		  * successfully before).
		  *
		  * The default implementation removes the flag dbsUpdating from
		  * the record status, and returns true, unless the dbsUpdateError
		  * flag has been set, in which case the flag is cleared and the
		  * function returns false.
		  *
		  * \return true if the update completed successfully, false
		  * if the database module must delete the record. */
		virtual bool end_update();
		
		/// Determines whether this record is owned by another record.
		/** Determines whether this record is owned by another record.
		  *
		  * The default implementation always returns false, so this
		  * has to be changed if the collection this records belongs to
		  * is owned by another collection.
		  *
		  * \param[in] coll The owner collection.
		  * \param[in] record The record of the owner collection being
		  * deleted.
		  * \return true if this record is owned by the given record,
		  * false otherwise.
		  * \sa remove_owned() */
		virtual bool owned(Collection const *coll, Record const *record)
				{ return false; }
		
		/// Returns the memory usage of the record.
		/** Returns the memory usage of the record.
		  *
		  * The default implementation always returns zero.
		  *
		  * \return The memory usage of the record. */
		virtual size_t memory_usage() const { return 0; }
		
		/// Returns the unique identifier of the record.
		identifier_type id() const { return m_id; }
		/// Sets the unique identifier of the record.
		/** Sets the unique identifier of the record.
		  *
		  * \param[in] id The unique identifier.
		  * \return The unique identifier of the record (which
		  * may be different than id if one was already set.) */
		identifier_type id(identifier_type id);
		
		/// Possible status flags.
		enum
		{
			/// The record is being updated by a database module.
			dbsUpdating = 0x0001,
			/// Error during the update phase
			dbsUpdateError = 0x0002,
			/// The record has been added to its collection.
			dbsMember = 0x0004,
			/// The record has been inserted in the database (and has not been deleted).
			dbsInsert = 0x0008
		};
		
		/// Returns the status of the record.
		uint16_t db_status() const { return m_db_status; }
		/// Adds flags to the status of the record.
		/** Adds flags to the status of the record.
		  *
		  * \param[in] flags The flags to add.
		  * \return The new record status. */ 
		uint16_t db_status_add(uint16_t flags)
				{ return (m_db_status |= flags); }
		/// Removes flags from the status of the record.
		/** Removes flags from the status of the record.
		  *
		  * \param[in] flags The flags to remove.
		  * \return The new record status. */ 
		uint16_t db_status_remove(uint16_t flags)
				{ return (m_db_status &= ~flags); }
	protected:
		/// The record identifier.
		identifier_type m_id;
		/// The record status
		uint16_t m_db_status;
};

/*************************************************************************/

/// Database named record class.

/** This record class adds a unique name to identify the record as
  * well as methods to compare names to the record.
  *
  * The name() function must be overriden in derived classes. The
  * matches() and operator==() functions may be overriden too. */

class NameRecord : public Record
{
	public:
		/// Default constructor.
		NameRecord() : Record() { }
		/// Constructor setting the record identifier.
		explicit NameRecord(identifier_type id) : Record(id) { }
		
		/// Checks whether a name is the same as the record's name.
		/** Checks whether the given name is the same as the
		  * record's name.
		  *
		  * The default implementation returns !strcasecmp(this->name(), name);
		  *
		  * \param[in] name The name to check.
		  * \return true if the two names are the same, false otherwise. */
		virtual bool operator==(char const *name)
				{ return !strcasecmp(this->name(), name); }
		
		/// Checks whether a pattern matches the record's name.
		/** Checks whether the given pattern checks the record's name.
		  * Implementation details are left entirely to the
		  * implementers.
		  *
		  * The default implementation returns misc::match(pattern, name());
		  *
		  * \param[in] pattern The pattern.
		  * \return true if the name matches the pattern, false
		  * otherwise. */
		virtual bool matches(char const *pattern);
		
		/// Returns the unique name of the record.
		virtual char const *name() const = 0;
	protected:
};

/*************************************************************************/

} /* namespace databases */

/*************************************************************************/

#endif /* DATABASES_RECORD_H */
