/* Database<->XML.
 *
 * PegSoft databases library (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#include <algorithm>
#include <map>

#include <tinyxml/tinyxml.h>

#include "databases/xml.h"

using namespace std;

/*************************************************************************/
	
namespace databases
{

/*************************************************************************/

XMLIndex &XMLIndex::operator=(XMLIndex const &right)
{
	if (this != &right)
	{
		m_files = right.m_files;
		m_error = right.m_error;
	}
	
	return *this;
}

/*************************************************************************/

bool XMLIndex::add(std::string const &filename)
{
	container_type::const_iterator it;
	
	it = std::find(m_files.begin(), m_files.end(), filename);
	if (it != m_files.end())
		return false;
	
	m_files.push_back(filename);
	return true;
}

/*************************************************************************/

void XMLIndex::clear()
{
	m_files.clear();
}

/*************************************************************************/

bool XMLIndex::exists(std::string const &filename) const
{
	container_type::const_iterator it;
	
	it = std::find(m_files.begin(), m_files.end(), filename);
	return (it != m_files.end());
}

/*************************************************************************/

bool XMLIndex::load(std::string const &filename)
{
	TiXmlDocument doc(filename.c_str());
	
	if (!doc.LoadFile(TIXML_ENCODING_LEGACY))
	{
		m_error = doc.ErrorDesc();
		return false;
	}
	
	m_files.clear();
	
	TiXmlElement *index = doc.RootElement();
	TiXmlElement *file = index->FirstChildElement("file");
	
	for (; file; file = file->NextSiblingElement("file"))
	{
		TiXmlText *text = file->FirstChild()->ToText();
		
		if (!text || !add(text->Value()))
		{
			m_error = "Invalid index file";
			return false;
		}
	}
	
	return true;
}

/*************************************************************************/

bool XMLIndex::remove(std::string const &filename)
{
	container_type::iterator it;
	
	it = std::remove(m_files.begin(), m_files.end(), filename);
	
	if (it == m_files.end())
		return false;
	
	m_files.erase(it, m_files.end());
	return true;
}

/*************************************************************************/

bool XMLIndex::save(std::string const &filename)
{
	TiXmlDocument doc(filename.c_str());
	
	TiXmlDeclaration decl("1.0", "", "yes");
	if (!doc.InsertEndChild(decl))
	{
		m_error = "XML construction error";
		return false;
	}
	
	TiXmlElement index("index");
	
	const_iterator it, last;
	
	for (it = begin(), last = end(); it != last; ++it)
	{
		TiXmlText text(it->c_str());
		TiXmlElement file("file");
		
		if (!file.InsertEndChild(text) || !index.InsertEndChild(file))
		{
			m_error = "XML construction error";
			return false;
		}
	}
	
	if (!doc.InsertEndChild(index))
	{
		m_error = "XML construction error";
		return false;
	}
	
	if (!doc.SaveFile())
	{
		m_error = doc.ErrorDesc();
		return false;
	}
	
	return true;
}

/*************************************************************************/

XMLReader &XMLReader::operator=(XMLReader const &right)
{
	if (this != &right)
		m_error = "";
	
	return *this;
}

/*************************************************************************/

bool XMLReader::load(std::string const &filename, Callback &cb)
{
	TiXmlDocument doc(filename.c_str());
	
	/* Load the file */
	if (!doc.LoadFile(TIXML_ENCODING_LEGACY))
	{
		m_error = doc.ErrorDesc();
		return false;
	}
	
	/* Find the base */
	TiXmlElement *base = doc.RootElement();
	if (!base || stricmp(base->Value(), "base"))
	{
		m_error = "Missing tag <base>";
		return false;
	}
	
	/* Find the header and data */
	TiXmlElement *header = base->FirstChildElement("header");
	TiXmlElement *data = base->FirstChildElement("data");
	
	if (!header || !data)
	{
		m_error = "Missing tags <header> and/or <data>";
		return false;
	}
	
	/* Set database information */
	TiXmlElement *database = header->FirstChildElement("database");

	if (!database)
	{
		m_error = "Missing tag <database>";
		return false;
	}
	
	int version;

	if (database->QueryIntAttribute("version", &version) != TIXML_SUCCESS)
	{
		m_error = "<database> tag is invalid.";
		return false;
	}
	
	if (!cb.database(version))
		return false;
	
	/* Add field information */
	map<string, uint16_t> ftype;
	
	char const *name, *defval;
	int type, size, attrs;
	
	TiXmlElement *field = header->FirstChildElement("field");
	
	for (; field; field = field->NextSiblingElement("field"))
	{
		if (!(name = field->Attribute("name")) ||
				!(defval = field->Attribute("default")) ||
				field->QueryIntAttribute("type", &type) != TIXML_SUCCESS ||
				field->QueryIntAttribute("size", &size) != TIXML_SUCCESS ||
				field->QueryIntAttribute("attrs", &attrs) != TIXML_SUCCESS)
		{
			m_error = "<field> tag is invalid.";
			return false;
		}
		
		if (!*defval)
			defval = NULL;
		
		if (!cb.add_field(name, type, size, attrs, defval))
			return false;
		
		ftype[name] = type;
	}
	
	/* Set records count */
	TiXmlElement *elcount = data->FirstChildElement("count");
	
	if (!elcount)
	{
		m_error = "Missing tag <count>";
		return false;
	}
	
	TiXmlText *txcount = elcount->FirstChild()->ToText();
	
	if (!txcount)
	{
		m_error = "Missing value for tag <count>";
		return false;
	}
	
	char *end;
	uint32_t count = strtoul(txcount->Value(), &end, 10);
	
	if (*end)
	{
		m_error = "Invalid value for tag <count>";
		return false;
	}
	
	if (!cb.count(count))
		return false;
	
	/* Don't look for records if count is zero. */
	if (!count)
		return true;
	
	uint32_t rcount = 0;
	
	TiXmlElement *record = data->FirstChildElement("record");
	
	for (; record; record = record->NextSiblingElement("record"))
	{
		if (!cb.begin_record())
			return false;
		
		uint16_t vtype;
		map<string, uint16_t>::const_iterator it;
		
		TiXmlElement *value = record->FirstChildElement();
		
		for (; value; value = value->NextSiblingElement())
		{
			if (!stricmp(value->Value(), "id")) {
				vtype = Field::ftIdentifier;
			} else if ((it = ftype.find(value->Value())) != ftype.end()) {
				vtype = it->second;
			} else 
			{	
				m_error = string("Unknown value ") + value->Value() + " in record";
				return false;
			}
			
			TiXmlText *txvalue = value->FirstChild()->ToText();
			if (!txvalue && vtype != Field::ftString)
			{
				m_error = string("Missing data for value ") + value->Value();
				return false;
			}
			
			switch (vtype)
			{
				case Field::ftString:
					if (!txvalue || !*txvalue->Value())
					{
						if (!cb.add_value(value->Value(), NULL))
							return false;
					} else
					{
						if (!cb.add_value(value->Value(), txvalue->Value()))
							return false;
					}
					break;
				case Field::ftInteger:
				{
					int32_t ivalue = strtol(txvalue->Value(), &end, 10);
					if (*end)
					{
						m_error = string("Bad type in data of value ") + value->Value();
						return false;
					}
					
					if (!cb.add_ivalue(value->Value(), ivalue))
						return false;
				}
					break;
				case Field::ftUnsigned:
				{
					uint32_t uivalue = strtoul(txvalue->Value(), &end, 10);
					if (*end)
					{
						m_error = string("Bad type in data of value ") + value->Value();
						return false;
					}
					
					if (!cb.add_uivalue(value->Value(), uivalue))
						return false;
				}
					break;
				case Field::ftUnixTime:
				{
					time_t utvalue = strtol(txvalue->Value(), &end, 10);
					if (*end)
					{
						m_error = string("Bad type in data of value ") + value->Value();
						return false;
					}
					
					if (!cb.add_utvalue(value->Value(), utvalue))
						return false;
				}
					break;
				case Field::ftIdentifier:
				{
					Record::identifier_type rivalue =
							Record::strtori(txvalue->Value(), &end, 10);
					if (*end)
					{
						m_error = string("Bad type in data of value ") + value->Value();
						return false;
					}
					
					if (!cb.add_rivalue(value->Value(), rivalue))
						return false;
				}
					break;
				default:
					m_error = string("Unknown type for field ") + value->Value();
					return false;
			}
		}
		
		if (!cb.end_record())
			return false;
		
		++rcount;
	}
	
	if (count != rcount)
	{
		m_error = "Incorrect number of records";
		return false;
	}
	
	return true;
}

/*************************************************************************/

XMLWriter::~XMLWriter()
{
	if (m_record)
		delete m_record;
}

/*************************************************************************/

XMLWriter &XMLWriter::operator=(XMLWriter const &right)
{
	if (this != &right)
	{
		m_doc = right.m_doc;
		m_header = right.m_header;
		m_data = right.m_data;
		if (m_record)
			delete m_record;
		m_record = (right.m_record ? new TiXmlElement(*right.m_record) : 0);
	}
	
	return *this;
}

/*************************************************************************/

bool XMLWriter::add_field(char const *name, uint16_t type, size_t size,
		uint16_t attrs, char const *defval)
{
	TiXmlElement field("field");
	
	field.SetAttribute("name", name);
	field.SetAttribute("type", type);
	field.SetAttribute("size", size);
	field.SetAttribute("attrs", attrs);
	field.SetAttribute("default", (defval ? defval : ""));
	
	if (!m_header.InsertEndChild(field))
		return false;
	
	return true;
}

/*************************************************************************/

bool XMLWriter::add_value(char const *field, char const *value)
{
	if (!m_record)
		return false;
	
	/* NULL values are written as "" */
	if (!value)
		value = "";
	
	TiXmlElement name(field);
	
	TiXmlText txvalue(value);
	if (!name.InsertEndChild(txvalue))
		return false;
	
	if (!m_record->InsertEndChild(name))
		return false;
	
	return true;
}

/*************************************************************************/

bool XMLWriter::add_ivalue(char const *field, int32_t value)
{
	char sval[12];
	snprintf(sval, sizeof(sval), "%d", value);
	
	return add_value(field, sval);
}

/*************************************************************************/

bool XMLWriter::add_rivalue(char const *field, Record::identifier_type value)
{
	char sval[12];
	snprintf(sval, sizeof(sval), RECORD_FORMAT, value);
	
	return add_value(field, sval);
}

/*************************************************************************/

bool XMLWriter::add_uivalue(char const *field, uint32_t value)
{
	char sval[12];
	snprintf(sval, sizeof(sval), "%u", value);
	
	return add_value(field, sval);
}

/*************************************************************************/

bool XMLWriter::add_utvalue(char const *field, time_t value)
{
	char sval[12];
	snprintf(sval, sizeof(sval), "%ld", value);
	
	return add_value(field, sval);
}

/*************************************************************************/

bool XMLWriter::begin_record()
{
	if (m_record)
		return false;
	
	m_record = new TiXmlElement("record");
	return true;
}

/*************************************************************************/

bool XMLWriter::count(uint32_t count)
{
	if (m_data.FirstChildElement("count"))
		return false;
	
	char scount[12];
	snprintf(scount, sizeof(scount), "%u", count);

	TiXmlElement elcount("count");
	
	TiXmlText txcount(scount);
	if (!elcount.InsertEndChild(txcount))
		return false;
	
	if (!m_data.InsertEndChild(elcount))
		return false;
	
	return true;
}

/*************************************************************************/

void XMLWriter::dump_record()
{
	if (m_record)
	{
		delete m_record;
		m_record = 0;
	}
}

/*************************************************************************/

bool XMLWriter::end_record()
{
	if (!m_record || !m_data.InsertEndChild(*m_record))
		return false;
	
	dump_record();
	return true;
}

/*************************************************************************/

bool XMLWriter::save(std::string const &filename)
{	
	TiXmlDeclaration decl("1.0", "", "yes");
	if (!m_doc.InsertEndChild(decl))
		return false;
	
	TiXmlElement base("base");
	
	if (!base.InsertEndChild(m_header) || !base.InsertEndChild(m_data) ||
			!m_doc.InsertEndChild(base))
		return false;
	
	return m_doc.SaveFile(filename.c_str());
}

/*************************************************************************/

bool XMLWriter::database(uint16_t version)
{
	if (m_header.FirstChildElement("database"))
		return false;
	
	TiXmlElement database("database");
	database.SetAttribute("version", version);
	
	if (!m_header.InsertEndChild(database))
		return false;
	
	return true;
}

/*************************************************************************/

} /* namespace databases */

/*************************************************************************/
