/* Mail agent.
 *
 * PegSoft mail library (c) 2006 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#include <algorithm>
#include <iostream>
#include <sstream>

#include <containers/hashlist.h>
#include <misc/misc.h>

#include "mail/mail.h"
#include "mail/transport.h"

#include "mail.h"

using namespace std;

/*************************************************************************/

namespace
{

/*************************************************************************/

class QuotaHash : public containers::Hash<mail::QuotaInfo *, std::string>
{
	public:
		virtual int compare(std::string const &key,
				mail::QuotaInfo *const &value) const 
				{ return stricmp(key.c_str(), value->name().c_str()); }	
		virtual uint32_t hash(std::string const &key, uint32_t max) const
				{ return (misc::crc16_str(key.c_str(), 
					misc::ctLowerCase) % max); }
		virtual bool match(std::string const &key,
				mail::QuotaInfo *const &value) const
				{ return !stricmp(key.c_str(), value->name().c_str()); }
};

QuotaHash quota_hash;

/*************************************************************************/

typedef containers::HashList<mail::QuotaInfo *, std::string, 1024> quota_hlist;

quota_hlist senders(quota_hash);
quota_hlist recipients(quota_hash);

/*************************************************************************/

/* Options */

std::string sname;
std::string saddress;
std::string ssender;
misc::Ratio lauth;
misc::Ratio lunauth;
misc::Ratio slimit;
	
time_t keep_data;
time_t last_check = 0;
	
bool configured = false;

/* Transport */

mail::Transport *transport = 0;

/*************************************************************************/

std::string rewrite_name(std::string const &name)
{
	std::string newname;
	newname.reserve(name.length() * 2);
	
	std::string::const_iterator it, end;
	
	for (it = name.begin(), end = name.end(); it != end; ++it)
	{
		if (*it == '"')
			newname += "\\\"";
		else
			newname += *it;
	}
	
	return newname;
}

/*************************************************************************/

void purge_quota()
{	
	time_t min = time(0) - keep_data;
	
	if (last_check >= min)
		return;
	
	quota_hlist::iterator it, cur, end;
		
	for (it = recipients.begin(), end = recipients.end(); it != end; )
	{	
		cur = it++;
		mail::QuotaInfo *qi = (mail::QuotaInfo *) *cur;
		
		if (qi->times().empty() || qi->times().back() < min)
		{
			recipients.erase(cur);
			delete qi;
		}
	}
	
	for (it = senders.begin(), end = senders.end(); it != end; )
	{
		cur = it++;
		mail::QuotaInfo *qi = (mail::QuotaInfo *) *cur;
		
		if (qi->times().empty() || qi->times().back() < min)
		{
			senders.erase(cur);
			delete qi;
		}
	}
	
	last_check = time(0);
}

/*************************************************************************/

/* Adds a quota to the list */
mail::QuotaInfo *add_quota(std::string const &name, quota_hlist &hlist)
{
	/* NOW! Do quota purge... */
	purge_quota();
	
	mail::QuotaInfo *qi = new mail::QuotaInfo(name);
	
	if (hlist.add_sorted(name, qi) == hlist.end())
	{
		delete qi;
		return 0;
	}
	
	return qi;
}

/*************************************************************************/

/* Finds a matching quota in the list (creates the quota if necessary) */
mail::QuotaInfo *find_quota(std::string const &name, quota_hlist &hlist)
{
	quota_hlist::const_iterator it = hlist.find_sorted(name);
	
	if (it == hlist.end())
		return add_quota(name, hlist);
	
	return *it;
}

/*************************************************************************/

} /* anonymous namespace */

/*************************************************************************/

namespace mail
{

/*************************************************************************/

Mail &Mail::operator=(Mail const &right)
{
	if (this != &right)
	{
		m_from_name = right.m_from_name;
		m_from_address = right.m_from_address;
		m_subject = right.m_subject;
		m_sender = right.m_sender;
		m_to_name = right.m_to_name;
		m_to_address = right.m_to_address;
		m_origin = right.m_origin;
		m_text = right.m_text;
		m_data = right.m_data;
	}
	
	return *this;
}

/*************************************************************************/

bool Mail::configured()
{
	return (::configured && transport);
}

/*************************************************************************/

void Mail::generate_data()
{
	char const *CRLF = "\r\n";
	
	char timebuf[1024];
	
	time_t now = time(0);
	strftime(timebuf, sizeof(timebuf),
			"%a, %d %b %Y %H:%M:%S +0000", gmtime(&now));
	
	std::stringstream ds;
	
	ds << "Date: " << timebuf << CRLF;
	if (!m_from_name.empty())
		ds << "From: \"" << rewrite_name(m_from_name) << "\" <" <<
				m_from_address << ">" << CRLF;
	else
		ds << "From: " << m_from_address << CRLF;
	
	if (!m_subject.empty())
		ds << "Subject: " << m_subject << CRLF;
	
	if (m_sender != m_from_address)
		ds << "Sender: " << m_sender << CRLF;
	
	if (!m_to_name.empty())
		ds << "To: \"" << rewrite_name(m_to_name) << "\" <" <<
				m_to_address << ">" << CRLF;
	else
		ds << "To: " << m_to_address << CRLF;
	
	if (!m_origin.empty())
		ds << "X-Origin: " << m_origin << CRLF;
	
	ds << CRLF;
	
	text_type::const_iterator it, end;
	
	for (it = m_text.begin(), end = m_text.end(); it != end; ++it)
	{
		if (*it == ".")
			ds << " ." << CRLF;
		else
			ds << *it << CRLF;
	}
	
	m_data = ds.str();
}

/*************************************************************************/

bool Mail::send(uint32_t flags, uint32_t *err, std::string *errdesc)
{
	/* Make sure we're configured and transportable */
	if (!::configured || !transport)
	{
		set_error(err, errdesc, meConfig, "Mail subsystem not configured");
		return false;
	}
	
	/* Check for missing fields, either returning an error or filling
	 * them accordingly. */
	
	if (m_to_address.empty())
	{
		set_error(err, errdesc, meIncomplete, "Missing fields");
		return false;
	}
	
	if (m_from_name.empty() && m_from_address.empty() && m_sender.empty())
		m_from_name = sname;
	
	if (m_sender.empty())
		m_sender = saddress;
	
	if (m_from_address.empty())
		m_from_address = m_sender;
	
	if (m_origin.empty())
		m_origin = m_from_address;
	
	/* Check the quota */
	QuotaInfo *rq = find_quota(m_to_address, recipients);
	QuotaInfo *sq = find_quota(m_origin, senders);
	
	if (!(flags & mfNoQuota) &&
			(rq->exceeds((flags & mfAuthenticated) ? lauth : lunauth) ||
			sq->exceeds(slimit)))
	{
		set_error(err, errdesc, meQuota | meTemporary, "Send quota exceeded");
		return false;
	}
	
	/* Generate the mail data */
	
	generate_data();
	
	/* Sends the mail */
	
	bool temp = false;
	
	if (!transport->deliver(*this, &temp))
	{
		set_error(err, errdesc, (meDelivery | (temp ? meTemporary : 0)),
				std::string("Delivery error: ") + transport->error());
		return false;
	}
	
	/* Once the mail has been sent, mark the quota */
	rq->mark();
	sq->mark();
	
	set_error(err, errdesc, meSuccess, "Success");
	return true;
}

/*************************************************************************/

void Mail::set_error(uint32_t *err, std::string *errdesc, uint32_t myerr,
		std::string const &mydesc) const
{
	if (err)
		*err = myerr;
	if (errdesc)
		*errdesc = mydesc;
}

/*************************************************************************/

Mail::text_type const &Mail::text_add(std::string const &line)
{ 
	string::const_iterator it, start, end;
	
	for (it = line.begin(), start = it, end = line.end(); it != end; ++it)
	{
		if (*it == '\n' || *it == '\r')
		{
			m_text.push_back(string(start, it));
			
			/* Handle that stupid \r character */
			if (*it == '\r' && (it + 1) != end && *(it + 1) == '\n')
				++it;
			
			start = it + 1;
		}
	}
	
	if (start == line.begin())
		m_text.push_back(line);
	else if (start != end)
		m_text.push_back(string(start, end));
	 
	return m_text; 
}

/*************************************************************************/

Mail::text_type const &Mail::text_addf(char const *format, ...)
{
	char line[16384];
	
	va_list args;
	
	va_start(args, format);
	vsnprintf(line, sizeof(line), format, args);
	va_end(args);
	
	return text_add(line);
}

/*************************************************************************/

bool Mail::valid_address(char const *address, bool noplus)
{
	if (!address || !*address)
		return false;
	
	static bool specials[128], sp_initialized = false;
	
	if (!sp_initialized)
	{
		memset(&specials, 0, sizeof(specials));
	
		specials[(int)'('] = true;
		specials[(int)')'] = true;
		specials[(int)'<'] = true;
		specials[(int)'>'] = true;
		specials[(int)'@'] = true;
		specials[(int)','] = true;
		specials[(int)';'] = true;
		specials[(int)':'] = true;
		specials[(int)'\\'] = true;
		specials[(int)'\"'] = true;
		specials[(int)'['] = true;
		specials[(int)']'] = true;
		specials[(int)' '] = true;
		specials[(int)'('] = true;
		
		sp_initialized = true;
	}
	
	bool has_dots = false;
	char const *at = 0, *it, *end;
	
	for (it = address, end = address + strlen(address); it != end; ++it)
	{
		if (*it == '@')
		{
			if (at || it == address)
				return false;
			
			at = it;
			continue;
		}
		
		if (*it == '.' && at)
		{
			if (it == at + 1 || it == end - 1)
				return false;
			
			has_dots = true;
			continue;
		}
		
		if (*it <= 31 || *it >= 127 || specials[(int)*it])
			return false;
		
		if (noplus && !at && *it == '+')
			return false;
	}
	
	return at && has_dots;
}

/*************************************************************************/

bool QuotaInfo::exceeds(misc::Ratio const &ratio)
{
	purge(ratio);
	
	return (m_times.size() >= (size_t)ratio.x());
}

/*************************************************************************/

void QuotaInfo::mark()
{
	m_times.push_back(time(0));
}

/*************************************************************************/

void QuotaInfo::purge(misc::Ratio const &ratio)
{
	time_t min = time(0) - ratio.y();
	
	while (!m_times.empty())
	{
		if (m_times.front() < min)
			m_times.pop_front();
		else
			break;
	}
}

/*************************************************************************/

void conf_options(std::string const &sname_, std::string const &saddress_,
		misc::Ratio const &lauth_, misc::Ratio const &lunauth_,
		misc::Ratio const &slimit_)
{
	sname = sname_;
	saddress = saddress_;
	lauth = lauth_;
	lunauth = lunauth_;
	slimit = slimit_;
	
	if (!sname.empty())
	{
		char buf[4096];
		snprintf(buf, sizeof(buf), "\"%s\" <%s>", sname.c_str(),
				saddress.c_str());
		ssender = buf;
	} else {
		ssender = saddress;
	}
	
	keep_data = max(max(lauth.y(), lunauth.y()), slimit.y());
	
	if (keep_data < 3600)
		keep_data = 3600;
	
	last_check = time(0);
	
	configured = true;
}

/*************************************************************************/

void conf_transport(Transport *transport_)
{
	transport = transport_;
}

/*************************************************************************/

} /* namespace mail */

/*************************************************************************/
