/* Tronc commun.
 *
 * PegSoft modules subsystem library (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software (the COPYING file); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 */

/** 
 * \file modules/interface.h
 * \brief Interface header file.
 */

#ifndef MODULES_INTERFACE_H
#define MODULES_INTERFACE_H

/*************************************************************************/

namespace modules
{

/*************************************************************************/

/// IID type.
typedef uint32_t PSIID;

/*************************************************************************/

/// Common ancestor for interfaces.
/** Interface is the base class for all interfaces.
  *
  * Interfaces are used by modules to expose functionality that may or may
  * not be available depending on the module. They are obtained by other
  * modules through the Module::factory() and Module::interface() functions,
  * depending on whether the interface object returned is unique to the
  * module (and owned by it) of if it is owned by the caller.
  *
  * Each interface must have its own IID (Interface IDentifier), a unique
  * number that will be used for identification when a module is looking
  * for the availability of a specific interface. 
  *
  * Optionally, interfaces may override the supports() function to add a
  * way to query for optional feature support.
  */

class Interface
{
	public:
		/// The interface unique identifier.
		static PSIID const IID = 0x4227894c;
	
		/// Destructor.
		virtual ~Interface() { }
		
		/// Determines whether the interface supports an optional feature.
		/** Determines whether the interface supports an optional feature.
		  *
		  * The default implementation always returns 0.
		  *
		  * \param[in] fid The ID of the feature to test for availability. 
		  * The possible values for this parameter is up to the implementer.
		  * \return The interface implementing the feature (which may be
		  * the same as interface being queried, depending on the 
		  * implementation) if the feature is available, 0 otherwise.
		  */
		virtual Interface *supports(PSIID sid) { return 0; }
};

/*************************************************************************/

/// An interface with error reporting capabilities.
/** This interface class may be used to as a base class for interface
  * that want to report errors. Error messages are limited to 1024
  * characters. */

class ErrorInterface : public Interface
{
	public:
		/// The interface unique identifier.
		static PSIID const IID = 0x42e694d6;
		
		/// The maximum length of error messages.
		static size_t const error_len = 1024;
		
		/// Constructor.
		ErrorInterface() : Interface() { m_error[0] = '\0'; }
		/// Copy constructor (does not copy the error message).
		ErrorInterface(ErrorInterface const &ei) : Interface(ei)
				{ m_error[0] = '\0'; }
		/// Assignment operator (does not copy the error message).
		virtual ErrorInterface &operator=(ErrorInterface const &right);
		
		/// Returns the last error message.
		char const *error() const { return m_error; }
	protected:
		/// The last error that occured.
		char m_error[error_len];
		
		/// Sets the last error message, printf-like.
		void errorf(char const *format, ...) FORMAT(printf,2,3);
};

/*************************************************************************/

} /* namespace modules */

/*************************************************************************/

#endif /* MODULES_INTERFACE_H */
