/* IP sockets base class.
 *
 * PegSoft sockets library (c) 2007 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** \file sockets/ipsocket.h
  * \brief IP sockets base class header file. */

#ifndef SOCKETS_IPSOCKET_H
#define SOCKETS_IPSOCKET_H

/*************************************************************************/

#include <string>

#include <sockets/socket.h>

/*************************************************************************/

#if defined(AF_INET6) && HAVE_GETADDRINFO && HAVE_INET_NTOP
# define HAVE_IPV6 1
#endif

/*************************************************************************/

namespace sockets
{

/*************************************************************************/

/// Base class for all IP sockets.

/** This is the base class for all IP sockets. It must not be instantiated
  * directly; rather, use a derived class or create your own.
  *
  * This class provides support to set up and retrieve information about
  * the local and remote end of a connection, as well as a function to
  * bind the socket to a specific address and/or port.
  *
  * Both IPv4 and IPv6 are supported. */

class IPSocket : public Socket
{
	public:
		/// Constructor.
		/** Constructor.
		  *
		  * \param[in] type The socket type (one of the SOCK_...
		  * constants defined in the system header files). May be set to
		  * 0 if you want to set it later.
		  * \param[in] protocol The socket protocol that needs to be
		  * used. This is seldom used, as there is generally only one
		  * protocol for a given family/type combination; use the value
		  * 0 to use the default protocol. */
		explicit IPSocket(int type = 0, int protocol = 0) :
				Socket(PF_UNSPEC, type, protocol), m_local(0), m_remote(0) { }
		/// Copy constructor.
		IPSocket(IPSocket const &ips) : Socket(ips), 
				m_local(ips.m_local ? new IPInfo(*ips.m_local) : 0),
				m_remote(ips.m_remote ? new IPInfo(*ips.m_remote) : 0) { }
		/// Assignment operator.
		virtual IPSocket &operator=(IPSocket const &right);
		/// Destructor.
		~IPSocket();
		
		/// Assigns a socket file descriptor.
		/** Assigns a socket file descriptor to the socket object,
		  * that is then used by the object for all operations on
		  * the socket.
		  *
		  * This function overrides Socket::assign() to set the
		  * socket as non-blocking.
		  *
		  * \param[in] fd The socket file descriptor to assign.
		  * \param[in] status The status flags that are added to
		  * the socket. If the ssMonitored flag is among them,
		  * the watch() function is automatically called.
		  * \returns true if successful, false otherwise. errno
		  * is set accordingly in the latter case. */
		virtual bool assign(int fd, uint32_t status);
		
		/// Binds the socket.
		/** Binds the socket to the local address and port that have
		  * been specified through the set_local() function.
		  *
		  * This function creates the socket file descriptor if it
		  * is not created already. The ssCreated and ssBound flags
		  * are guaranteed to be set if this function succeeds. 
		  *
		  * \return true if successful, false otherwise. In the latter
		  * case, errno is set appropriately, unless you didn't call
		  * set_local(). */
		virtual bool bind();
		
		/// Sets local address and port.
		/** Sets local address and port. It can't be changed once the
		  * ssBound, ssListening, ssConnecting or ssConnected status
		  * flags are set on the socket.
		  *
		  * This function determines the socket protocol family if
		  * the ssRemoteInfo and ssCreated status flags are not set on
		  * the socket.
		  *
		  * \param[in] address The address to bind to, either as a
		  * hostname that will be resolved or an IP address expressed
		  * as a character string.
		  * \param[in] port The port to bind to, either as a service
		  * listed in /etc/services or a port number expressed as
		  * a character string.
		  * \param[out] error (optional) Points to a character wtring
		  * describing the error that happened if the function fails.
		  * \return true if successful, false otherwise. */
		virtual bool set_local(char const *address, char const *port,
				char const **error = 0);
		/// Same as the other set_local(), but port is an integer.
		virtual bool set_local(char const *address, unsigned short port,
				char const **error = 0);
		
		/// Sets remote address and port.
		/** Sets remote address and port. It can't be changed once the
		  * ssListening, ssConnecting or ssConnected status flags are
		  * set on the socket.
		  *
		  * This function determines the socket protocol family if
		  * the ssLocalInfo and ssCreated status flags are not set on
		  * the socket.
		  *
		  * \param[in] address The address to connect to, either as a
		  * hostname that will be resolved or an IP address expressed
		  * as a character string.
		  * \param[in] port The port to connect to, either as a service
		  * listed in /etc/services or a port number expressed as
		  * a character string.
		  * \param[out] error (optional) Points to a character wtring
		  * describing the error that happened if the function fails.
		  * \return true if successful, false otherwise. */
		virtual bool set_remote(char const *address, char const *port,
				char const **error = 0);
		/// Same as the other set_remote(), but port is an integer.
		virtual bool set_remote(char const *address, unsigned short port,
				char const **error = 0);
		
		/// Returns the local address.
		/** Returns the local address. This is either the address set
		  * through the set_local() function if it was used, or the
		  * one the socket is actually bound to if the ssCreated
		  * status flag is set and ssLocalInfo is not.
		  *
		  * \return The local address, or 0 if it is unavailable. */
		virtual char const *laddress() const;
		
		/// Returns the local port.
		/** Returns the local port. This is the port set through
		  * the set_local() function if it was used, or the
		  * one the socket is actually bound to if the ssCreated
		  * status flag is set and ssLocalInfo is not.
		  *
		  * \return The local port, or 0 if it is unavailable. */
		virtual unsigned short lport() const;
		
		/// Returns the remote address.
		/** Returns the remote address. This is the address set
		  * through the set_remote() function.
		  *
		  * \return The remote address, or 0 if it is unavailable. */
		virtual char const *raddress() const;
		
		/// Returns the remote port.
		/** Returns the remote port. This is the port set through the
		  * set_remote() function.
		  *
		  * \return The remote port, or 0 if it is unavailable. */
		virtual unsigned short rport() const;
	protected:
		/// IP information class
		/** This class stores information about the local and remote
		  * ends of a connection. */
		class IPInfo
		{
			public:
				/// Constructor.
				IPInfo(int family, size_t addrlen, struct sockaddr const *addr,
						char const *name, IPInfo *next);
				/// Copy constructor.
				IPInfo(IPInfo const &ipi);
				/// Destructor.
				~IPInfo();
				
				/// Returns the address protocol family
				int family() const { return m_family; }
				
				/// Returns the address size
				size_t addrlen() const { return m_addrlen; }
				/// Returns the address
				struct sockaddr const *addr() const 
						{ return (struct sockaddr *)&m_addr; }
				
				/// Returns the canonical name
				char const *cname() const { return m_name.c_str(); }
				/// Returns the canonical name (std::string version)
				std::string const &name() const { return m_name; }
					
				/// Next IP information (non-NULL if multiple answers available)
				IPInfo const *next() const { return m_next; }
			private:
				int m_family;
				
				size_t m_addrlen;
#if HAVE_IPV6
				struct sockaddr_storage m_addr;
#else
				struct sockaddr_in m_addr;
#endif
				
				std::string m_name;
				
				IPInfo *m_next;
				
				IPInfo &operator=(IPInfo const &);
		};
		
		/// Overrides Socket::create() to set the socket as non-blocking.
		virtual bool create();
		
		/// Resolves an address and port.
		/** Resolves an address and port.
		  *
		  * \param[in] address The address to resolve. It may either be
		  * an IP address or a hostname.
		  * \param[in] port The port to resolve. It may either be a
		  * numeric port or a service name.
		  * \param[out] error A pointer to a character string receiving
		  * an error message when the function returns false. May be NULL.
		  * \param[in] ipi On input, points to either a NULL pointer or
		  * a pointer to an IPInfo object that will be the next object in
		  * the resolved list. On output, points to either the original
		  * pointer or to the first new object of the resolved list.
		  * \return true if successful, false otherwise. */
		bool resolve(char const *address, char const *port,
				char const **error, IPInfo **ipi);
		
		/// Returns the local information
		IPInfo *local() { return m_local; }
		/// Returns the local information
		IPInfo const *local() const { return m_local; }
		/// Returns the remote information
		IPInfo *remote() { return m_remote; }
		/// Returns the remote information
		IPInfo const *remote() const { return m_remote; }
		
		/// Returns the port out of an IPInfo object.
		unsigned short port(IPInfo const *ipi) const;
	private:
		IPInfo *m_local;
		IPInfo *m_remote;
};

/*************************************************************************/

} /* namespace sockets */

/*************************************************************************/

#endif /* SOCKETS_IPSOCKET_H */
