/* BotServ header file.
 *
 * Epona (c) 2000-2002, 2004 PegSoft
 * Contact us at epona@pegsoft.net
 *
 * This program is free but copyrighted software; see the file COPYING for
 * details.
 *
 * Based on the original code of Services by Andy Church.
 *
 */

#ifndef BOTSERV_H
#define BOTSERV_H

/*************************************************************************/

#include "tweak.h"

/*************************************************************************/

#include <databases/collection.h>
#include <databases/record.h>
#include <log/levels.h>

#include "services.h"

/*************************************************************************/

namespace modules { class Module; }

class ChannelInfo;

/*************************************************************************/

/* Bot info structures.
 */
 
struct BotInfo : public databases::NameRecord
{
	public:
		/* Data */
		
		static databases::Field *f_nick;
		static databases::Field *f_username;
		static databases::Field *f_hostname;
		static databases::Field *f_realname;
		static databases::Field *f_created;
		static databases::Field *f_flags;
		
		/* Constants */
		
		/// Maximum nickname length.
		static size_t const nick_len = 63;
		/// Maximum username length.
		static size_t const username_len = 31;
		/// Maximum hostname length.
		static size_t const hostname_len = 63;
		/// Maximum realname length.
		static size_t const realname_len = 127;
		
		/* Functions */
		
		/// Constructor.
		BotInfo();
		/// Constructor setting the record identifier.
		explicit BotInfo(identifier_type id);
		
		/// Overrides NameRecord::get_s().
		virtual char const *get_s(databases::Field const *field) const;
		/// Overrides NameRecord::set_s().
		virtual char const *set_s(databases::Field const *field,
				char const *value);
		/// Overrides NameRecord::get_ui().
		virtual uint32_t get_ui(databases::Field const *field) const;
		/// Overrides NameRecord::set_ui().
		virtual uint32_t set_ui(databases::Field const *field, uint32_t value);
		/// Overrides NameRecord::get_ut().
		virtual time_t get_ut(databases::Field const *field) const;
		/// Overrides NameRecord::set_ut().
		virtual time_t set_ut(databases::Field const *field, time_t value);
		
		/// Overrides Record::end_update().
		virtual bool end_update();
		/// Updates the pseudo-client (after changes).
		void update_client();
		
		/// Overrides NameRecord::name().
		virtual char const *name() const { return m_nick; }
		
		/// Overrides NameRecord::memory_usage().
		virtual size_t memory_usage() const;
		
		/// Returns the bot's nickname
		char const *nick() const { return m_nick; }
		/// Sets the bot's nickname
		char const *nick(char const *nick);
		
		/// Returns the bot's username
		char const *username() const { return m_username; }
		/// Sets the bot's username
		char const *username(char const *username);
		
		/// Returns the bot's hostname
		char const *hostname() const { return m_hostname; }
		/// Sets the bot's hostname
		char const *hostname(char const *hostname);
		
		/// Returns the bot's realname
		char const *realname() const { return m_realname; }
		/// Sets the bot's realname
		char const *realname(char const *realname);
		
		/// Returns the bot's creation date
		time_t created() const { return m_created; }
		/// Sets the bot's creation date
		time_t created(time_t created);
		
		/* Flags */
		
		/// Bot can't be assigned by non-IRC operators
		static uint16_t const bfPrivate = 0x0001;
		/// Unused bot flags
		static uint16_t const bfUnused  = 0xFFFE;
		
		/// Returns the bot flags (see the bf* constants).
		uint16_t flags() const { return m_flags; }
		/// Sets the bot flags -- for use by database modules only!
		uint16_t flags(uint16_t flags);
		/// Adds flags to the current bot flags
		uint16_t flags_add(uint16_t flags);
		/// Removes flags from the current bot flags
		uint16_t flags_remove(uint16_t flags);
		
		/// Returns the number of channels used by the bot.
		uint16_t chancount() const { return m_chancount; }
		/// Sets the number of channels used by the bot.
		uint16_t chancount(uint16_t chancount)
				{ return (m_chancount = chancount); }
		
		/// Returns the last time the bot said something.
		time_t lastmsg() const { return m_lastmsg; }
		/// Sets the last time the bot said something.
		time_t lastmsg(time_t lastmsg) { return (m_lastmsg = lastmsg); }
	 private:
	 	enum
		{
			/// Username, hostname or realname have been updated.
			dbsUHRUpdated = 0x8000
		};

		misc::cstring m_nick;
		misc::cstring m_username;
		misc::cstring m_hostname;
		misc::cstring m_realname;
		
		time_t m_created;
		
		uint16_t m_flags;
		
		/* Volatile data */
		uint16_t m_chancount;
		time_t m_lastmsg;		/* Last time we said something */
};

/*************************************************************************/

class Bots : public databases::NameCollection
{
	public:
		/// Constructor.
		Bots();
		
		/// Implements Collection::create().
		virtual databases::Record *
				create(databases::Record::identifier_type id = 0) const;
		/// Overrides NameCollection::add().
		virtual bool add(databases::Record *record, int *err = 0);
		/// Overrides NameCollection::remove().
		virtual bool remove(databases::Record *record, int *err = 0);
		/// Overrides NameCollection::clear().
		virtual bool clear(int *err = 0) { return false; }
		
		/// Finds a record by ID.
		virtual BotInfo *ifind(databases::Record::identifier_type id)
				{ return (BotInfo *)NameCollection::find(id); }
		/// Finds a record by name.
		virtual BotInfo *find(char const *name) const
				{ return (BotInfo *)NameCollection::nfind(name); }
		
		/// Removes matching nick record and kills any user using the nick (returns true if successful).
		bool nick_cleanup(char const *nick) const;
		
		/// Sets the database up.
		/** Sets the database up. Called by the Module::notify()
		  * function when the database module is loaded. */
		void setup_database(databases::Database *db);
	private:
		Bots(Bots const &);
		Bots &operator=(Bots const &);
};

/*************************************************************************/

extern Bots bots;

/*************************************************************************/

class KBB : public databases::Record
{
	public:
		/* Data */
		static databases::Field *f_owner;
		static databases::Field *f_name;
		static databases::Field *f_count;
		
		/* Constants */
		
		/// Maximum length of the name.
		static size_t const name_len = 31;
		
		/* Functions */
		
		/// Constructor.
		KBB();
		/// Constructor setting the record ID.
		KBB(identifier_type id);
		
		/// Overrides Record::get_s().
		virtual char const *get_s(databases::Field const *field) const;
		/// Overrides Record::set_s().
		virtual char const *set_s(databases::Field const *field,
				char const *value);
		/// Overrides Record::get_ui().
		virtual uint32_t get_ui(databases::Field const *field) const;
		/// Overrides Record::set_ui().
		virtual uint32_t set_ui(databases::Field const *field, uint32_t value);
		/// Overrides Record::get_ri().
		virtual identifier_type get_ri(databases::Field const *field) const;
		/// Overrides Record::set_ri().
		virtual identifier_type set_ri(databases::Field const *field,
				identifier_type value);
		
		/// Overrides Record::end_update().
		virtual bool end_update();
		
		/// Overrides Record::memory_usage().
		virtual size_t memory_usage() const;
		
		/// Returns the owner channel of the KBB.
		ChannelInfo *owner() const { return m_owner; }
		/// Sets the owner channel of the KBB.
		ChannelInfo *owner(ChannelInfo *owner);
		/// Sets the owner channel ID of the KBB (for databases only!)
		identifier_type owner_id(identifier_type owner_id);
		
		/// Returns the KBB index.
		uint8_t index() const { return m_index; }
		/// Sets the KBB index.
		uint8_t index(uint8_t index);
		
		/// Returns the number of kicks before ban.
		uint8_t count() const { return m_count; }
		/// Sets the the number of kicks before ban.
		uint8_t count(uint8_t count);
	private:
		ChannelInfo *m_owner;
		
		uint8_t m_index;
		uint8_t m_count;
		
		char const *name(char const *name);
};

/*************************************************************************/

class KBBs : public databases::Collection
{
	public:
		/* Functions */
		
		/// Implements Collection::create().
		virtual databases::Record *
				create(databases::Record::identifier_type id = 0) const;
		/// Overrides Collection::add().
		virtual bool add(databases::Record *record, int *err = 0);
		
		/// Overrides Collection::clear().
		virtual bool clear(int *err = 0) { return false; }
		
		/// Configures the map.
		void setup_map();
		
		/// Configures owner collection.
		void setup_owner();
		
		/// Sets the database up.
		/** Sets the database up. This is called by the Module::notify()
		  * function when the database module is loaded. */
		void setup_database(databases::Database *db);
		
		/// Overrides Collection::owner_field().
		virtual databases::Field const
				*owner_field(Collection const *coll) const
				{ return KBB::f_owner; }
	protected:
		/// Overrides Collection::add_list().
		virtual bool add_list(databases::Record *record);
		/// Overrides Collection::remove_list().
		virtual void remove_list(databases::Record *record);
		
		/// Overrides Collection::remove_owned_db().
		virtual bool remove_owned_db(databases::Collection const *coll,
				databases::Record const *record, int *err = 0);
		/// Overrides Collection::remove_owned_list().
		virtual bool remove_owned_list(databases::Collection const *coll,
				databases::Record const *record, int *err = 0);
};

#define KBB_SIZE  8

/*************************************************************************/

extern KBBs kbbs;

/*************************************************************************/

/* Structure used to contain bad words. */

class BadWord : public databases::NameRecord
{
	public:
		/* Data */
		static databases::Field *f_owner;
		static databases::Field *f_word;
		static databases::Field *f_type;
		
		/* Constants */
		
		/// Maximum length of the word.
		static size_t const word_len = 63;
		
		// Word types
		
		/// Matches anywhere
		static uint8_t const bwAny     = 0;
		/// Matches entire word
		static uint8_t const bwSingle  = 1;
		/// Matches at beginning of a word
		static uint8_t const bwStart   = 2;
		/// Matches at end of a word
		static uint8_t const bwEnd     = 3;
		
		/// Maximum value for the word type
		static uint8_t const bwMax     = 3;
		
		/* Functions */
		
		/// Constructor.
		BadWord();
		/// Constructor setting the record ID.
		BadWord(identifier_type id);
		
		/// Overrides Record::get_s().
		virtual char const *get_s(databases::Field const *field) const;
		/// Overrides Record::set_s().
		virtual char const *set_s(databases::Field const *field,
				char const *value);
		/// Overrides Record::get_ui().
		virtual uint32_t get_ui(databases::Field const *field) const;
		/// Overrides Record::set_ui().
		virtual uint32_t set_ui(databases::Field const *field, uint32_t value);
		/// Overrides Record::get_ri().
		virtual identifier_type get_ri(databases::Field const *field) const;
		/// Overrides Record::set_ri().
		virtual identifier_type set_ri(databases::Field const *field,
				identifier_type value);
		
		/// Implements NameRecord::name().
		virtual char const *name() const { return word(); }
		
		/// Overrides Record::memory_usage().
		virtual size_t memory_usage() const;
		
		/// Returns the owner channel of the bad word.
		ChannelInfo *owner() const { return m_owner; }
		/// Sets the owner channel of the bad word.
		ChannelInfo *owner(ChannelInfo *owner);
		/// Sets the owner channel ID of the bad word (for databases only!)
		identifier_type owner_id(identifier_type owner_id);
		
		/// Returns the bad word.
		char const *word() const { return m_word; }
		/// Sets the bad word.
		char const *word(char const *word);
		
		/// Returns the bad word type (see BW_* constants)
		uint8_t type() const { return m_type; }
		/// Sets the bad word type (see BW_* constants)
		uint8_t type(uint8_t type);
	private:
		ChannelInfo *m_owner;
		
		misc::cstring m_word;
		uint8_t m_type;
};

/*************************************************************************/

class BadWords : public databases::Collection
{
	public:
		/* Consts */
		
		/// Custom error codes
		enum
		{
			/// Bad word entry with the same name already exists.
			ceName = ceCustom + 0
		};
		
		/* Functions */
		
		/// Implements Collection::create().
		virtual databases::Record *
				create(databases::Record::identifier_type id = 0) const;
		/// Overrides Collection::add().
		virtual bool add(databases::Record *record, int *err = 0);
		
		/// Overrides Collection::clear().
		virtual bool clear(int *err = 0) { return false; }
		
		/// Configures owner collection.
		void setup_owner();
		
		/// Sets the database up.
		/** Sets the database up. This is called by the Module::notify()
		  * function when the database module is loaded. */
		void setup_database(databases::Database *db);
		
		/// Overrides Collection::owner_field().
		virtual databases::Field const
				*owner_field(Collection const *coll) const
				{ return BadWord::f_owner; }
	protected:
		/// Overrides Collection::add_list().
		virtual bool add_list(databases::Record *record);
		/// Overrides Collection::remove_list().
		virtual void remove_list(databases::Record *record);
		
		/// Overrides Collection::remove_owned_db().
		virtual bool remove_owned_db(databases::Collection const *coll,
				databases::Record const *record, int *err = 0);
		/// Overrides Collection::remove_owned_list().
		virtual bool remove_owned_list(databases::Collection const *coll,
				databases::Record const *record, int *err = 0);
};

/*************************************************************************/

extern BadWords badwords;

/*************************************************************************/

LOGLEVELH(ErrorBotServ, xlog::LogError, "Error::BotServ", "Error: botserv: ");
LOGLEVELH(ErrorBotServList, ErrorBotServ, "Error::BotServ::List", "Error: botserv list: ");

/*************************************************************************/

extern void bs_init();

extern bool botserv_load();
extern void botserv_notify(modules::Module &sender, int msg, void *arg);
extern void botserv_start();

extern void botserv(User *u, char *buf);
extern void botmsgs(User *u, BotInfo *bi, char *buf);
extern void botchanmsgs(User *u, ChannelInfo *ci, char *buf);

extern void bot_join(ChannelInfo *ci);
extern void bot_rejoin_all(BotInfo *bi);

/*************************************************************************/

#endif /* BOTSERV_H */
