/* Declarations for command data.
 *
 * Epona (c) 2000-2002, 2004 PegSoft
 * Contact us at epona@pegsoft.net
 *
 * This program is free but copyrighted software; see the file COPYING for
 * details.
 *
 * Based on the original code of Services by Andy Church.
 *
 */

#ifndef COMMANDS_H
#define COMMANDS_H

/*************************************************************************/

#include <containers/vectorsp.h>
#include <databases/record.h>
#include <misc/string.h>

#include "services.h"

/*************************************************************************/

/// Services command class

/** This class implements a service command.
  *
  * The execute function() must be overriden in derived classes. */

class Cmd
{
	public:
		/// Execution results
		enum
		{
			/// Don't execute the next command with the same name
			crStop = 0,
			/// Execute the next command with the same name
			crContinue = 1
		};
		
		/// Constructor.
		/** Constructor.
		  *
		  * \param[in] The name of the command.
		  * \param[in] The name of the service to which this
		  * command belongs. */
		Cmd(char const *name, char const *service) : m_name(name),
				m_service(service) { }
		/// Destructor
		virtual ~Cmd() { }
	
		/// Executes the command.
		/** Executes the command. What this function actually does
		  * is up to the implementer.
		  *
		  * You may obtain command parameters by using the
		  * strtok(NULL, delimiter) function.
		  *
		  * \param[in] u The user who executes the command.
		  * \return crStop or crContinue, depending on what the
		  * command handler should do next.
		  * \sa crStop, crContinue */
		virtual int execute(User *u) = 0;
	protected:
		/// The name of the command.
		misc::cstring m_name;
		/// The service to which this command belongs.
		misc::cstring m_service;
	private:
		Cmd(Cmd const &);
		Cmd &operator=(Cmd const &);
};

/*************************************************************************/

/// List command class

/** This class implements a command specialized in manipulating 
  * containers::VectorSP-based lists. 
  *
  * It provides an implementation for the sub-commands DEL, LIST and CLEAR
  * while letting derived classes override their default behaviour and
  * add new sub-commands.
  *
  * The container(), changeable(), del_record() and list_record() functions
  * must be overriden in derived classes. The execute(), context(),
  * allow_dispatch(), dispatch(), is_member(), add(), del(), allow_del(),
  * list(), allow_list(), clear(), allow_clear() and clear_records() functions
  * may be overriden too.
  */

class ListCmd : public Cmd
{
	public:
		/* Typedefs */
		/// The container type.
		typedef containers::VectorSP<databases::Record *, uint16_t> container_type;
		
		/* Constants */
	
		/// Indexes of the language file messages in the m_msgs array.
		enum
		{
			/// Syntax error message
			cmSyntax = 0,
			/// Modification disabled message
			cmDisabled = 1,
			/// Empty list message
			cmEmpty = 2,
			/// DEL syntax
			cmSyntaxDel = 3,
			/// No such entry message
			cmNoSuchEntry = 4,
			/// No matching entries message
			cmNoMatch = 5,
			/// Deleted one entry message
			cmDeletedOne = 6,
			/// Deleted several entries message
			cmDeletedSeveral = 7,
			/// Name not found in the list
			cmNotFound = 8,
			/// Deleted an entry (by name)
			cmDeleted = 9,
			/// LIST format
			cmSyntaxList = 10,
			/// List header
			cmListHeader = 11,
			/// Listed one entry
			cmListedOne = 12,
			/// Listed several entries
			cmListedSeveral = 13,
			/// List cleared partially
			cmClearPartially = 14,
			/// List cleared
			cmClear = 15,
			
			/// Total messages for ListCmd
			cmListCmdTotal = 16
		};
	
		/* Functions */
	
		/// Constructor.
		/** Constructor.
		  *
		  * \param[in] msgs The language file messages used by the
		  * commands. See the cm* constants to know the order of
		  * the messages in the array.
		  * \param[in] named true if the records are upcasted from
		  * the databases::NameRecord class, false otherwise. */
		ListCmd(char const *name, char const *service, int const *msgs, bool named) :
				Cmd(name, service), m_msgs(msgs), m_named(named) { }
		
		/// Implements Cmd::execute().
		/** Implements Cmd::execute().
		  *
		  * This method will first call the context() method to get
		  * the container_type object on which the list command
		  * operates. If it obtains one, it will get the sub-command
		  * name and call allow_dispatch() the user has access to
		  * this sub-command, then the dispatch() method so that
		  * appropriate action can be taken.
		  *
		  * \param[in] u The user who executes the command.
		  * \return The return value of dispatch(), or csStop if
		  * it wasn't called. */
		virtual int execute(User *u);
	protected:
		/* Data */
		/// Array of language file messages
		int const *m_msgs;
		/// Are the records upcasted from databases::NameRecord ?
		bool m_named;
		
		/* Functions */
		
		/// Returns a context string for the list.
		/** Returns a context string for the list, that'll be used as a
		  * parameter for language strings. This function may be called
		  * multiple times by all other functions.
		  *
		  * The default implementation always returns "".
		  *
		  * \param[in] u The user who executes the command.
		  * \return A pointer to a context string, or "" if no context
		  * string is available. */
		virtual char const *context(User *u) const { return ""; }
		
		/// Returns the container on which the list command operates.
		/** Returns the container on which the list command operates.
		  * How it is determined is up to the implementer.
		  *
		  * \param[in] u The user who executes the command.
		  * \return A pointer to the container, or 0 on error (an
		  * appropriate error message should have been sent to the user
		  * in that case). */
		virtual container_type *container(User *u) = 0;
		
		/// Determines whether the user has access to a sub-command.
		/** Determines whether the user has access to the given sub-command.
		  *
		  * The default implementation always returns true. 
		  *
		  * \param[in] u The user who executes the command.
		  * \param[in] ctn The container on which this list operates. 
		  * \param[in] sub The sub-command name.
		  * \return true if the user has access to the sub-command,
		  * false otherwise. */
		virtual bool allow_dispatch(User *u, container_type &ctn,
				char const *sub) { return true; }
		
		/// Dispatches the sub-command.
		/** Dispatches the sub-command.
		  *
		  * The default implementation calls the add() method if the
		  * sub-command is "ADD", del() if it is "DEL", list() if it is
		  * "LIST", clear() if it is "CLEAR", or sends an error message
		  * to the user if it is something else.
		  *
		  * \param[in] u The user who executes the command.
		  * \param[in] ctn The container on which this list operates. 
		  * \param[in] sub The sub-command name.
		  * \return The return value of the method it calls, or crStop
		  * if the sub-command is invalid. */
		virtual int dispatch(User *u, container_type &ctn, char const *sub);
		
		/// Returns whether the list can be changed.
		/** Returns whether the list can be changed.
		  *
		  * \param[in] u The user who executes the command.
		  * \param[in] ctn The container on which this list operates.
		  * \param[in] sub The sub-command name.
		  * \return true if the list can be changed, false otherwise. */
		virtual bool changeable(User *u, container_type &ctn,
				char const *sub) const = 0;
		
		/// Returns whether the given record is a member of this list.
		/** Returns whether the given record is a member of this list.
		  * That way, you may have commands managing containers with
		  * records of different types.
		  *
		  * The default implementation always return true.
		  *
		  * \param[in] u The user who executes the command.
		  * \param[in] ctn The container on which this list operates. 
		  * \param[in] sub The sub-command name.
		  * \param[in] rec The record. */
		virtual bool is_member(User *u, container_type &ctn,
				char const *sub, databases::Record *rec) const { return true; }
		
		/// Processes the ADD sub-command.
		/** Processes the ADD sub-command. The default implementation
		  * does nothing.
		  *
		  * \param[in] u The user who executes the command.
		  * \param[in] ctn The container on which this list operates. 
		  * \param[in] sub The sub-command name.
		  * \return Returns the same potiential return values than
		  * Cmd::execute(). The default implementation returns crStop. */
		virtual int add(User *u, container_type &ctn, char const *sub)
				{ return crStop; }
		
		/// Processes the DEL sub-command.
		/** Processes the DEL sub-command. The default implementation
		  * deletes records using a numeric range parameter or, if the
		  * list contains records upcasted from databases::NameRecord,
		  * using a name parameter. Before deleting a record, the
		  * is_member(), allow_del() and del_record() methods are
		  * called. If they are all successful, the pointer to the
		  * record in the container is set to NULL.
		  *
		  * \param[in] u The user who executes the command.
		  * \param[in] ctn The container on which this list operates. 
		  * \param[in] sub The sub-command name.
		  * \return Returns the same potiential return values than
		  * Cmd::execute(). The default implementation returns crStop. */
		virtual int del(User *u, container_type &ctn, char const *sub);
		
		/// Determines whether the user can delete a record.
		/** Determines whether the user can delete the given record.
		  *
		  * The default implementation always returns true. 
		  *
		  * \param[in] u The user who executes the command.
		  * \param[in] ctn The container on which this list operates. 
		  * \param[in] sub The sub-command name.
		  * \param[in] rec The record.
		  * \return true if the user can delete the record, false
		  * otherwise. */
		virtual bool allow_del(User *u, container_type &ctn,
				char const *sub, databases::Record const *rec) { return true; }
		
		/// Deletes a record.
		/** Deletes the given record. How this is achieved is up to
		  * the implementer.
		  *
		  * \param[in] u The user who executes the command.
		  * \param[in] ctn The container on which this list operates. 
		  * \param[in] sub The sub-command name.
		  * \param[in] rec The record.
		  * \return true if the record has been deleted successfully,
		  * false otherwise. */
		virtual bool del_record(User *u, container_type &ctn,
				char const *sub, databases::Record *rec) = 0;
		
		/// Processes the LIST sub-command.
		/** Processes the LIST sub-command. The default implementation
		  * lists records using a numeric range parameter or, if the
		  * list contains records upcasted from databases::NameRecord,
		  * by matching an optional parameter to the name of the record.
		  * The is_member() and allow_list() methods are then called for
		  * each record and, if they both return true, list_record() is
		  * called to list the record.
		  *
		  * \param[in] u The user who executes the command.
		  * \param[in] ctn The container on which this list operates. 
		  * \param[in] sub The sub-command name.
		  * \return Returns the same potiential return values than
		  * Cmd::execute(). The default implementation returns crStop. */
		virtual int list(User *u, container_type &ctn, char const *sub);
		
		/// Determines whether the user can list a record.
		/** Determines whether the user can list the given record.
		  *
		  * The default implementation always returns true. 
		  *
		  * \param[in] u The user who executes the command.
		  * \param[in] ctn The container on which this list operates. 
		  * \param[in] sub The sub-command name.
		  * \param[in] rec The record.
		  * \return true if the user can list the record, false
		  * otherwise. */
		virtual bool allow_list(User *u, container_type &ctn,
				char const *sub, databases::Record const *rec) { return true; }
		
		/// Lists a record.
		/** Lists the given record.
		  *
		  * \param[in] u The user who executes the command.
		  * \param[in] ctn The container on which this list operates. 
		  * \param[in] sub The sub-command name.
		  * \param[in] num The record number (which is its index in
		  * container + 1).
		  * \param[in] rec The record.
		  * \param[in] count How many records were listed before.
		  * \return true if the record has been listed and should be
		  * included in the count, false otherwise. */
		virtual bool list_record(User *u, container_type &ctn,
				char const *sub, uint16_t num, databases::Record const *rec,
				uint16_t count) = 0;
		
		/// Processes the CLEAR sub-command.
		/** Processes the CLEAR sub-command. The default implementation
		  * first calls allow_clear(), then calls is_member() and
		  * del_record() for each record in the list and clears the container.
		  *
		  * \param[in] u The user who executes the command.
		  * \param[in] ctn The container on which this list operates. 
		  * \param[in] sub The sub-command name.
		  * \return Returns the same potiential return values than
		  * Cmd::execute(). The default implementation returns crStop. */
		virtual int clear(User *u, container_type &ctn, char const *sub);
		
		/// Determines whether the user can clear the list.
		/** Determines whether the user can clear the list.
		  *
		  * The default implementation always returns true. 
		  *
		  * \param[in] u The user who executes the command.
		  * \param[in] ctn The container on which this list operates. 
		  * \param[in] sub The sub-command name.
		  * \return true if the user can clear the list, false
		  * otherwise. */
		virtual bool allow_clear(User *u, container_type &ctn,
				char const *sub) { return true; }
		
		/// Clears all records in the list.
		/** Clears all records in the list.
		  *
		  * The default implementation calls del_record() for each
		  * record that is determined to be a member of the list
		  * by the is_member() method. ctn.clear() is then called
		  * if there are no records in the container anymore.
		  * Feel free to override the function and optimize it
		  * if you can.
		  *
		  * \param[in] u The user who executes the command.
		  * \param[in] ctn The container on which this list operates. 
		  * \param[in] sub The sub-command name.
		  * \return true if there are still member records in the
		  * list, false otherwise. */
		virtual bool clear_records(User *u, container_type &ctn,
				char const *sub);
};

/*************************************************************************/

/* Structure for information about a *Serv command. */

struct Command 
{
    const char *name;
    void (*routine)(User *u);
    int (*has_priv)(User *u);	/* Returns 1 if user may use command, else 0 */

    /* Regrettably, these are hard-coded to correspond to current privilege
     * levels (v4.0).  Suggestions for better ways to do this are
     * appreciated.
     */
    int helpmsg_all;	/* Displayed to all users; -1 = no message */
    int helpmsg_reg;	/* Displayed to regular users only */
    int helpmsg_oper;	/* Displayed to Services operators only */
    int helpmsg_admin;	/* Displayed to Services admins only */
    int helpmsg_root;	/* Displayed to Services root only */
    const char *help_param1;
    const char *help_param2;
    const char *help_param3;
    const char *help_param4;
};

/*************************************************************************/

/* Routines for looking up commands.  Command lists are arrays that must be
 * terminated with a NULL name.
 */

extern Command *lookup_cmd(Command *list, const char *name);
extern void run_cmd(const char *service, User *u, Command *list,
		const char *name);
extern void help_cmd(const char *service, User *u, Command *list,
		const char *name);

/*************************************************************************/

#endif /* COMMANDS_H */
