/* Epona core directives.
 *
 * Epona (c) 2000-2002, 2004 PegSoft
 * Contact us at epona@pegsoft.net
 *
 * Based on the original code of Services by Andy Church.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** \file core/directives.h
  * \brief Header file for Epona core configuration directives.
  *
  * This header file provides access to the configuration directives
  * used by the core Epona module. Some of them may be renamed and/or
  * removed as the development of Epona progresses, so using them in
  * your modules is NOT recommended unless one module wouldn't have
  * any purpose anyways if the feature that the directive controls gets
  * removed.
  *
  * Directives aren't documented unless there's something special to them,
  * because the epona.conf.dist does a better job for that. */

#ifndef CORE_DIRECTIVES_H
#define CORE_DIRECTIVES_H

/*************************************************************************/

#include <conf/directives.h>
#include <conf/quick.h>
#include <log/conf.h>
#include <modules/conf.h>

/*************************************************************************/

namespace core
{
	
/*************************************************************************/

/// Namespace for Epona's core private declarations.

namespace hidden
{

/*************************************************************************/

/// Class encapsulating all core directives.

class Directives : public conf::DirectivesQ
{
	public:
		/* Classes */
		class AllowRoot : public conf::ConfBool
		{
			public:
				AllowRoot();
			protected:
				virtual bool apply(conf::Conf &conf);
		};
		
		class Directory : public conf::ConfStrChar
		{
			public:
				Directory();
			protected:
				virtual bool apply(conf::Conf &conf);
		};
		
		class DumpCore : public conf::ConfBool
		{
			public:
				DumpCore();
			protected:
				virtual bool apply_change(conf::Conf &conf);
		};
		
		class PIDFile : public conf::ConfStrChar
		{
			public:
				PIDFile();
			protected:
				virtual bool apply_change(conf::Conf &conf);
		};
		
		class Umask : public conf::ConfUint<mode_t>
		{
			public:
				Umask();
			protected:
				virtual bool apply(conf::Conf &conf);
		};
		
		class BadPasswords : public conf::ConfBoolBlock
		{
			public:
				/* Data */
				conf::ConfUint<uint16_t> limit;
				conf::ConfDuration timeout;
				
				/* Functions */
				BadPasswords();
			protected:
				virtual bool add(conf::Conf &conf, conf::Conf &new_conf);
		};
		
		class Mail : public conf::ConfBoolBlock
		{
			public:
				/* Classes */
				
				class Transport : public conf::ConfString
				{
					public:
						/// Constructor.
						Transport();
					protected:
						/// Loads new module.
						virtual bool apply_change(conf::Conf &conf);
				};
				
				class Sender : public conf::ConfBlock
				{
					public:
						/* Classes */
						class Address : public conf::ConfString
						{
							public:
								/// Constructor.
								Address();
							protected:
								/// Ensures the e-mail address is valid.
								virtual bool check(conf::Conf &conf,
										std::string const &value) const;
						};
						
						
						/* Data */
						
						conf::ConfString name;
						Address address;
						
						/* Functions */
						
						/// Constructor.
						Sender();
					protected:
						/// Adds directives.
						virtual bool add(conf::Conf &conf,
								conf::Conf &new_conf);
				};
				
				class Limit : public conf::ConfBlock
				{
					public:
						/* Data */
						
						conf::ConfRatio authentified;
						conf::ConfRatio unauthentified;
						
						/* Functions */
						
						/// Constructor.
						Limit();
					protected:
						/// Adds directives.
						virtual bool add(conf::Conf &conf,
								conf::Conf &new_conf);
				};
				
				/* Data */
				
				Transport transport;
				Sender sender;
				Limit limit;
				conf::ConfRatio sender_limit;
				
				/* Functions */
				Mail();
			protected:
				virtual bool add(conf::Conf &conf, conf::Conf &new_conf);
				virtual bool apply_changes(conf::Conf &conf);
		};
		
		class Services : public conf::ConfBlock 
		{
			public:
				/* Data */
				conf::ConfStrChar sname;
				conf::ConfStrChar description;
				conf::ConfHostPort connect;
				conf::ConfHostPort bind;
				conf::ConfDuration read_timeout;
				conf::ConfStrChar password;
				conf::ConfStrChar network_name;
				conf::ConfBool log_users;
				conf::ConfBool log_maxusers;
				conf::ConfBool log_input;
				conf::ConfBool log_output;
				conf::ConfStrChar motd;
				
				conf::ConfStrChar guest_nick;
				conf::ConfDuration inhabit;
				
				conf::ConfStrChar help_channel;
		
				/* Functions */
				Services();
			protected:
				virtual bool add(conf::Conf &conf, conf::Conf &new_conf);
		};
		
		class Session : public conf::ConfBlock
		{
			public:
				/* Data */
				conf::ConfBool enable;
				conf::ConfUint<uint16_t> default_limit;
				conf::ConfStrCharA message;
				conf::ConfUint<uint16_t> max_kills;
				conf::ConfDuration expire_akills;
				conf::ConfString database_exceptions;
				conf::ConfUint<uint16_t> max_limit;
				conf::ConfDuration expire_exceptions;

				/* Functions */
				Session();
			protected:
				virtual bool add(conf::Conf &conf, conf::Conf &new_conf);
		};
		
		class Blacklist : public conf::ConfBlock
		{
			public:
				/* Data */
				conf::ConfBool enable;
				conf::ConfStrCharA master;
				conf::ConfStrCharA diff;
				conf::ConfDuration update;
				conf::ConfStrChar ban_reason;
				conf::ConfUint<uint16_t> max;
				
				/* Functions */
				Blacklist();
			protected:
				virtual bool add(conf::Conf &conf, conf::Conf &new_conf);
		};
		
		class Service : public conf::ConfBlock
		{
			public:
				/* Data */
				conf::ConfStrChar nick;
				conf::ConfStrChar username;
				conf::ConfStrChar hostname;
				conf::ConfStrChar realname;
				
				/* Functions */
				Service(std::string const &name, std::string const &defnick,
						std::string const &defreal);
			protected:
				virtual bool add(conf::Conf &conf, conf::Conf &new_conf);
		};
		
		class NickServ : public conf::ConfBlock
		{
			public:
				/* Classes */
				class PasswordModule : public conf::ConfStrChar
				{
					public:
						PasswordModule();
					protected:
						virtual bool apply_change(conf::Conf &conf);
				};
				
				class DefaultOptions : public conf::ConfSet<uint32_t>
				{
					public:
						DefaultOptions();
					protected:
						virtual bool add(conf::Conf &conf);
						virtual bool check(conf::Conf &conf,
								uint32_t const &value);
				};
				
				class DefaultLanguage : public conf::ConfInt<int16_t>
				{
					public:
						DefaultLanguage();
					protected:
						virtual bool apply_change(conf::Conf &conf);
				};
				
				class Auth : public conf::ConfBoolBlock
				{
					public:
						/* Classes */
						class Restrict : public conf::ConfSet<uint32_t>
						{
							public:
								/* Classes */
								enum
								{
									rfGroup           = 0x00000001,
									rfIdentify        = 0x00000002,
									rfIdentifyImplied = 0x00000003,
									rfAll             = 0x00000003
								};
								
								/* Functions */
								Restrict();
								
								bool group() const
										{ return (m_value & rfGroup); }
								bool identify() const
										{ return (m_value & rfIdentify); }
							protected:
								virtual bool add(conf::Conf &conf);
						};
						
						/* Data */
						conf::ConfBool required;
						conf::ConfStrChar web_form;
						conf::ConfUint<uint8_t> attempts;
						Restrict restrict;
						
						/* Functions */
						Auth();
					protected:
						virtual bool add(conf::Conf &conf,
								conf::Conf &new_conf);
				};
				
				class Expire : public conf::ConfBoolBlock
				{
					public:
						/* Data */
						conf::ConfDuration delay;
						conf::ConfDuration grace;
					
						/* Functions */
						Expire();
					protected:
						virtual bool add(conf::Conf &conf,
								conf::Conf &new_conf);
				};
				
				class Enforcer : public conf::ConfBlock
				{
					public:
						/* Data */
						conf::ConfStrChar username;
						conf::ConfStrChar hostname;
						
						/* Functions */
						Enforcer();
					protected:
						virtual bool add(conf::Conf &conf,
								conf::Conf &new_conf);
				};
				
				class CustomHost : public conf::ConfBlock
				{
					public:
						/* Data */
						conf::ConfBool enable;
						conf::ConfBool restrict;
						conf::ConfDuration age;
						conf::ConfDuration delay;
						conf::ConfBool enforce_strict;
						
						/* Functions */
						CustomHost();
					protected:
						virtual bool add(conf::Conf &conf,
								conf::Conf &new_conf);
				};
			
				/* Data */
				Service service;
				
				PasswordModule password_module;
				
				conf::ConfString database_groups;
				conf::ConfString database_aliases;
				conf::ConfString database_access;
				conf::ConfString database_memos;
				
				conf::ConfBool require_email;
				DefaultOptions default_options;
				int16_t mutable default_language;
				/// For internal use only -- use default_language instead
				DefaultLanguage real_deflang;
				conf::ConfDuration reg_delay;
				Auth auth;
				Expire expire;
				conf::ConfUint<uint16_t> max_aliases;
				conf::ConfUint<uint16_t> max_access;
				Enforcer enforcer;
				conf::ConfDuration release_timeout;
				conf::ConfBool kill_immed;
				conf::ConfBool group_change;
				conf::ConfBool list_opers;
				conf::ConfUint<uint16_t> list_max;
				conf::ConfBool sendpass_opers;
				conf::ConfBool protect_admins;
				conf::ConfBool strict_privileges;
				CustomHost custom_host;
				
				/* Functions */
				NickServ();
			protected:
				virtual bool add(conf::Conf &conf, conf::Conf &new_conf);
		};
		
		class ChanServ : public conf::ConfBlock
		{
			public:
				/* Classes */
				class PasswordModule : public conf::ConfStrChar
				{
					public:
						PasswordModule();
					protected:
						virtual bool apply_change(conf::Conf &conf);
				};
				
				class DefaultOptions : public conf::ConfSet<uint32_t>
				{
					public:
						DefaultOptions();
					protected:
						virtual bool add(conf::Conf &conf);
						virtual bool check(conf::Conf &conf,
								uint32_t const &value);
				};
				
				class RestrictAuth : public conf::ConfSet<uint32_t>
				{
					public:
						/* Classes */
						enum
						{
							rfFounder  = 0x00000001,
							rfAccess   = 0x00000002,
							rfIdentify = 0x00000004,
							rfRegister = 0x00000008,
							rfAll      = 0x0000000F
						};
						
						/* Functions */
						RestrictAuth();
						
						bool founder() const
								{ return (m_value & rfFounder); }
						bool access() const
								{ return (m_value & rfAccess); }
						bool identify() const
								{ return (m_value & rfIdentify); }
						bool reg() const
								{ return (m_value & rfRegister); }
					protected:
						virtual bool add(conf::Conf &conf);
				};
				
				class Expire : public conf::ConfBoolBlock
				{
					public:
						/* Data */
						conf::ConfDuration delay;
						conf::ConfDuration grace;
					
						/* Functions */
						Expire();
					protected:
						virtual bool add(conf::Conf &conf,
								conf::Conf &new_conf);
				};
				
				/* Data */
				Service service;
				
				conf::ConfString database_channels;
				conf::ConfString database_levels;
				conf::ConfString database_access;
				conf::ConfString database_akicks;
				
				conf::ConfUint<uint16_t> max_reg;
				PasswordModule password_module;
				DefaultOptions default_options;
				conf::ConfUint<uint16_t> default_bantype;
				RestrictAuth restrict_auth;
				Expire expire;
				conf::ConfUint<uint16_t> max_access;
				conf::ConfUint<uint16_t> max_akicks;
				
				conf::ConfStrChar akick_reason;
				conf::ConfDuration inhabit;
				
				conf::ConfBool list_opers;
				conf::ConfUint<uint16_t> list_max;
				conf::ConfBool clist_opers;
				conf::ConfBool sendpass_opers;
				
				/* Functions */
				ChanServ();
			protected:
				virtual bool add(conf::Conf &conf, conf::Conf &new_conf);
		};
		
		class MemoServ : public conf::ConfBlock
		{
			public:
				/* Classes */
				class RestrictAuth : public conf::ConfSet<uint32_t>
				{
					public:
						/* Classes */
						enum
						{
							rfSend  = 0x00000001,
							rfGet   = 0x00000002,
							rfAll   = 0x00000003
						};
						
						/* Functions */
						RestrictAuth();
						
						bool send() const
								{ return (m_value & rfSend); }
						bool get() const
								{ return (m_value & rfGet); }
					protected:
						virtual bool add(conf::Conf &conf);
				};
				
				/* Data */
				Service service;
				
				RestrictAuth restrict_auth;
				conf::ConfUint<uint16_t> max_memos;
				conf::ConfDuration send_delay;
				conf::ConfBool channel_target;
				conf::ConfBool forward;
				
				/* Functions */
				MemoServ();
			protected:
				virtual bool add(conf::Conf &conf, conf::Conf &new_conf);
		};
		
		class BotServ : public conf::ConfBlock
		{
			public:
				/* Classes */
				class DefaultOptions : public conf::ConfSet<uint32_t>
				{
					public:
						DefaultOptions();
					protected:
						virtual bool add(conf::Conf &conf);
				};
				
				/* Data */
				Service service;
				
				conf::ConfString database_bots;
				conf::ConfString database_kbb;
				conf::ConfString database_badwords;
				
				conf::ConfUint<uint16_t> min_users;
				conf::ConfBool smart_join;
				conf::ConfDuration keep_data;
				
				DefaultOptions default_options;
				conf::ConfUint<uint16_t> max_badwords;
				conf::ConfBool soft_reason;
				
				/* Functions */
				BotServ();
			protected:
				virtual bool add(conf::Conf &conf, conf::Conf &new_conf);
		};
		
		class OperServ : public conf::ConfBlock
		{
			public:
				/* Classes */
				class Root : public conf::ConfStrCharA
				{
					public:
						Root();
					protected:
						virtual bool apply_change(conf::Conf &conf);
				};
				
				class Wall : public conf::ConfSet<uint32_t>
				{
					public:
						/* Classes */
						enum
						{
							waNewOper         = 0x00000001,
							waNonOper         = 0x00000002,
							waGlobal          = 0x00000004,
							waMode            = 0x00000008,
							waClearModes      = 0x00000010,
							waKick            = 0x00000020,
							waAkill           = 0x00000040,
							waSGLine          = 0x00000080,
							waSQLine          = 0x00000100,
							waSZLine          = 0x00000200,
							waNoOp            = 0x00000400,
							waJupe            = 0x00000800,
							waRaw             = 0x00001000,
							waReload          = 0x00002000,
							waModules         = 0x00004000,
							waAkillExpire     = 0x00008000,
							waSGLineExpire    = 0x00010000,
							waSQLineExpire    = 0x00020000,
							waSZLineExpire    = 0x00040000,
							waExceptionExpire = 0x00080000,
							waDrop            = 0x00100000,
							waGetpass         = 0x00200000,
							waSetpass         = 0x00400000,
							waAll             = 0x007FFFFF
						};
						
						/* Functions */
						Wall();
					protected:
						virtual bool add(conf::Conf &conf);
				};
			
				/* Data */
				Service service;
				
				conf::ConfString database_stats;
				conf::ConfString database_akills;
				conf::ConfString database_sglines;
				conf::ConfString database_sqlines;
				conf::ConfString database_szlines;
				conf::ConfString database_news;
				
				Root root;
				
				Wall wall;
				
				conf::ConfUint<uint16_t> default_akilltype;
				conf::ConfDuration expire_akills;
				conf::ConfDuration expire_sglines;
				conf::ConfDuration expire_sqlines;
				conf::ConfDuration expire_szlines;
				
				conf::ConfUint<uint16_t> list_max;

				conf::ConfBool allow_raw;
				conf::ConfBool allow_modules;
				
				/* Functions */
				OperServ();
			protected:
				virtual bool add(conf::Conf &conf, conf::Conf &new_conf);
		};
		
		class Global : public conf::ConfBlock
		{
			public:
				/* Data */
				Service service;
				conf::ConfStrChar network_domain;
				
				/* Functions */
				Global();
			protected:
				virtual bool add(conf::Conf &conf, conf::Conf &new_conf);
		};
		
		/* Data */
		AllowRoot allow_root;
		Directory directory;
		DumpCore dump_core;
		PIDFile pidfile;
		Umask umask;
		conf::ConfBool strict_passwords;
		BadPasswords bad_passwords;
		conf::ConfDuration expire_timeout;
		conf::ConfDuration warning_timeout;
		conf::ConfDuration timeout_check;
		
		xlog::LogRules log;
		
		modules::ConfModuleA module;
		modules::LoadModule load_module;
		conf::ConfStringA lock_module;
		
		Mail mail;
		
		Services services;
		Session session;
		Blacklist blacklist;
		
		NickServ nickserv;
		ChanServ chanserv;
		MemoServ memoserv;
		BotServ botserv;
		Service helpserv;
		OperServ operserv;
		Global global;
		
		/* Functions */
		Directives(conf::ConfLogger &logger);
		
		virtual bool add(conf::Conf &conf);
	private:
		Directives(Directives const &);
		Directives &operator=(Directives const &);
};

/*************************************************************************/
	
} /* namespace hidden */

/*************************************************************************/

} /* namespace core */

/*************************************************************************/

#endif /* CORE_DIRECTIVES_H */
