/* Multi-language support.
 *
 * Epona (c) 2000-2002, 2004 PegSoft
 * Contact us at epona@pegsoft.net
 *
 * This program is free but copyrighted software; see the file COPYING for
 * details.
 *
 * Based on the original code of Services by Andy Church.
 *
 */

#include <system/headers.h>

#include <core/directives.h>
#include <core/epona.h>
#include <log/levels.h>

#include "epona.h"
#include "language.h"
#include "nickserv.h"
#include "services.h"

using namespace std;

/*************************************************************************/

/* The list of lists of messages. */
char **langtexts[NUM_LANGS];

/* The list of names of languages. */
char *langnames[NUM_LANGS];

/* Indexes of available languages: */
int langlist[NUM_LANGS];

/* Order in which languages should be displayed: (alphabetical) */
static int langorder[NUM_LANGS] = {
    LANG_EN_US,		/* English (US) */
    LANG_FR,		/* French */
    LANG_DE,		/* German */
    LANG_JA_JIS,	/* Japanese (JIS encoding) */
    LANG_JA_EUC,	/* Japanese (EUC encoding) */
    LANG_JA_SJIS,	/* Japanese (SJIS encoding) */
    LANG_PT,		/* Portugese */
    LANG_ES,		/* Spanish */
    LANG_TR,		/* Turkish */
    LANG_CAT,		/* Catalan */
    LANG_GR,		/* Greek */
    LANG_IT,		/* Italian */
};

/*************************************************************************/

LOGLEVELH(DebugLanguage, xlog::LogDebug, "Debug::Language", "debug: language: ");

LOGLEVELH(WarningLanguage, xlog::LogWarning, "Warning::Language", "warning: language: ");

LOGLEVELH(ErrorLanguage, xlog::LogError, "Error::Language", "Error: language: ");

/*************************************************************************/

/* Load a language file. */

static int read_int32(int32_t *ptr, FILE *f)
{
    int a = fgetc(f);
    int b = fgetc(f);
    int c = fgetc(f);
    int d = fgetc(f);
    if (a == EOF || b == EOF || c == EOF || d == EOF)
	return -1;
    *ptr = a<<24 | b<<16 | c<<8 | d;
    return 0;
}

static void load_lang(int index, const char *filename)
{
    char buf[PATH_MAX];
    FILE *f;
    int num, i;

    if (epona->debuglevel()) {
	epona->log(DebugLanguage(), "Loading language %d from file `languages/%s'",
		index, filename);
    }
    snprintf(buf, sizeof(buf), "%s/languages/%s", EPONA_DIR, filename);
    if (!(f = fopen(buf, "r"))) {
	epona->log_perror(ErrorLanguage(), "Failed to load language %d (%s)",
			index, filename);
	return;
    } else if (read_int32(&num, f) < 0) {
	epona->log(ErrorLanguage(),
		"Failed to read number of strings for language %d (%s)", index,
		filename);
	return;
    } else if (num != NUM_STRINGS) {
	epona->log(WarningLanguage(), "Bad number of strings (%d, wanted %d) "
	    "for language %d (%s)", num, NUM_STRINGS, index, filename);
    }
    langtexts[index] = static_cast<char **>(scalloc(sizeof(char *), NUM_STRINGS));
    if (num > NUM_STRINGS)
	num = NUM_STRINGS;
    for (i = 0; i < num; i++) {
	int32_t pos, len;
	fseek(f, i*8+4, SEEK_SET);
	if (read_int32(&pos, f) < 0 || read_int32(&len, f) < 0) {
	    epona->log(ErrorLanguage(),
	    		"Failed to read entry %d in language %d (%s) TOC", i, index,
	    		filename);
	    while (--i >= 0) {
		if (langtexts[index][i])
		    free(langtexts[index][i]);
	    }
	    free(langtexts[index]);
	    langtexts[index] = NULL;
	    return;
	}
	if (len == 0) {
	    langtexts[index][i] = NULL;
	} else if (len >= 65536) {
	    epona->log(ErrorLanguage(),
	    	"Entry %d in language %d (%s) is too long (over 64k)--"
	    	"corrupt TOC?", i, index, filename);
	    while (--i >= 0) {
		if (langtexts[index][i])
		    free(langtexts[index][i]);
	    }
	    free(langtexts[index]);
	    langtexts[index] = NULL;
	    return;
	} else if (len < 0) {
	    epona->log(ErrorLanguage(),
	    		"Entry %d in language %d (%s) has negative length--"
	    		"corrupt TOC?", i,
	    		index, filename);
	    while (--i >= 0) {
		if (langtexts[index][i])
		    free(langtexts[index][i]);
	    }
	    free(langtexts[index]);
	    langtexts[index] = NULL;
	    return;
	} else {
	    langtexts[index][i] = static_cast<char *>(smalloc(len+1));
	    fseek(f, pos, SEEK_SET);
	    if (fread(langtexts[index][i], 1, len, f) != 
	    		static_cast<size_t>(len)) {
		epona->log(ErrorLanguage(),
			"Failed to read string %d in language %d (%s)",
			i, index, filename);
		while (--i >= 0) {
		    if (langtexts[index][i])
			free(langtexts[index][i]);
		}
		free(langtexts[index]);
		langtexts[index] = NULL;
		return;
	    }
	    langtexts[index][i][len] = 0;
	}
    }
    fclose(f);
}

/*************************************************************************/

/* Initialize list of lists. */

void lang_init()
{
    int i, j, n = 0;

    load_lang(LANG_CAT,     "cat");
    load_lang(LANG_DE,      "de");
    load_lang(LANG_EN_US,   "en_us");
    load_lang(LANG_ES,      "es");
    load_lang(LANG_FR,      "fr");
    load_lang(LANG_GR,      "gr");
    load_lang(LANG_IT,      "it");
    load_lang(LANG_PT,      "pt");
    load_lang(LANG_TR,      "tr");

    for (i = 0; i < NUM_LANGS; i++) {
	if (langtexts[langorder[i]] != NULL) {
	    langnames[langorder[i]] = langtexts[langorder[i]][LANG_NAME];
	    langlist[n++] = langorder[i];
	    for (j = 0; j < NUM_STRINGS; j++) {
		if (!langtexts[langorder[i]][j]) {
		    langtexts[langorder[i]][j] =
				langtexts[DEF_LANGUAGE][j];
		}
		if (!langtexts[langorder[i]][j]) {
		    langtexts[langorder[i]][j] =
				langtexts[LANG_EN_US][j];
		}
	    }
	}
    }
    while (n < NUM_LANGS)
		langlist[n++] = -1;

    if (!langtexts[DEF_LANGUAGE])
		epona->fatal("Unable to load default language");
    for (i = 0; i < NUM_LANGS; i++) {
		if (!langtexts[i])
	    	langtexts[i] = langtexts[DEF_LANGUAGE];
    }
    
    /* Not what I intended to do, but these services are so archac
	 * that it's difficult to do more. */
	if ((epona->cd().nickserv.default_language =
			langlist[epona->cd().nickserv.real_deflang - 1]) < 0)
		epona->cd().nickserv.default_language = DEF_LANGUAGE;
    
    if (epona->debuglevel())
	epona->log(DebugLanguage(), "Loaded languages");
}

/*************************************************************************/
/*************************************************************************/

/* Format a string in a strftime()-like way, but heed the user's language
 * setting for month and day names.  The string stored in the buffer will
 * always be null-terminated, even if the actual string was longer than the
 * buffer size.
 * Assumption: No month or day name has a length (including trailing null)
 * greater than BUFSIZE.
 */

int strftime_lang(char *buf, int size, User *u, int format, struct tm *tm)
{
    int language = u && u->na ? u->na->ng()->language() :
    		epona->cd().nickserv.default_language;
    char tmpbuf[BUFSIZE], buf2[BUFSIZE];
    char *s;
    int i, ret;

    strscpy(tmpbuf, langtexts[language][format], sizeof(tmpbuf));
    if ((s = langtexts[language][STRFTIME_DAYS_SHORT]) != NULL) {
	for (i = 0; i < tm->tm_wday; i++)
	    s += strcspn(s, "\n")+1;
	i = strcspn(s, "\n");
	strncpy(buf2, s, i);
	buf2[i] = 0;
	strnrepl(tmpbuf, sizeof(tmpbuf), "%a", buf2);
    }
    if ((s = langtexts[language][STRFTIME_DAYS_LONG]) != NULL) {
	for (i = 0; i < tm->tm_wday; i++)
	    s += strcspn(s, "\n")+1;
	i = strcspn(s, "\n");
	strncpy(buf2, s, i);
	buf2[i] = 0;
	strnrepl(tmpbuf, sizeof(tmpbuf), "%A", buf2);
    }
    if ((s = langtexts[language][STRFTIME_MONTHS_SHORT]) != NULL) {
	for (i = 0; i < tm->tm_mon; i++)
	    s += strcspn(s, "\n")+1;
	i = strcspn(s, "\n");
	strncpy(buf2, s, i);
	buf2[i] = 0;
	strnrepl(tmpbuf, sizeof(tmpbuf), "%b", buf2);
    }
    if ((s = langtexts[language][STRFTIME_MONTHS_LONG]) != NULL) {
	for (i = 0; i < tm->tm_mon; i++)
	    s += strcspn(s, "\n")+1;
	i = strcspn(s, "\n");
	strncpy(buf2, s, i);
	buf2[i] = 0;
	strnrepl(tmpbuf, sizeof(tmpbuf), "%B", buf2);
    }
    ret = strftime(buf, size, tmpbuf, tm);
    if (ret == size)
	buf[size-1] = 0;
    return ret;
}

/*************************************************************************/
/*************************************************************************/

/* Send a syntax-error message to the user. */

void syntax_error(const char *service, User *u, const char *command, int msgnum, ...)
{
	const char *str = getstring(u->na, msgnum);

	char buf[BUFSIZE];

	va_list args;
	va_start(args, msgnum);
	vsnprintf(buf, sizeof(buf), str, args);
	va_end(args);

	notice_lang(service, u, SYNTAX_ERROR, buf);
	notice_lang(service, u, MORE_INFO, service, command);
}

/*************************************************************************/
