/* MemoServ header file.
 *
 * Epona (c) 2000-2002, 2004 PegSoft
 * Contact us at epona@pegsoft.net
 *
 * This program is free but copyrighted software; see the file COPYING for
 * details.
 *
 * Based on the original code of Services by Andy Church.
 *
 */

#ifndef MEMOSERV_H
#define MEMOSERV_H

/*************************************************************************/

#include <containers/vectorsp.h>
#include <databases/collection.h>
#include <databases/record.h>
#include <log/levels.h>

#include "tweak.h"
#include "services.h"

/*************************************************************************/

class NickGroup;

/*************************************************************************/

/* Memo info structures. */

class Memo : public databases::Record
{
	public:
		/* Data */
		static databases::Field *f_owner;
		static databases::Field *f_number;
		static databases::Field *f_flags;
		static databases::Field *f_sender;
		static databases::Field *f_target;
		static databases::Field *f_created;
		static databases::Field *f_text;
		
		/* Consts */
	
		/// Maximum length of the sender.
		static size_t const sender_len = 63;
		/// Maximum length of the target.
		static size_t const target_len = 63;
		/// Maximum length of the text.
		static size_t const text_len = 4096;
		/// Maximum length of a line in text.
		static size_t const line_len = 475;
		
		/// Memo has not yet been read.
		static uint16_t const mfUnread    = 0x0001;
		/// Memo has multiple lines
		static uint16_t const mfMultiLine = 0x0002;
		/// Unused flags.
		static uint16_t const mfUnused    = 0xFFFC;
		
		/* Functions */
		
		/// Constructor.
		Memo();
		/// Constructor setting the record ID.
		Memo(identifier_type id);
		
		/// Overrides Record::get_s().
		virtual char const *get_s(databases::Field const *field) const;
		/// Overrides Record::set_s().
		virtual char const *set_s(databases::Field const *field,
				char const *value);
		/// Overrides Record::get_ui().
		virtual uint32_t get_ui(databases::Field const *field) const;
		/// Overrides Record::set_ui().
		virtual uint32_t set_ui(databases::Field const *field, uint32_t value);
		/// Overrides Record::get_ut().
		virtual time_t get_ut(databases::Field const *field) const;
		/// Overrides Record::set_ut().
		virtual time_t set_ut(databases::Field const *field, time_t value);
		/// Overrides Record::get_ri().
		virtual identifier_type get_ri(databases::Field const *field) const;
		/// Overrides Record::set_ri().
		virtual identifier_type set_ri(databases::Field const *field,
				identifier_type value);
		
		/// Overrides Record::end_update().
		virtual bool end_update();
		
		/// Overrides Record::memory_usage().
		virtual size_t memory_usage() const;
		
		/// Returns the owner group of the memo.
		NickGroup *owner() const { return m_owner; }
		/// Sets the owner group of the memo.
		NickGroup *owner(NickGroup *owner);
		/// Sets the owner group ID of the memo (for databases only!)
		identifier_type owner_id(identifier_type owner_id);
		
		/// Returns the index number (not necessarily array position!)
		uint32_t number() const { return m_number; }
		/// Sets the index number.
		uint32_t number(uint32_t number);

		/// Returns the flags (see mf*)
		uint16_t flags() const { return m_flags; }
		/// Sets the flags -- must be called by databases only!
		uint16_t flags(uint16_t flags);
		/// Adds flags.
		uint16_t flags_add(uint16_t flags);
		/// Removes flags.
		uint16_t flags_remove(uint16_t flags);
		
		/// Returns the sender of the memo.
		char const *sender() const { return m_sender; }
		/// Sets the sender of the memo.
		char const *sender(char const *sender);
		
		/// Returns the target of the memo.
		char const *target() const { return m_target; }
		/// Sets the target of the memo.
		char const *target(char const *target);
		
		/// Returns the creation time of the memo.
		time_t created() const { return m_created; }
		/// Sets the creation time of the memo.
		time_t created(time_t created);
	    
	    /// Returns the text of the memo.
		char const *text() const { return m_text; }
		/// Sets the text of the memo.
		char const *text(char const *text);
	private:
		NickGroup *m_owner;
		
		uint32_t m_number;
		uint16_t m_flags;
		
		misc::cstring m_sender;
		misc::cstring m_target;
		time_t m_created;
		misc::cstring m_text;
			
		bool check_text() const;
};

/*************************************************************************/

class MemoInfo 
{	
	public:
		/* Typedefs */
		typedef containers::VectorSP<Memo *, uint16_t> list_type;
		
		/* Functions */
		
		MemoInfo(NickGroup *ng);
		
		list_type &memos() { return m_memos; }
		list_type const &memos() const { return m_memos; }
	    
		int16_t memomax() const;
		int16_t memomax(int16_t memomax);
	private:
		NickGroup *m_ng;
		
		list_type m_memos;
};

/*************************************************************************/

class Memos : public databases::Collection
{
	public:
		/// Implements Collection::create().
		virtual databases::Record *
				create(databases::Record::identifier_type id = 0) const;
		/// Overrides Collection::add().
		virtual bool add(databases::Record *record, int *err = 0);
		
		/// Overrides Collection::clear().
		virtual bool clear(int *err = 0) { return false; }
		
		/// Configures owner collection.
		void setup_owner();
		
		/// Sets the database up.
		/** Sets the database up. This is called by the Module::notify()
		  * function when the database module is loaded. */
		void setup_database(databases::Database *db);
		
		/// Overrides Collection::owner_field().
		virtual databases::Field const
				*owner_field(Collection const *coll) const
				{ return Memo::f_owner; }
	protected:
		/// Overrides Collection::add_list().
		virtual bool add_list(databases::Record *record);
		/// Overrides Collection::remove_list().
		virtual void remove_list(databases::Record *record);
		
		/// Overrides Collection::remove_owned_db().
		virtual bool remove_owned_db(databases::Collection const *coll,
				databases::Record const *record, int *err = 0);
		/// Overrides Collection::remove_owned_list().
		virtual bool remove_owned_list(databases::Collection const *coll,
				databases::Record const *record, int *err = 0);
	private:
		void mail_memo(Memo const *m) const;
};

/*************************************************************************/

extern Memos memos;

/*************************************************************************/

LOGLEVELH(InfoMemoServ, xlog::LogInfo, "Info::MemoServ", "memoserv: ");
LOGLEVELH(ErrorMemoServ, xlog::LogError, "Error::MemoServ", "Error: memoserv: ");
LOGLEVELH(ErrorMemoServList, xlog::LogError, "Error::MemoServ::List",
		"Error: memoserv list: ");

/*************************************************************************/

#endif /* MEMOSERV_H */
