/* Compressor.
 *
 * Bzip2 file module (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#if HAVE_SYS_STAT_H
# include <sys/stat.h>
#endif

#if HAVE_FCNTL_H
# include <fcntl.h>
#endif

#include <misc/string.h>

#include "bz2file.h"

using namespace std;

using misc::strncpyzt;

/*************************************************************************/

static void cleanup(int fd, FILE *out, BZFILE *bzf, char const *newpath)
{
	if (fd != -1)
		while ((close(fd) == -1) && (errno == EINTR))
			;
	if (bzf)
	{
		int bzerror;
		BZ2_bzWriteClose(&bzerror, bzf, 1, 0, 0);
	}
	if (out)
	{
		while ((fclose(out) == EOF) && (errno == EINTR))
			;
	}
	remove(newpath);
}

/*************************************************************************/

namespace bz2file
{

/*************************************************************************/

bool Compressor::compress(char const *path, int level)
{
	*m_error = '\0';
	
	int fd = -1;
	
	if ((fd = open(path, O_RDONLY)) == -1)
	{
		set_errno_error("Read error");
		return false;
	}
	
	char newpath[PATH_MAX];
	snprintf(newpath, sizeof(newpath), "%s.bz2", path);
	
	FILE *out = fopen(newpath, "wb");
	if (!out)
	{
		set_errno_error("Write error");
		cleanup(fd, 0, 0, newpath);
		return false;
	}
	
	if ((level == 0) || (level < 1) || (level > 9))
		level = 4;
	
	int bzerror;
	BZFILE *bzf = BZ2_bzWriteOpen(&bzerror, out, level, 0, 0);
	
	if (bzerror != BZ_OK)
	{
		strncpyzt(m_error, "Not enough memory", sizeof(m_error));
		cleanup(fd, out, bzf, newpath);
		return false;
	}
	
	ssize_t bytes = 0;
	char buf[32768];
	
	for (;;)
	{
		while ((bytes = read(fd, buf, sizeof(buf))) == -1)
		{
			if (errno == EINTR)
				continue;
			
			set_errno_error("Read error");
			cleanup(fd, out, bzf, newpath);
			return false;
		}
		
		if (bytes == 0)
			break;
		
		BZ2_bzWrite(&bzerror, bzf, buf, bytes);
		
		if (bzerror != BZ_OK)
		{
			set_errno_error("Write error");
			cleanup(fd, out, bzf, newpath);
			return false;
		}
	}
	
	BZ2_bzWriteClose(&bzerror, bzf, 0, 0, 0);
	if (bzerror != BZ_OK)
	{
		set_errno_error("Write error");
		cleanup(fd, out, bzf, newpath);
		return false;
	}
	
	while ((fclose(out) == EOF))
	{
		if (errno != EINTR)
		{
			set_errno_error("Write error");
			cleanup(fd, out, 0, newpath);
			return false;
		}
	}
	
	while ((close(fd) == -1) && (errno == EINTR))
		;
	
	return true;
}

/*************************************************************************/

void Compressor::set_errno_error(char const *prefix)
{
	if (errno)
		errorf("%s: %s", prefix, strerror(errno));
	else
		errorf("%s", prefix);
}

/*************************************************************************/

} /* namespace bz2file */

/*************************************************************************/
