/* MySQL connection class.
 *
 * MySQL client module (c) 2006 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#include <errmsg.h>
#include <mysql.h>

#include <core/version.h>
#include <mysql-cli/conn.h>
#include <mysql-cli/result.h>

using namespace std;
using namespace mysql_cli;

/*************************************************************************/

namespace mysql_cli
{

/*************************************************************************/

char const *Conn::client_info()
{
	return mysql_get_client_info();
}

/*************************************************************************/

unsigned long Conn::client_version()
{
	return mysql_get_client_version();
}

/*************************************************************************/

Conn::~Conn()
{
	close();
	
	if (m_e)
		delete m_e;
}

/*************************************************************************/

Conn &Conn::operator=(Conn const &right)
{
	if (this != &right)
	{
		close();
		
		m_ctimeout = right.m_ctimeout;
		m_reconnect = right.m_reconnect;
		
		m_host = right.m_host;
		m_port = right.m_port;
		m_unixsocket = right.m_unixsocket;
		m_user = right.m_user;
		m_password = right.m_password;
		m_db = right.m_db;
		
		m_charset = right.m_charset;
		
		if (m_e)
			delete m_e;
		m_e = (right.m_e ? right.m_e->copy() : 0);
		
		m_logqueries = right.m_logqueries;
		m_logwarnings = right.m_logwarnings;
	}
	
	return *this;
}

/*************************************************************************/

my_ulonglong Conn::affected_rows()
{
	if (!m_connected)
		return 0;
	
	return mysql_affected_rows(m_handle);
}

/*************************************************************************/

bool Conn::change_user(char const *user, char const *password, char const *db)
{
	if (!m_connected)
		return false;
	
	m_gone = false;
	
	if (mysql_change_user(m_handle, user, password, db))
	{
		if (errnum() == CR_CONNECTION_ERROR || errnum() == CR_CONN_HOST_ERROR ||
				errnum() == CR_SERVER_GONE_ERROR || errnum() == CR_SERVER_LOST)
			m_gone = true;
		
		if (m_e)
			m_e->log_event(*this, Events::evError, 0, "Change of user failed");
		return false;
	}
	
	this->user(user);
	this->password(password);
	this->db(db);
	
	return true;
}

/*************************************************************************/

char const *Conn::charset()
{
	if (!m_connected)
		return m_charset;
	else
		return mysql_character_set_name(m_handle);
}

/*************************************************************************/

char const *Conn::charset(char const *charset)
{
	if (m_charset.compare(charset))
	{
		if (m_connected && charset &&
				mysql_set_character_set(m_handle, charset))
		{
			if (m_e)
				m_e->log_event(*this, Events::evError, 0,
						"Setting of connection character set failed");
			return m_charset;
		}
		
		m_charset = charset;
	}
	
	return this->charset();
}

/*************************************************************************/

void Conn::close()
{
	if (m_connected)
	{
		if (m_e)
			m_e->log_event(*this, Events::evClosed, 0, 0);
		m_connected = m_gone = false;
		m_queries = m_since = 0;
		m_exectime = 0;
	}
	
	if (m_handle)
	{
		mysql_close(m_handle);
		m_handle = 0;
	}
}

/*************************************************************************/

bool Conn::connect()
{
	if (m_connected)
		return true;
	
	if (!m_handle && !(m_handle = mysql_init(0)))
	{
		if (m_e)
			m_e->log_event(*this, Events::evError, 0, "Out of memory");
		return false;
	}
	
	if (m_ctimeout && mysql_options(m_handle, MYSQL_OPT_CONNECT_TIMEOUT,
			(char const *)&m_ctimeout))
	{
		if (m_e)
			m_e->log_event(*this, Events::evError, 0,
					"Setting of connection timeout failed");
		return false;
	}
	
	if (mysql_options(m_handle, MYSQL_READ_DEFAULT_GROUP,
			core::program_name_lc))
	{
		if (m_e)
			m_e->log_event(*this, Events::evError, 0,
					"Setting of default group failed");
		return false;
	}
	
#if defined(MYSQL_VERSION_ID) && MYSQL_VERSION_ID >= 50013
	if (mysql_options(m_handle, MYSQL_OPT_RECONNECT,
			(char const *)&m_reconnect))
	{
		if (m_e)
			m_e->log_event(*this, Events::evError, 0,
				"Setting of automatic reconnection option failed");
		return false;
	}
#else
	m_handle->reconnect = 1;
#endif

	if (mysql_real_connect(m_handle, m_host, m_user, m_password, m_db,
			m_port, m_unixsocket, 0) != m_handle)
	{
		if (m_e)
			m_e->log_event(*this, Events::evError, 0, "Connect error");
		return false;
	}
	
	/* May be necessary to set it again after the connection has been
	 * established. */
#if defined(MYSQL_VERSION_ID) && MYSQL_VERSION_ID >= 50013
	if (mysql_options(m_handle, MYSQL_OPT_RECONNECT,
			(char const *)&m_reconnect))
	{
		if (m_e)
			m_e->log_event(*this, Events::evError, 0,
				"Setting of automatic reconnection option failed");
		return false;
	}
#else
	m_handle->reconnect = 1;
#endif

	if (m_charset && mysql_set_character_set(m_handle, m_charset))
	{
		if (m_e)
			m_e->log_event(*this, Events::evError, 0,
				"Setting of connection character set failed");
		return false;
	}

	m_connected = true;
	m_gone = false;
	m_queries = 0;
	m_since = time(0);
	
	if (m_e)
		m_e->log_event(*this, Events::evConnected, 0, 0);
	return true;
}

/*************************************************************************/

unsigned int Conn::errnum()
{
	if (!m_handle)
		return 0;
	
	return mysql_errno(m_handle);
}

/*************************************************************************/

char const *Conn::error()
{
	if (!m_handle)
		return 0;
	
	return mysql_error(m_handle);
}

/*************************************************************************/

unsigned long Conn::escape_string(misc::cstring &dest,
		char const *src, long len)
{
	if (len == -1)
		len = strlen(src);
	
	dest.reserve(2 * len + 1);
	
	return mysql_real_escape_string(m_handle, dest.str(), src, len);
}

/*************************************************************************/

char const *Conn::escape_string(char const *str, long len)
{
	if (len == -1)
		len = strlen(str);
	
	if (m_essize < (size_t) (2 * len + 1))
	{
		m_essize = (2 * len + 1);
		m_es.reserve(m_essize);
	}
	
	mysql_real_escape_string(m_handle, m_es.str(), str, len);
	return m_es;
}

/*************************************************************************/

Events *Conn::events(Events *events)
{
	if (m_e)
		delete m_e;
	
	return (m_e = events);
}

/*************************************************************************/

unsigned int Conn::field_count()
{
	if (!m_connected)
		return 0;
	
	return mysql_field_count(m_handle);
}

/*************************************************************************/

char const *Conn::host_info()
{
	if (!m_connected)
		return 0;
	
	return mysql_get_host_info(m_handle);
}

/*************************************************************************/

char const *Conn::info()
{
	if (!m_connected)
		return 0;
	
	return mysql_info(m_handle);
}

/*************************************************************************/

my_ulonglong Conn::insert_id()
{
	if (!m_connected)
		return 0;
	
	return mysql_insert_id(m_handle);
}

/*************************************************************************/

bool Conn::query(char const *query, unsigned long length,
		Result *result, bool use)
{
	if (!m_connected)
		return false;
	
	m_gone = false;
	
	/* Always increase the query count, even if it fails... */
	++m_queries;
	
#if HAVE_GETTIMEOFDAY
	struct timeval exs;
	gettimeofday(&exs, 0);
#endif
	
	if (mysql_real_query(m_handle, query, length))
	{
		/* An error occured... */
		
		/* Has the server gone away? */
		if (errnum() == CR_CONNECTION_ERROR || errnum() == CR_CONN_HOST_ERROR ||
				errnum() == CR_SERVER_GONE_ERROR | errnum() == CR_SERVER_LOST)
			m_gone = true;
		
		/* Log the failing query, if requested */
		if (m_logqueries)
			if (m_e)
				m_e->log_event(*this, Events::evQuery, 0, query);
		/* Then log the error */
		if (m_e)
			m_e->log_event(*this, Events::evError, 0, "Query failure");
		/* Then return false */
		return false;
	}
	
	MYSQL_RES *r = (use ? mysql_use_result(m_handle) :
			mysql_store_result(m_handle));

#if HAVE_GETTIMEOFDAY
	struct timeval exe;
	gettimeofday(&exe, 0);
	
	m_exectime = (exe.tv_sec + (double)exe.tv_usec / 1000000) -
			  (exs.tv_sec + (double)exs.tv_usec / 1000000);
#endif
	
	if (!r)
	{
		/* There is no result set... Is this an error? */
		if (field_count())
		{
			/* Yes it is! */
			
			/* Has the server gone away? */
			if (errnum() == CR_CONNECTION_ERROR ||
					errnum() == CR_CONN_HOST_ERROR ||
					errnum() == CR_SERVER_GONE_ERROR ||
					errnum() == CR_SERVER_LOST)
				m_gone = true;
			
			/* Log the failing query, if requested */
			if (m_logqueries && m_e)
				m_e->log_event(*this, Events::evQuery, 0, query);
			
			/* Report the error */
			if (m_e)
				m_e->log_event(*this, Events::evError, 0, "Retrieving rows failed");
	
			/* Return false */
			return false;
		}
		
		/* The query really returned no rows, so proceed */
	}
	
	/* Log the query now -- AFTER mysql_store_result() has been called so
	 * we don't get a -1 return value for SELECT statements in
	 * affected_rows(). */
	if (m_logqueries && m_e)
		m_e->log_event(*this, Events::evQuery, 0, query);
	
	/* Assign the result to the given result object, if available.
	 * Else, just free it. */ 
	if (result)
		*result = r;
	else if (r)
		mysql_free_result(r);
	
	if (m_logqueries && m_logwarnings && m_e && warning_count())
	{
		Result wres;
		if (this->query("SHOW WARNINGS", &wres))
		{
			MYSQL_ROW row;
			
			while ((row = wres.fetch_row()))
			{
				if (!stricmp(row[0], "Note"))
					m_e->log_event(*this, Events::evWarningNote, atoi(row[1]), row[2]);
				else if (!stricmp(row[0], "Warning"))
					m_e->log_event(*this, Events::evWarning, atoi(row[1]), row[2]);
				else if (!stricmp(row[0], "Error"))
					m_e->log_event(*this, Events::evWarningError, atoi(row[1]), row[2]);
			}
		}
	}
	
	return true;
}

/*************************************************************************/

bool Conn::query(char const *query, Result *result, bool use)
{
	return this->query(query, strlen(query), result, use);
}

/*************************************************************************/

bool Conn::queryf(Result *result, char const *fmt, ...)
{
	char buf[16384];
	
	va_list args;
	va_start(args, fmt);
	vsnprintf(buf, sizeof(buf), fmt, args);
	va_end(args);	

	return this->query(buf, result);
}

/*************************************************************************/

bool Conn::ping()
{
	if (!m_connected)
		return false;
	
	m_gone = false;
	
	if (mysql_ping(m_handle))
	{
		m_gone = true;
		
		if (m_e)
			m_e->log_event(*this, Events::evError, 0, "Dead connection");
		return false;
	}
	
	if (m_e)
		m_e->log_event(*this, Events::evPing, 0, "Connection is alive");	
	
	return true;
}

/*************************************************************************/

unsigned int Conn::proto_info()
{
	if (!m_connected)
		return 0;
	
	return mysql_get_proto_info(m_handle);
}

/*************************************************************************/

bool Conn::select_db(char const *db)
{
	if (!m_connected)
		return false;
	
	m_gone = false;
	
	if (mysql_select_db(m_handle, db))
	{
		if (errnum() == CR_CONNECTION_ERROR || errnum() == CR_CONN_HOST_ERROR ||
				errnum() == CR_SERVER_GONE_ERROR || errnum() == CR_SERVER_LOST)
			m_gone = true;
		
		if (m_e)
			m_e->log_event(*this, Events::evError, 0, "Change of database failed");
		return false;
	}
	
	this->db(db);
	
	return true;
}

/*************************************************************************/

char const *Conn::server_info()
{
	if (!m_connected)
		return 0;
	
	return mysql_get_server_info(m_handle);
}

/*************************************************************************/

unsigned long Conn::server_version()
{
	if (!m_connected)
		return 0;
	
	return mysql_get_server_version(m_handle);
}

/*************************************************************************/

char const *Conn::stat()
{
	if (!m_connected)
		return 0;
	
	m_gone = false;
	
	char const *status = mysql_stat(m_handle);
	
	if (!status)
	{
		if (errnum() == CR_CONNECTION_ERROR || errnum() == CR_CONN_HOST_ERROR ||
				errnum() == CR_SERVER_GONE_ERROR || errnum() == CR_SERVER_LOST)
			m_gone = true;
		
		if (m_e)
			m_e->log_event(*this, Events::evError, 0,
					"Retrieving server status failed");
		return 0;
	}
	
	return status;
}

/*************************************************************************/

unsigned long Conn::thread_id()
{
	if (!m_connected)
		return 0;
	
	return mysql_thread_id(m_handle);
}

/*************************************************************************/

unsigned int Conn::warning_count()
{
	if (!m_connected)
		return 0;
	
	return mysql_warning_count(m_handle);
}

/*************************************************************************/

} /* namespace mysql_cli */

/*************************************************************************/
