/* MySQL connection class.
 *
 * MySQL client module (c) 2006 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/** \file mysql-cli/conn.h
  * \brief MySQL connection header file. */

#ifndef MYSQL_CLI_CONN_H
#define MYSQL_CLI_CONN_H

/*************************************************************************/

#include <mysql.h>

#include <misc/string.h>

/*************************************************************************/

/// Namespace of the MySQL client classes

namespace mysql_cli
{

/*************************************************************************/

class Conn;
class Result;

/*************************************************************************/

/// MySQL connection events class.

/** This class is used by the Conn class as an event notifier. It must 
  * be inherited in another class and the log_event() function must
  * be overriden. Then, Conn::events() must be called to register
  * the event notifier. */

class Events
{
	public:
		/// Log events
		enum
		{
			/// Connected to the MySQL database.
			evConnected = 1,
			/// Successful ping
			evPing = 2,
			/// Connection closed.
			evClosed = 3,
			/// Connection error.
			evError = 4,
			/// An executed query
			evQuery = 5,
			/// Warning: note level
			evWarningNote = 6,
			/// Warning: warning level
			evWarning = 7,
			/// Warning: error level
			evWarningError = 8
		};

		/// Destructor.
		virtual ~Events() { }
		
		/// Makes a copy of the object.
		/** Makes a copy of the object. This is called by Conn's
		  * copy constructor and assignment operator.
		  *
		  * \return A copy of the object. */
		virtual Events *copy() const = 0;
		
		/// Logs an event.
		/** Logs an event.
		  *
		  * \param[in] conn The connection where the event happened.
		  * \param[in] event Can be one of the following values:
		  * evConnected, evClosed, evError, evQuery, evWarningNote,
		  * evWarning or evWarningError.
		  * \param[in] code For evWarningNote, evWarning and
		  * evWarningError, the corresponding warning code.
		  * \param[in] msg A message associated to the event, if any.
		  * For evError events, it is the context of the error, for
		  * evQuery it is the query that has been executed, and for
		  * evWarningNote, evWarning and evWarningError, the contents
		  * of the warnings. */
		virtual void log_event(Conn &conn, int event, unsigned int code,
				char const *msg) const = 0;
};

/*************************************************************************/

/// MySQL connection class.

/** This class encapsulates a MySQL connection.
  *
  * See the MySQL C API for more detailed information on some of the
  * functions. */

class Conn
{
	public:
		/// Returns the MySQL client library version string.
		static char const *client_info();
		/// Returns the MySQL client library version.
		static unsigned long client_version();
	
		/// Default constructor.
		Conn() : m_handle(0), m_connected(false), m_gone(false), m_queries(0),
				m_since(0), m_exectime(0), m_ctimeout(0), m_reconnect(true),
				m_host(0), m_port(0), m_unixsocket(0), m_user(0),
				m_password(0), m_db(0), m_charset(0), m_es(0), m_essize(0),
				m_e(0), m_logqueries(false), m_logwarnings(false) { }
	 	/// Constructor with the same parameters as mysql_connect().
	 	Conn(char const *host, char const *user, char const *password,
	 				char const *db, unsigned int port, char const *us) :
	 			m_handle(0), m_connected(false), m_gone(false), m_queries(0),
	 			m_since(0), m_exectime(0), m_ctimeout(0), m_reconnect(true),
	 			m_host(host), m_port(port), m_unixsocket(us), m_user(user),
	 			m_password(password), m_db(db), m_charset(0), m_es(0),
	 			m_essize(0), m_e(0), m_logqueries(false),
	 			m_logwarnings(false) { }
		/// Copy constructor.
		/** Copy constructor. It copies the various data fields of the
		  * connection (like host, username, password...) but does not
		  * reestablish a connection.
		  *
		  * \param[in] c The connection to copy. */
		Conn(Conn const &c) : m_handle(0), m_connected(false), m_gone(false),
				m_queries(0), m_since(0), m_exectime(0),
				m_ctimeout(c.m_ctimeout), m_reconnect(c.m_reconnect),
				m_host(c.m_host), m_port(c.m_port),
				m_unixsocket(c.m_unixsocket), m_user(c.m_user),
				m_password(c.m_password), m_db(c.m_db),
				m_charset(c.m_charset), m_es(0), m_essize(0),
				m_e(c.m_e ? c.m_e->copy() : 0), m_logqueries(c.m_logqueries),
				m_logwarnings(c.m_logwarnings) { }
		/// Assignment operator.
		/** Assignment operator. Like the copy constructor, a new
		  * connection is not established.
		  *
		  * \param[in] right The connection that is assigned.
		  * \return The connection object.
		  * \sa Conn(Conn const &) */
		Conn &operator=(Conn const &right);
		/// Destructor.
		~Conn();
		
		/// Connects to the MySQL server.
		/** Connects to the MySQL server.
		  *
		  * \return true if successful, false otherwise. An evConnected
		  * event is logged if successful; an evError event is logged
		  * otherwise. */
		bool connect();
		
		/// Changes the current MySQL user.
		/** Changes the current MySQL user. If the user, password or
		  * database is invalid, then no change is made and the
		  * connection is still valid. Else, the user(), password()
		  * and db() fields are updated automatically.
		  *
		  * \param[in] user The user to change to.
		  * \param[in] password The user's password.
		  * \param[in] db The database to use.
		  * \return true if successful, false otherwise. */
		bool change_user(char const *user, char const *password,
				char const *db);
		
		/// Changes the current MySQL database.
		/** Changes the current MySQL database. If the database can't
		  * be changed, then no change is made and the connection is
		  * still valid. Else, the db() field is updated automatically.
		  *
		  * \param[in] db The database to use.
		  * \return true if successful, false otherwise. */
		bool select_db(char const *db);
		
		/// Reconnects to the MySQL server.
		/** Reconnects to the MySQL server. This is equivalent to calling
		  * close() and connect().
		  *
		  * \return The return value of connect(). */
		bool reconnect() { close(); return connect(); }
		
		/// "Pings" the MySQL server.
		/** "Pings" the MySQL server, i.e. checks whether the connection
		  * is working. If the connection has gone down, a reconnection
		  * is attempted.
		  *
		  * \return true if the connection works, false otherwise. */
		bool ping();
		
		/// Closes the connection to the MySQL server.
		/** Closes the connection to the MySQL server, if a connection
		  * is established.
		  *
		  * An evClosed event is logged. */
		void close();
		
		/// Escapes a string.
		/** Escapes a string in a way that is suitable to use in a query.
		  *
		  * \param[in] dest The destination string, a misc::cstring object
		  * that will automatically be sized accordingly.
		  * \param[in] src The string to escape.
		  * \param[in] len The length of the string to escape. If this is -1,
		  * then the length is computed automatically.
		  * \return The length of the destination string. */
		unsigned long escape_string(misc::cstring &dest, char const *src,
				long len = -1);
		
		/// Escapes a string.
		/** Escapes a string in a way that is suitable to use in a query.
		  * The returned string is stored in the object instance for which
		  * the method is called; therefore, further calls to this method
		  * for the same object will overwrite the contents of the returned
		  * string.
		  *
		  * \param[in] str The string to escape.
		  * \param[in] len The length of the string to escape. If this is -1,
		  * then the length is computed automatically.
		  * \return The escaped string. */
		char const *escape_string(char const *str, long len = -1);
		
		/// Executes a query.
		/** Executes a SQL query.
		  *
		  * \param[in] query The query to execute.
		  * \param[in] length The length of the query, in bytes.
		  * \param[in] result If different than NULL, the result of
		  * the query will be assigned to this result object. If it
		  * is NULL, and the query returned a result, the function
		  * will automatically retrieve and free it, so that
		  * another query can take place afterwards.
		  * \param[in] use If true, mysql_use_result() is used to
		  * get the result set instead of mysql_store_result().
		  * \return true if the query was successful and the 
		  * retrieval of the result set succeeded, false otherwise. */
		bool query(char const *query, unsigned long length,
				Result *result, bool use = false);
		/// Same as query(query, strlen(query), result, use);
		bool query(char const *query, Result *result, bool use = false);
		/// printf()-like interface to query() -- limited to queries with 16384 characters.
		bool queryf(Result *result, char const *fmt, ...) FORMAT(printf,3,4);
		
		/// Returns whether the MySQL connection is established.
		bool connected() const { return m_connected; }
		/// Returns whether the server is gone or lost.
		/** Returns whether the server is gone or lost (i.e. if one of
		  * the MySQL functions returned CR_SERVER_GONE_ERROR or
		  * CR_SERVER_LOST), if the connection has not been reestablished.
		  *
		  * \return true if it is gone or lost, false otherwise. */
		bool gone() const { return m_gone; }
		
		/// Returns the number of queries that have been issued since connection.
		unsigned int queries() const { return m_queries; }
		/// Returns when the connection was established.
		time_t since() const { return m_since; }
		/// Returns the execution time for the last query.
		double exectime() const { return m_exectime; }
		
		/// Returns the MySQL connection thread ID.
		unsigned long thread_id();
		
		/// Returns the last error number that happened on the mysql connection.
		unsigned int errnum();
		/// Returns a string describing the last error that happened on the mysql connection.
		char const *error();
		
		/// Returns a string describing the connection, including server hostname.
		char const *host_info();
		/// Returns the protocol version used by the connection.
		unsigned int proto_info();
		/// Returns a string that represents the server version number.
		char const *server_info();
		/// Returns the version number of the server.
		unsigned long server_version();
		/// Returns server status string.
		char const *stat();
		
		/// Returns the last insert ID.
		my_ulonglong insert_id();
		
		/// Returns the number of rows affected by the last query.
		my_ulonglong affected_rows();
		/// Returns the number of warnings for the last query.
		unsigned int warning_count();
		/// Returns information about the last query.
		/** Returns information about the last query, but only for a
		  * few statements, including INSERT, ALTER TABLE and UPDATE.
		  *
		  * \return Information about the last query. */
		char const *info();
		
		/// Returns the number of fields for the last query.
		unsigned int field_count();
		
		/// Returns the connection timeout (in seconds).
		/** Returns the connection timeout (in seconds).
		  *
		  * \return The connection timeout, which may be zero if the
		  * default value must be used. */
		unsigned int connect_timeout() const { return m_ctimeout; }
		/// Sets the connection timeout (in seconds).
		/** Sets the connection timeout (in seconds).
		  *
		  * \param[in] timeout The connection timeout. If it is zero,
		  * then the default value is used.
		  * \return The new connection timeout. */
		unsigned int connect_timeout(unsigned int timeout)
				{ return (m_ctimeout = timeout); }
		
		/// Returns whether the connection is reestablished automatically.
		/** Returns whether the connection is reestablished automatically if
		  * it was dropped and a query is made.
		  *
		  * \return true if the connection is reestablished automatically,
		  * false otherwise. By default, it is true. */
		my_bool reconnect() const { return m_reconnect; }
		
		/// Sets whether the connection is reestablished automatically.
		/** Sets whether the connection is reestablished automatically if
		  * it was dropped and a query is made.
		  *
		  * \param[in] reco true if the connection must be reestablished,
		  * false otherwise.
		  * \return The option's new value. */
		my_bool reconnect(my_bool reco) { return (m_reconnect = reco); }
		
		/// Returns the MySQL hostname.
		char const *host() const { return m_host; }
		/// Sets the MySQL hostname.
		char const *host(char const *host) { return (m_host = host); }
		
		/// Returns the MySQL port.
		unsigned int port() const { return m_port; }
		/// Sets the MySQL port.
		unsigned int port(unsigned int port) { return (m_port = port); }
		
		/// Returns the MySQL hostname.
		char const *unix_socket() const { return m_unixsocket; }
		/// Sets the MySQL hostname.
		char const *unix_socket(char const *us) { return (m_unixsocket = us); }
		
		/// Returns the MySQL username.
		char const *user() const { return m_user; }
		/// Sets the MySQL username.
		char const *user(char const *user) { return (m_user = user); }
		
		/// Returns the MySQL password.
		char const *password() const { return m_password; }
		/// Sets the MySQL username.
		char const *password(char const *password)
				{ return (m_password = password); }
		
		/// Returns the MySQL database.
		char const *db() const { return m_db; }
		/// Sets the MySQL database.
		char const *db(char const *db) { return (m_db = db); }
		
		/// Returns the connection character set.
		char const *charset();
		/// Sets the connection character set.
		char const *charset(char const *charset);
		
		/// Returns the event notifier object.
		Events const *events() const { return m_e; }
		/// Sets the event notifier object.
		/** Sets the event notifier object.
		  *
		  * \param[in] events A pointer to the event notifier object.
		  * Conn will automatically free it when it's no longer
		  * needed.
		  * \return The new event notifier object. */
		Events *events(Events *events);
		
		/// Returns whether queries are logged.
		bool log_queries() const { return m_logqueries; }
		/// Sets whether queries are logged.
		bool log_queries(bool value)
				{ return (m_logqueries = value); }
		
		/// Returns whether warnings are logged.
		bool log_warnings() const { return m_logwarnings; }
		/// Sets whether warnings are logged.
		/** Sets whether warnings are logged. Warnings may only be
		  * logged after a call to the execute() method.
		  *
		  * \param[in] value true if the warnings need to be logged,
		  * false otherwise. */
		bool log_warnings(bool value)
				{ return (m_logwarnings = value); }
	protected:
		/// The MySQL connection handle.
		MYSQL *m_handle;
		/// Is it currently connected?
		bool m_connected;
		/// Is the server lost or gone away?
		bool m_gone;
		
		/// How much queries have been done?
		unsigned int m_queries;
		/// Since when are we connected?
		time_t m_since;
		/// Last query execution time.
		double m_exectime;
		
		/// Connect timeout (in seconds).
		unsigned int m_ctimeout;
		/// Reconnect automatically ?
		my_bool m_reconnect;
		
		/// Server's hostname
		misc::cstring m_host;
		/// Server's port
		unsigned int m_port;
		/// Server's unix socket
		misc::cstring m_unixsocket;
		
		/// MySQL username
		misc::cstring m_user;
		/// MySQL password
		misc::cstring m_password;
		/// MySQL database
		misc::cstring m_db;
		
		/// Connection character set
		misc::cstring m_charset;
		
		/// A shared buffer for escape_string().
		misc::cstring m_es;
		/// Maximum size of the escape_string buffer.
		size_t m_essize;
		
		/// Events notifier.
		Events *m_e;
		
		/// Log queries?
		bool m_logqueries;
		/// Log warnings?
		bool m_logwarnings;
};

/*************************************************************************/

} /* namespace mysql_cli */

/*************************************************************************/

#endif /* MYSQL_CLI_CONN_H */
