/* MySQL connection.
 *
 * MySQL database module (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#include <core/version.h>
#include <modules/core.h>
#include <log/levels.h>

#include "conn.h"
#include "decl.h"
#include "module.h"

using namespace std;

/*************************************************************************/

namespace mysql
{

/*************************************************************************/

void Events::log_event(mysql_cli::Conn &conn, int event, unsigned int code,
		char const *msg) const
{
	switch (event)
	{
		case evConnected:
			if (conn.server_version() < 40102)
				me->core().fatal("mysql: The MySQL server version needs to be "
						"at least 4.1.2");
		
			me->core().log(InfoMySQLConn(),
					"Connected to the MySQL server (%s; version %s; "
					"charset %s) [ID #%lu]",
					conn.host_info(), conn.server_info(), conn.charset(),
					conn.thread_id());
			break;
		case evPing:
			me->core().log(InfoMySQLConn(),
					"Reconnected to the MySQL server (%s; version %s; "
					"charset %s) [ID #%lu]",
					conn.host_info(), conn.server_info(), conn.charset(),
					conn.thread_id());
			break;
		case evClosed:
			me->core().log(InfoMySQLConn(),
					"Connection to the MySQL server closed [ID #%lu]",
					conn.thread_id());
			break;
		case evError:
			if (conn.thread_id())
				log_perror(conn, "%s [ID #%lu]", msg, conn.thread_id());
			else
				log_perror(conn, "%s", msg);
			break;
		case evQuery:
			if (conn.errnum())
				me->core().log(InfoMySQLQueries(), "%s [ID #%lu]",
						msg, conn.thread_id());
			else if (conn.info())
				me->core().log(InfoMySQLQueries(), "%s (%s) (%.3fs) [ID #%lu]",
						msg, conn.info(), conn.exectime(), conn.thread_id());
			else if (conn.field_count())
				me->core().log(InfoMySQLQueries(),
						"%s (Rows: %lu; Warnings: %u) (%.3fs) [ID #%lu]",
						msg, (unsigned long)conn.affected_rows(),
						conn.warning_count(), conn.exectime(),
						conn.thread_id());
			else
				me->core().log(InfoMySQLQueries(),
						"%s (Affected: %lu; Warnings: %u) (%.3fs) [ID #%lu]",
						msg, (unsigned long)conn.affected_rows(),
						conn.warning_count(), conn.exectime(),
						conn.thread_id());
			break;
		case evWarningNote:
			me->core().log(InfoMySQLWarnings(), "%s (note; code %u) [ID #%lu]",
					msg, code, conn.thread_id());
			break;
		case evWarning:
			me->core().log(InfoMySQLWarnings(),
					"%s (warning; code %u) [ID #%lu]", msg, code,
					conn.thread_id());
			break;
		case evWarningError:
			me->core().log(InfoMySQLWarnings(),
					"%s (error; code %u) [ID #%lu]", msg, code,
					conn.thread_id());
			break;
		default:
			break;
	};
}

/*************************************************************************/

void Events::log_perror(mysql_cli::Conn &conn, char const *format, ...) const
{
	char buf[4096], *end = buf;
	
	va_list args;
	va_start(args, format);
	end += vsnprintf(buf, sizeof(buf), format, args);
	va_end(args);
	
	if (conn.error())
		snprintf(end, sizeof(buf)-(end-buf), ": %s", conn.error());
	
	me->core().log(ErrorMySQLConn(), "%s", buf);
}

/*************************************************************************/

} /* namespace mysql */

/*************************************************************************/
