/* Database interface.
 *
 * MySQL database module (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free sofmyare; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Sofmyare Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#include <sstream>

#include <databases/collection.h>
#include <databases/database.h>
#include <databases/field.h>
#include <databases/record.h>
#include <log/levels.h>
#include <misc/misc.h>
#include <misc/string.h>
#include <modules/core.h>
#include <mysql-cli/conn.h>
#include <mysql-cli/result.h>

#include "conf.h"
#include "database.h"
#include "dbinfo.h"
#include "decl.h"
#include "module.h"

using namespace std;
using databases::Collection;
using databases::Record;

/*************************************************************************/

namespace mysql
{

/*************************************************************************/

Database::mycontainer_type Database::g_mylist;

/*************************************************************************/

Database::~Database()
{
	g_mylist.remove(this);
}

/*************************************************************************/

bool Database::delete_(databases::Record const *record)
{
	if (!databases::Database::delete_(record))
		return false;
	
	return real_delete(record);
}

/*************************************************************************/

bool Database::delete_owned(databases::Field const *field,
				databases::Record::identifier_type id)
{
	if (!databases::Database::delete_owned(field, id))
		return false;
	
	/* Do the delete */
	ostringstream q;
	
	q << "DELETE FROM " << tname() << " WHERE " << ((Field *)field)->fname() <<
			" = " << id;
	
	if (!me->conn().query(q.str().c_str(), 0))
	{
		me->core().log(ErrorMySQL(),
				"Can't delete records whose field %s is " RECORD_FORMAT
				" in database %s", field->name(), id, m_name.c_str());
		return false;
	}
	
	return true;
}

/*************************************************************************/

databases::Field *Database::field_create(char const *name, uint16_t minver,
		uint16_t type, size_t size, uint16_t attrs, char const *def,
		uint16_t tag)
{
	Field *field = new Field;
	
	field->name(name);
	field->version(vtMinimum, minver);
	field->type(type);
	field->size(size);
	field->attrs_add(attrs);
	field->default_value(def);
	if (tag != 0)
		field->tag(tag);
	
	return field;
}

/*************************************************************************/

void Database::free_all()
{
	mycontainer_type::iterator it, end;
	
	for (it = g_mylist.begin(), end = g_mylist.end(); it != end; ++it)
		delete (*it);
	
	g_mylist.clear();
}

/*************************************************************************/

void Database::generate_expfields()
{
	/* Generate the internal fields list so that we can use it. */
	databases::Database::generate_expfields();
	
	/* Now! Build our internal import field list. */
	
	ostringstream fields;
	fields << "id";
	
	fields_type::const_iterator it, end;
	
	for (it = m_expfields.begin(), end = m_expfields.end(); it != end; ++it)
		fields << ", `" << (*it)->name() << "`";
	
	m_eflist = fields.str().c_str();
}

/*************************************************************************/

void Database::generate_impfields()
{
	/* Generate the internal fields list so that we can use it. */
	databases::Database::generate_impfields();
	
	/* Now! Build our internal import field list. */
	
	DBInfo *dbi = me->index().find(name().c_str());
	
	/* If the database doesn't exist, there's no fields to import. */
	if (!dbi)
		return;
	
	ostringstream fields;
	fields << "id";
	
	fields_type::const_iterator it, end;
	
	for (it = m_impfields.begin(), end = m_impfields.end(); it != end; ++it)
	{
		if (!dbi->find((*it)->name()))
			continue;
		
		if ((*it)->type() == Field::ftUnixTime)
			fields << ", UNIX_TIMESTAMP(`" << (*it)->name() << "`)";
		else
			fields << ", `" << (*it)->name() << "`";
	}
	
	fields << ", `res_act` + 0";
	
	m_iflist = fields.str().c_str();
}

/*************************************************************************/

bool Database::import(bool initial)
{
	ostringstream q;

#if 0
	/* During initial import... */
	if (initial)
	{
		/* Delete all records marked as DELETE (if allowed) */
		if (perms() & dpDelete)
		{
			/* Delete all records marked as DELETE */
			q << "DELETE FROM " << tname() << " WHERE `res_act` IN (4,5,6,7)";
			
			if (!me->conn().query(q.str().c_str(), 0))
				return false;
			
			q.str("");
		}
	}
#endif
	
	/* Process the records */
	
	mysql_cli::Result res;
	
	for (;;)
	{
		/* First lock the table.
		 * (No need for a lock during initial import -- there's only
		 *  one query afterwards). */
		if (!initial)
		{
			q << "LOCK TABLES " << tname() << " WRITE";
			if (!me->conn().query(q.str().c_str(), 0))
				return false;
			q.str("");
		}
		
		/* Do the SELECT */
		if (!initial)
			q << "SELECT " << m_iflist << " FROM " << tname() <<
					" WHERE `res_act` IN (1,2,3,4,5,6,7) LIMIT 0, " <<
					me->cd().select_max;
		else
			/* We need to ignore records marked CREATE during
			 * initial import because if new rows are inserted
			 * by external programs during initial imports (it
			 * can happen if there are more records than the
			 * limit), some of the non-CREATE rows may get ignored
			 * in the process depending on the order the
			 * rows are returned. */
			q << "SELECT " << m_iflist << " FROM " << tname() <<
					" WHERE `res_act` NOT IN (1,3,5,7) LIMIT 0, " <<
					me->cd().select_max;
		
		if (!me->conn().query(q.str().c_str(), &res))
		{
			if (!initial)
				me->conn().query("UNLOCK TABLES", 0);
			return false;
		}
		q.str("");
		
		/* Now remove the CREATE/UPDATE/DELETE markers
		 * (but not during initial import) */
		if (!initial && res.num_rows())
		{
			q << "UPDATE " << tname() << " SET `res_act` = 0 "
					"WHERE `res_act` IN (1,2,3,4,5,6,7) LIMIT " <<
					me->cd().select_max;
			
			if (!me->conn().query(q.str().c_str(), 0))
			{
				me->conn().query("UNLOCK TABLES", 0);
				return false;
			}
			q.str("");
		}
		
		if (!initial && !me->conn().query("UNLOCK TABLES", 0))
			return false;
		
		if (!initial && (!coll()->begin_import(Collection::csDeletion) ||
				(perms() & dpDelete) ||
				!coll()->end_import(Collection::csDeletion)))
		{
			if (!import_delete(res))
			{
				me->core().log(ErrorMySQL(),
					"Can't delete marked records from database %s",
					m_name.c_str());
				return false;
			}
			
			res.data_seek(0);
		}
		
		if (initial || (perms() & dpCreate))
		{
			if ((!initial && !coll()->begin_import(Collection::csCreation)) ||
					!import_create(res, initial) ||
					(!initial && !coll()->end_import(Collection::csCreation)))
			{
				me->core().log(ErrorMySQL(),
					"Can't import created records from database %s",
					m_name.c_str());
				return false;
			}
			
			res.data_seek(0);
		}
		
		if (!initial && (!coll()->begin_import(Collection::csUpdate) ||
				(perms() & dpUpdate) ||
				!coll()->end_import(Collection::csUpdate)))
		{
			if (!import_update(res))
			{
				me->core().log(ErrorMySQL(),
					"Can't import record updates from database %s",
					m_name.c_str());
				return false;
			}
		}
		
		if (res.num_rows() < me->cd().select_max)
			break;
		q.str("");
	}
	
	return true;
}

/*************************************************************************/

bool Database::import_create(mysql_cli::Result &res, bool initial)
{
	MYSQL_ROW row;
	bool is_error = false;
	
	while ((row = res.fetch_row()))
	{
		if (!initial)
		{
			uint8_t res_act = strtoul(row[res.num_fields() - 1], 0, 10);
		
			if (!(res_act & 1) || (res_act & 4))
				continue;
		}
		
		Record *record;
		
		Record::identifier_type id = Record::strtori(row[0], 0, 10);
		
		if ((record = coll()->find(id)))
		{
			me->core().log(WarningMySQL(),
					"Record " RECORD_FORMAT " of database %s marked "
					"for creation but it already exists, ignoring it",
					id, m_name.c_str());
			continue;
		}
		
		if (!(record = coll()->create()))
		{
			me->core().log(ErrorMySQL(),
					"Can't import new record " RECORD_FORMAT
					" from database %s", id, m_name.c_str());
			if (!real_delete_id(id))
				is_error = true;
			continue;
		}
		
		/* If this returns false, the record has been deleted, so don't
		 * use it afterwards; */
		if (!import_update_record(record, row, is_error))
			continue;
		
		if (!coll()->add(record, 0))
		{
			me->core().log(InfoMySQL(), "Deleting record #" RECORD_FORMAT
					" from database %s (could not be added to collection)",
					record->id(), m_name.c_str());
			
			if (!real_delete(record))
				is_error = true;
			
			delete record;
		}
	}
	
	if (is_error)
	{
		/* No need for an error message if we are during initial loading,
		 * which is going to fail anyway. */
		if (!initial)
			me->core().log(ErrorMySQL(), "There was an error during the "
					"import of records of database %s which prevented "
					"some error corrections to take place; if you "
					"experience differences between data stored in the "
					"MySQL table and the cached data, restart the "
					"program.", m_name.c_str());
		return false;
	}
	
	return true;
}

/*************************************************************************/

bool Database::import_delete(mysql_cli::Result &res)
{	
	MYSQL_ROW row;
	bool is_error = false;
	
	while ((row = res.fetch_row()))
	{
		uint8_t res_act = strtoul(row[res.num_fields() - 1], 0, 10);
		
		if (!(res_act & 4))
			continue;
		
		Record *record;
		Record::identifier_type id = Record::strtori(row[0], 0, 10);
		
		if (!(record = coll()->find(id)))
		{
			me->core().log(ErrorMySQL(),
					"Can't delete non-existing record " RECORD_FORMAT
					" of database %s", id, m_name.c_str());
			if (!real_delete_id(id))
				is_error = true;
			continue;
		}
		
		/* Let Collection::remove remove the record from the database
		 * so we're sure to not remove records when writable() yields
		 * false. */

		uint32_t ro = (me->core().flags() & modules::Core::pfReadOnly);

		me->core().flags_remove(modules::Core::pfReadOnly);
		if (!coll()->remove(record))
		{
			me->core().flags_add(ro);
			return false;
		}
		me->core().flags_add(ro);
	}
	
	if (is_error)
		me->core().log(ErrorMySQL(), "There was an error during the "
				"deletion of records of database %s which prevented "
				"some error corrections to take place; if you "
				"experience differences between data stored in the "
				"MySQL table and the cached data, restart the "
				"program.", m_name.c_str());
	
	return true;
}

/*************************************************************************/

bool Database::import_update(mysql_cli::Result &res)
{
	MYSQL_ROW row;
	bool is_error = false;
	
	while ((row = res.fetch_row()))
	{
		uint8_t res_act = strtoul(row[res.num_fields() - 1], 0, 10);
		
		if (!(res_act & 2) || (res_act & (1 | 4)))
			continue;
		
		Record *record;
		Record::identifier_type id = Record::strtori(row[0], 0, 10);
		
		if (!(record = coll()->find(id)))
		{
			me->core().log(ErrorMySQL(),
					"Can't update non-existing record " RECORD_FORMAT
					" of database %s", id, m_name.c_str());
			if (!real_delete_id(id))
				is_error = true;
			continue;
		}
		
		import_update_record(record, row, is_error);
	}
	
	if (is_error)
	{
		me->core().log(ErrorMySQL(), "There was an error during the "
				"import of records of database %s which prevented "
				"some error corrections to take place; if you "
				"experience differences between data stored in the "
				"MySQL table and the cached data, restart the "
				"program.", m_name.c_str());
		return false;
	}
	
	return true;
}

/*************************************************************************/

bool Database::import_update_record(databases::Record *record, MYSQL_ROW row,
		bool &is_error)
{
	record->begin_update();
	/* The record ID comes first! (it's not updated if it's already set). */
	record->id(Record::strtori(row[0], 0, 10));
	
	int i;
	fields_type::const_iterator it, end;

	for (i = 1, it = m_impfields.begin(), end = m_impfields.end();
			it != end; ++i, ++it)
	{
		Field *field = (Field *)(*it);
		
		switch (field->type())
		{
			case Field::ftString:
			{
				char const *val = (*row[i] ? row[i] : 0);
				char const *rval = record->set_s(field, val);
				
				if ((val && !rval) || (!val && rval) ||
						(rval && val && strcmp(rval, val)))
					if (!real_update_s(record, field, rval))
						is_error = true;
			}
				break;
			case Field::ftInteger:
			{
				int32_t val = strtol(row[i], 0, 10);
				int32_t rval = record->set_i(field, val);
				
				if (val != rval)
					if (!real_update_i(record, field, rval))
						is_error = true;
			}
				break;
			case Field::ftUnsigned:
			{
				uint32_t val = strtoul(row[i], 0, 10);
				uint32_t rval = record->set_ui(field, val);
				
				if (val != rval)
					if (!real_update_ui(record, field, rval))
						is_error = true;
			}
				break;
			case Field::ftUnixTime:
			{
				time_t val = strtol(row[i], 0, 10);
				time_t rval = record->set_ut(field, val);
				
				if (val != rval)
					if (!real_update_ut(record, field, rval))
						is_error = true;
			}
				break;
			case Field::ftIdentifier:
			{
				Record::identifier_type val = Record::strtori(row[i], 0, 10);
				Record::identifier_type rval = record->set_ri(field, val);
				
				if (val != rval)
					if (!real_update_ri(record, field, rval))
						is_error = true;
			}
				break;
			default:
				break;
		}
	}
	
	if (!record->end_update())
	{
		me->core().log(InfoMySQL(), "Deleting bad record #" RECORD_FORMAT
				" from database %s", record->id(), m_name.c_str());
		
		/* Make sure writable() doesn't return false because of read-only
		 * stuff. */
		if (record->db_status() & Record::dbsMember)
		{
			uint32_t ro = (me->core().flags() & modules::Core::pfReadOnly);
			me->core().flags_remove(modules::Core::pfReadOnly);
		
			if (!coll()->remove(record))
				me->core().fatalf("Can't remove bad record #" RECORD_FORMAT
						" from database %s", record->id(), m_name.c_str());
			
			me->core().flags_add(ro);
		}
		else
		{
			if (!real_delete(record))
				is_error = true;	
			delete record;
		}
		
		return false;
	}
	
	return true;
}

/*************************************************************************/

bool Database::initial_import()
{	
	/* If the database doesn't exist, don't try to import it */
	DBInfo *dbi = me->index().find(m_name.c_str());
	
	if (dbi && !readable())
		return false;
	
	if (dbi)
	{
		/* Set the import version if the database exists (by default, == 0). */
		version(vtImport, dbi->version());
		/* Generate the list of imported fields. */
		generate_impfields();
	}
	
	/* Check if we need to DO initial import, based on version numbers. */
	if (version(vtImport) >= version(vtMinimum) &&
			(!version(vtMaximum) || version(vtImport) <= version(vtMaximum)))
	{
		/* Say to the collection initial import is started. */
		if (!coll()->begin_initial())
		{
			me->core().log(CriticalMySQL(),
					"Aborting initial import (initialization failed).");
			return false;
		}

		if (dbi && (!readable() || !import(true)))
			return false;
		
		/* Say to the collection initial import is finished. */
		if (!coll()->end_initial())
		{
			me->core().log(CriticalMySQL(),
					"Aborting initial import (finalization failed).");
			return false;
		}
	}
	
	/* Woohoo! Initial import is done, set the corresponding status flag
	 * (even if the version number didn't match). */
	m_status |= dsInitial;
	
	if (me->core().debuglevel())
		me->core().log(DebugMySQL(), "Loaded database %s", m_name.c_str());
	
	return true;
}

/*************************************************************************/

bool Database::insert(databases::Record *record)
{
	if (!databases::Database::insert(record))
		return false;
	
	ostringstream q;
	
	q << "INSERT INTO " << tname() << " (" << m_eflist << ") VALUES(" <<
			record->id();
	
	fields_type::const_iterator it, end;
	
	for (it = m_expfields.begin(), end = m_expfields.end(); it != end; ++it)
	{
		q << ", ";
		
		Field *field = (Field *)(*it);
		
		switch (field->type())
		{
			case Field::ftString:
			{
				char const *val = record->get_s(field);
				q << "'" << me->conn().escape_string((val ? val : "")) << "'";
			}
				break;
			case Field::ftInteger:
				q << record->get_i(field);
				break;
			case Field::ftUnsigned:
				q << record->get_ui(field);
				break;
			case Field::ftUnixTime:
				q << "FROM_UNIXTIME(" << record->get_ut(field) << ")";
				break;
			case Field::ftIdentifier:
				q << record->get_ri(field);
				break;
			default:
				break;
		}
	}
	
	q << ')';
	
	if (!me->conn().query(q.str().c_str(), 0))
	{
		me->core().log(ErrorMySQL(),
				"Can't insert record " RECORD_FORMAT " in database %s",
						record->id(), m_name.c_str());
		return false;
	}
	
	if (record->id() == 0)
		record->id(me->conn().insert_id());
	
	return true;
}

/*************************************************************************/

bool Database::readable() const
{
	return (databases::Database::readable() && !(me->conn().gone()));
}

/*************************************************************************/

bool Database::real_delete(databases::Record const *record)
{
	return real_delete_id(record->id());
}

/*************************************************************************/

bool Database::real_delete_id(databases::Record::identifier_type id)
{
	/* Don't update if dpExport is not set, but still return true. */
	if (!(perms() & dpExport))
		return true;
	
	/* If it's gone though, return an error. */
	if (me->conn().gone())
		return false;
	
	/* Do the delete */
	ostringstream q;
	
	q << "DELETE FROM " << tname() << " WHERE `id` = " << id;
	
	if (!me->conn().query(q.str().c_str(), 0))
	{
		me->core().log(ErrorMySQL(),
				"Can't delete record " RECORD_FORMAT " in database %s",
				id, m_name.c_str());
		return false;
	}
	
	return true;
}

/*************************************************************************/

bool Database::real_truncate()
{
	/* Don't update if dpExport is not set, but still return true. */
	if (!(perms() & dpExport))
		return true;
	
	/* If it's gone though, return an error. */
	if (me->conn().gone())
		return false;
	
	/* Do the truncate */
	ostringstream q;
	
	q << "DELETE FROM " << tname();
	
	if (!me->conn().query(q.str().c_str(), 0))
	{
		me->core().log(ErrorMySQL(),
				"Can't clear all records from database %s",
				m_name.c_str());
		return false;
	}
	
	return true;
}

/*************************************************************************/

bool Database::real_update_i(databases::Record const *record,
		Field const *field, int32_t value)
{
	/* Don't update if dpExport is not set, but still return true. */
	if (!(perms() & dpExport))
		return true;
	
	/* If it's gone though, return an error. */
	if (me->conn().gone())
		return false;
	
	/* Do the update */
	ostringstream q;
	
	q << "UPDATE " << tname() << " SET " << field->fname() << " = " <<
			value << " WHERE `id` = " << record->id();
	
	if (!me->conn().query(q.str().c_str(), 0))
	{
		me->core().log(ErrorMySQL(),
				"Can't update field %s of record " RECORD_FORMAT
				" in database %s", field->name(), record->id(),
				m_name.c_str());
		return false;
	}
	
	return true;
}

/*************************************************************************/

bool Database::real_update_ri(databases::Record const *record,
		Field const *field, Record::identifier_type value)
{
	/* Don't update if dpExport is not set, but still return true. */
	if (!(perms() & dpExport))
		return true;
	
	/* If it's gone though, return an error. */
	if (me->conn().gone())
		return false;
	
	/* Do the update */
	ostringstream q;
	
	q << "UPDATE " << tname() << " SET " << field->fname() << " = " <<
			value << " WHERE `id` = " << record->id();
	
	if (!me->conn().query(q.str().c_str(), 0))
	{
		me->core().log(ErrorMySQL(),
				"Can't update field %s of record " RECORD_FORMAT
				" in database %s", field->name(), record->id(),
				m_name.c_str());
		return false;
	}
	
	return true;
}

/*************************************************************************/

bool Database::real_update_s(databases::Record const *record,
		Field const *field, char const *value)
{
	/* Don't update if dpExport is not set, but still return true. */
	if (!(perms() & dpExport))
		return true;
	
	/* If it's gone though, return an error. */
	if (me->conn().gone())
		return false;
	
	/* Do the update */
	ostringstream q;
	
	q << "UPDATE " << tname() << " SET " << field->fname() << " = '" <<
			me->conn().escape_string((value ? value : "")) <<
			"' WHERE `id` = " << record->id();
	
	if (!me->conn().query(q.str().c_str(), 0))
	{
		me->core().log(ErrorMySQL(),
				"Can't update field %s of record " RECORD_FORMAT
				" in database %s", field->name(), record->id(),
				m_name.c_str());
		return false;
	}
	
	return true;
}

/*************************************************************************/

bool Database::real_update_ui(databases::Record const *record,
		Field const *field, uint32_t value)
{
	/* Don't update if dpExport is not set, but still return true. */
	if (!(perms() & dpExport))
		return true;
	
	/* If it's gone though, return an error. */
	if (me->conn().gone())
		return false;
	
	/* Do the update */
	ostringstream q;
	
	q << "UPDATE " << tname() << " SET " << field->fname() << " = " <<
			value << " WHERE `id` = " << record->id();
	
	if (!me->conn().query(q.str().c_str(), 0))
	{
		me->core().log(ErrorMySQL(),
				"Can't update field %s of record " RECORD_FORMAT
				" in database %s", field->name(), record->id(),
				m_name.c_str());
		return false;
	}
	
	return true;
}

/*************************************************************************/

bool Database::real_update_ut(databases::Record const *record,
		Field const *field, time_t value)
{
	/* Don't update if dpExport is not set, but still return true. */
	if (!(perms() & dpExport))
		return true;
	
	/* If it's gone though, return an error. */
	if (me->conn().gone())
		return false;
	
	/* Do the update */
	ostringstream q;
	
	q << "UPDATE " << tname() << " SET " << field->fname() <<
			" = FROM_UNIXTIME(" << value << ") WHERE `id` = " << record->id();
	
	if (!me->conn().query(q.str().c_str(), 0))
	{
		me->core().log(ErrorMySQL(),
				"Can't update field %s of record " RECORD_FORMAT
				" in database %s", field->name(), record->id(),
				m_name.c_str());
		return false;
	}
	
	return true;
}

/*************************************************************************/

bool Database::reg()
{
	g_mylist.push_back(this);
	
	return databases::Database::reg();
}

/*************************************************************************/

modules::Interface *Database::supports(modules::PSIID sid)
{
	if (sid == Database::sIDGen)
		return this;
	else
		return databases::Database::supports(sid);
}

/*************************************************************************/

bool Database::start_real()
{
	/* Set the table name */
	
	m_tname.reserve(1024);
	snprintf(m_tname, 1024, "`%s%s`",
			me->cd().table_prefix.value().c_str(), name().c_str());
	
	/* Do initial import if requested. */
	if ((perms() & dpInitial) && !initial_import())
	{
		me->core().log(CriticalMySQL(),
				"Loading of database %s failed", m_name.c_str());
		return false;
	}
	
	/* Create/update the table */
	if ((perms() & dpExport) && !update_table())
		return false;
	
	/* Generate the list of imported fields again now that initial
	 * loading and table update are done. */
	generate_impfields();
	
	/* Truncate the table if dpInitial is not set but dpExport is. */
	if (!(perms() & dpInitial) && (perms() & dpExport) && !real_truncate())
		return false;
	
	if ((perms() & dpExport) && writable() && !coll()->update_new())
	{
		me->core().log(CriticalMySQL(),
				"Updating new fields of database %s failed", m_name.c_str());
		return false;
	}
	
	return true;
}

/*************************************************************************/

void Database::sync()
{
	if (me->core().debuglevel())
		me->core().log(DebugMySQL(), "Syncing database %s",
				m_name.c_str());
	
	if ((perms() & dpImport) && readable() && !import(false))
		me->core().log(ErrorMySQL(), 
				"Can't import changes from database %s", m_name.c_str());
}

/*************************************************************************/

void Database::sync_all()
{
	mycontainer_type::iterator it, end;
	
	for (it = g_mylist.begin(), end = g_mylist.end(); it != end; ++it)
	{
		bool found = false;
		conf::ConfStrCharA::container_type::const_iterator bit, bend;
		
		for (bit = me->cd().sync.blacklist.value().begin(),
				bend = me->cd().sync.blacklist.value().end();
				bit != bend; ++bit)
		{
			if (misc::match(bit->value().c_str(), (*it)->name().c_str(), true))
				found = true;
		}
		
		if (found)
		{
			if (me->cd().sync.mem_usage)
				(*it)->update_memusage();
			continue;
		}
		
		(*it)->sync();
		
		if (me->cd().sync.mem_usage)
			(*it)->update_memusage();
	}
}

/*************************************************************************/

bool Database::truncate()
{
	if (!databases::Database::truncate())
		return false;
	
	return real_truncate();
}

/*************************************************************************/

bool Database::update_i(databases::Record const *record,
		databases::Field const *field, int32_t value)
{
	if (!databases::Database::update_i(record, field, value))
		return false;
	
	return real_update_i(record, (Field *)field, value);
}

/*************************************************************************/

void Database::update_memusage()
{	
	if (coll()->status() & Collection::csChanges)
	{
		DBInfo *dbi = me->index().find(m_name.c_str());
	
		if (!dbi)
			return;
		
		dbi->mem(coll()->memory_usage());
		coll()->status_remove(Collection::csChanges);
	}
}

/*************************************************************************/

bool Database::update_multi(databases::Record const *record,
		unsigned int valcount, va_list args)
{
	/* Ensure everything's OK */
	if (!databases::Database::update_multi(record, valcount, args))
		return false;
	
	/* Do the update */
	ostringstream q;
	unsigned int count = 0;
	
	q << "UPDATE " << tname() << " SET ";
	
	for (unsigned int i = 0; i < valcount; ++i)
	{
		Field *field = (Field *)va_arg(args, databases::Field *);
		
		if (!field_find(field->name(), Database::vtExport))
			continue;
		
		char const *s = 0;
		char buf[64];
		
		switch (field->type())
		{
			case Field::ftString:
				s = va_arg(args, char *);
				break;
			case Field::ftInteger:
				snprintf(buf, sizeof(buf), "%d", va_arg(args, int32_t));
				s = buf;
				break;
			case Field::ftUnsigned:
				snprintf(buf, sizeof(buf), "%u", va_arg(args, uint32_t));
				s = buf;
				break;
			case Field::ftUnixTime:
				snprintf(buf, sizeof(buf), "FROM_UNIXTIME(%ld)",
						va_arg(args, time_t));
				s = buf;
				break;
			case Field::ftIdentifier:
				snprintf(buf, sizeof(buf), RECORD_FORMAT, va_arg(args,
						databases::Record::identifier_type));
				s = buf;
				break;
			default:
				return false;
		}
		
		++count;
		
		if (count > 1)
			q << ", ";
		
		if (field->type() == Field::ftString)
			q << field->fname() << " = '" <<
					me->conn().escape_string((s ? s : "")) << "'";
		else
			q << field->fname() << " = " << s;
	}
	
	/* If no valid fields... */
	if (count == 0)
		return true;
	
	 q << " WHERE `id` = " << record->id();
	
	if (!me->conn().query(q.str().c_str(), 0))
	{
		me->core().log(ErrorMySQL(),
				"Can't update multiple fields of record " RECORD_FORMAT
				" in database %s", record->id(), m_name.c_str());
		return false;
	}
	
	return true;
}

/*************************************************************************/

bool Database::update_owned(databases::Field const *field,
				databases::Record::identifier_type id,
				databases::Record::identifier_type new_id)
{
	if (!databases::Database::update_owned(field, id, new_id))
		return false;
	
	/* Do the update */
	ostringstream q;
	
	q << "UPDATE " << tname() << " SET " << ((Field *)field)->fname() <<
		" = " << new_id << " WHERE " << ((Field *)field)->fname() <<
			" = " << id;
	
	if (!me->conn().query(q.str().c_str(), 0))
	{
		me->core().log(ErrorMySQL(),
				"Can't update records whose field %s is " RECORD_FORMAT
				" in database %s", field->name(), id, m_name.c_str());
		return false;
	}
	
	return true;
}

/*************************************************************************/

bool Database::update_ri(databases::Record const *record,
		databases::Field const *field, Record::identifier_type value)
{
	if (!databases::Database::update_ri(record, field, value))
		return false;
	
	return real_update_ri(record, (Field *)field, value);
}

/*************************************************************************/

bool Database::update_s(databases::Record const *record,
		databases::Field const *field, char const *value)
{
	if (!databases::Database::update_s(record, field, value))
		return false;
	
	return real_update_s(record, (Field *)field, value);
}

/*************************************************************************/

bool Database::update_table()
{
	if (me->conn().gone())
		return false;
	
	bool existed = true;
	DBInfo *dbi = me->index().find(m_name.c_str());
	
	if (!dbi)
	{
		dbi = new DBInfo(&me->conn(), 0, DBInfo::current_meta, m_name.c_str(),
				version(vtExport), 0);
		existed = false;
	}
	
	fields_type::const_iterator it, end;
	
	for (it = m_expfields.begin(), end = m_expfields.end(); it != end; ++it)
	{
		databases::Field *field = (*it);
		FieldInfo *fi = dbi->find(field->name());
		
		if (!fi)
		{
			/* The field is brand new */
			fi = new FieldInfo(&me->conn(), 0, field->name(), field->type(),
					field->size(), field->attrs(), field->default_value());
			dbi->add(fi);
		} else {
			/* Check if the field data match */
			if (fi->type() != field->type() || fi->size() != field->size() ||
					fi->attrs() != field->attrs())
			{
				me->core().log(ErrorMySQL(), "Field %s (#%u) info mismatch "
						"(has: type=%u, size=%u, attrs=%u; wants: type=%u; "
						"size=%u; attrs=%u)", fi->name(), fi->id(), fi->type(),
						fi->size(), fi->attrs(), field->type(), field->size(),
						field->attrs());
				return false;
			}
		}
	}
	
	if (!dbi->update_table())
		return false;
	
	if (!existed)
		me->index().add(dbi);
	else if (!dbi->version(version(vtExport)))
	{
		me->core().log(ErrorMySQL(),
				"Can't update version of database %s", dbi->name());
		return false;
	}
	
	return true;
}

/*************************************************************************/

bool Database::update_ui(databases::Record const *record,
		databases::Field const *field, uint32_t value)
{
	if (!databases::Database::update_ui(record, field, value))
		return false;
	
	return real_update_ui(record, (Field *)field, value);
}

/*************************************************************************/

bool Database::update_ut(databases::Record const *record,
		databases::Field const *field, time_t value)
{
	if (!databases::Database::update_ut(record, field, value))
		return false;
	
	return real_update_ut(record, (Field *)field, value);
}

/*************************************************************************/

bool Database::writable() const
{
	return (databases::Database::writable() && !(me->conn().gone()));
}

/*************************************************************************/

bool DBRemove::clear()
{
	if (!me->index().clear())
	{
		errorf("Removal of databases from the index failed");
		return false;
	}
	
	return true;
}

/*************************************************************************/

bool DBRemove::remove(std::string const &database)
{
	DBInfo *dbi;
	
	if (!(dbi = me->index().find(database.c_str())))
	{
		errorf("Database %s does not exist", database.c_str());
		return false;
	}
	
	if (!me->index().remove(dbi))
	{
		errorf("Removal of the database from the index failed");
		return false;
	}
	
	me->core().log(InfoMySQL(), "Removed database %s", database.c_str());
	return true;
}

/*************************************************************************/

bool Field::setup()
{
	if (!databases::Field::setup())
		return false;
	
	m_fname.reserve(1024);
	snprintf(m_fname, 1024, "`%s`", name());

	return true;
}

/*************************************************************************/

} /* namespace mysql */

/*************************************************************************/
