/* MySQL connection.
 *
 * MySQL database module (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#include <sstream>

#include <databases/field.h>
#include <databases/record.h>
#include <databases/xml.h>
#include <log/levels.h>
#include <misc/misc.h>
#include <misc/string.h>
#include <mysql-cli/conn.h>
#include <mysql-cli/result.h>
#include <modules/core.h>

#include "conf.h"
#include "dbinfo.h"
#include "decl.h"
#include "module.h"
#include "xml.h"

using namespace std;
using databases::Field;
using databases::Record;
using databases::XMLReader;
using databases::XMLWriter;

/*************************************************************************/

namespace mysql
{

/*************************************************************************/

class XICallback : public XMLReader::Callback
{
	public:
		/// Constructor.
		XICallback(char const *name) : XMLReader::Callback(), m_name(name),
				m_dbi(0), m_ins(), m_needcomma(false), m_curid(0) { }
	
		/// Overrides XMLReader::Callback::database().
		virtual bool database(uint16_t version);
		
		/// Overrides XMLReader::Callback::add_field().
		virtual bool add_field(char const *name, uint16_t type,
				size_t size, uint16_t attrs, char const *defval);
		
		/// Overrides XMLReader::Callback::count().
		virtual bool count(uint32_t count);
		
		/// Overrides XMLReader::Callback::begin_record().
		virtual bool begin_record();
		
		/// Overrides XMLReader::Callback::add_value().
		virtual bool add_value(char const *field,
				char const *value);
		
		/// Overrides XMLReader::Callback::add_ivalue().
		virtual bool add_ivalue(char const *field, int32_t value);
		/// Overrides XMLReader::Callback::add_uivalue().
		virtual bool add_uivalue(char const *field,
				uint32_t value);
		/// Overrides XMLReader::Callback::add_utvalue().
		virtual bool add_utvalue(char const *field, time_t value);
		/// Overrides XMLReader::Callback::add_rivalue().
		virtual bool add_rivalue(char const *field,
				Record::identifier_type value);
		
		/// Overrides XMLReader::Callback::end_record().
		virtual bool end_record();
		
		/// Returns the database info.
		DBInfo *dbinfo() { return m_dbi; }
	private:
		misc::cstring m_name;
		DBInfo *m_dbi;
		ostringstream m_ins;
		bool m_needcomma;
		Record::identifier_type m_curid;
		
		XICallback(XICallback const &);
		XICallback &operator=(XICallback const &);
};

/*************************************************************************/

bool XICallback::database(uint16_t version)
{
	m_dbi = new DBInfo(&me->conn(), 0, DBInfo::current_meta, m_name, version, 0);
	return true;
}

/*************************************************************************/

bool XICallback::add_field(char const *name, uint16_t type,
		size_t size, uint16_t attrs, char const *defval)
{
	FieldInfo *fi = new FieldInfo(&me->conn(), 0, name, type,
			size, attrs, defval);
	m_dbi->add(fi);
	
	return true;
}

/*************************************************************************/

bool XICallback::count(uint32_t count)
{
	return m_dbi->update_table();
}

/*************************************************************************/

bool XICallback::begin_record()
{
	m_ins << "INSERT INTO `" << me->cd().table_prefix.value() << m_name <<
			"` SET ";
	
	return true;
}

/*************************************************************************/

bool XICallback::add_value(char const *field,
		char const *value)
{
	if (m_needcomma)
		m_ins << ", ";
	
	m_ins << "`" << field << "` = '" <<
			me->conn().escape_string((value ? value : "")) << "'";
	return (m_needcomma = true);
}

/*************************************************************************/

bool XICallback::add_ivalue(char const *field, int32_t value)
{
	if (m_needcomma)
		m_ins << ", ";
	
	m_ins << "`" << field << "` = " << value;
	return (m_needcomma = true);
}

/*************************************************************************/

bool XICallback::add_uivalue(char const *field,
		uint32_t value)
{
	if (m_needcomma)
		m_ins << ", ";
	
	m_ins << "`" << field << "` = " << value;
	return (m_needcomma = true);
}

/*************************************************************************/

bool XICallback::add_utvalue(char const *field, time_t value)
{
	if (m_needcomma)
		m_ins << ", ";
	
	m_ins << "`" << field << "` = FROM_UNIXTIME(" << value << ")";
	return (m_needcomma = true);
}

/*************************************************************************/

bool XICallback::add_rivalue(char const *field,
		Record::identifier_type value)
{
	if (m_needcomma)
		m_ins << ", ";
	
	m_ins << "`" << field << "` = " << value;
	
	if (!strcmp(field, "id"))
		m_curid = value;
	
	return (m_needcomma = true);
}

/*************************************************************************/

bool XICallback::end_record()
{
	if (!me->conn().query(m_ins.str().c_str(), 0))
	{
		me->core().log(ErrorMySQL(), "Can't insert record " RECORD_FORMAT 
				" in database %s", m_curid, (char const *)m_name);
		return false;
	}
	
	m_ins.str("");
	m_needcomma = false;
	m_curid = 0;
	
	return true;
}

/*************************************************************************/

bool XMLExport::export_data(std::string const &dir, databases::XMLIndex &index,
		std::string const &database)
{
	DBIndex::container_type::const_iterator it, end;
	
	for (it = me->index().dbs().begin(), end = me->index().dbs().end();
			it != end; ++it)
	{
		DBInfo *dbi = (*it);
		
		if (!database.empty() && (database != dbi->name()))
			continue;
		
		XMLWriter xw;
		
		if (!xw.database(dbi->version()))
		{
			errorf("XML write error: %s", xw.error());
			return false;
		}
		
		ostringstream fields;
		fields << "`id`";
		
		DBInfo::container_type::const_iterator fit, fend;
		
		for (fit = dbi->fields().begin(), fend = dbi->fields().end();
				fit != fend; ++fit)
		{	
			FieldInfo *fi = (*fit);
			
			if (!xw.add_field(fi->name(), fi->type(), fi->size(),
					fi->attrs(), fi->defval()))
			{
				errorf("XML write error: %s", xw.error());
				return false;
			}
			
			if (fi->type() != Field::ftUnixTime)
				fields << ", `" << fi->name() << "`";
			else
				fields << ", UNIX_TIMESTAMP(`" << fi->name() << "`)";
		}
		
		mysql_cli::Result res;
		ostringstream q;
		
		q << "SELECT COUNT(*) FROM `" << me->cd().table_prefix.value() <<
				dbi->name() << "`";
		
		if (!me->conn().query(q.str().c_str(), &res))
		{
			errorf("Failed to get record count for database %s",
					dbi->name());
			return false;
		}
		
		MYSQL_ROW row = res.fetch_row();
		
		if (!xw.count(strtoul(row[0], 0, 10)))
		{
			errorf("XML write error: %s", xw.error());
			return false;
		}
		
		q.str("");
		
		for (uint32_t offset = 0; ; offset += me->cd().select_max)
		{
			q << "SELECT " << fields.str() << " FROM `" <<
					me->cd().table_prefix.value() << dbi->name() <<
					"` ORDER BY `id` LIMIT " << offset << "," <<
					me->cd().select_max;
			
			if (!me->conn().query(q.str().c_str(), &res))
			{
				errorf("Failed to fetch records for database %s",
						dbi->name());
				return false;
			}
			
			while ((row = res.fetch_row()))
			{
				if (!xw.begin_record())
				{
					errorf("XML write error: %s", xw.error());
					return false;
				}
			
				if (!xw.add_rivalue("id", Record::strtori(row[0], 0, 10)))
				{
					errorf("XML write error: %s", xw.error());
					return false;
				}
				
				int i;
				
				for (i = 1, fit = dbi->fields().begin(),
						fend = dbi->fields().end(); fit != fend; ++i, ++fit)
				{
					FieldInfo *fi = (*fit);
					
					switch (fi->type())
					{
						case Field::ftString:
							if (!xw.add_value(fi->name(),
									(*row[i] ? row[i] : 0)))
							{
								errorf("XML write error: %s", xw.error());
								return false;
							}
							break;
						case Field::ftInteger:
						case Field::ftUnsigned:
						case Field::ftUnixTime:
						case Field::ftIdentifier:
							/* No need to use the non-string versions of
							 * addvalue... */
							if (!xw.add_value(fi->name(), row[i]))
							{
								errorf("XML write error: %s", xw.error());
								return false;
							}
							break;
						default:
							errorf("Bad type %u for field %s of database %s "
								"(corrupted index?)", fi->type(), fi->name(),
								dbi->name());
							break;
					}
				}
			
				if (!xw.end_record())
				{
					errorf("XML write error: %s", xw.error());
					return false;
				}
			}
			
			if (res.num_rows() < me->cd().select_max)
				break;
			
			q.str("");
		}
			
		if (!xw.save(misc::absolute_path(dir,
				std::string(dbi->name()) + ".xml")))
		{
			errorf("XML save error: %s", xw.error());
			return false;
		}
		
		if (!index.add(std::string(dbi->name()) + ".xml"))
		{
			errorf("Could not add database %s to index file", dbi->name());
			return false;
		}
		
		me->core().log(InfoMySQL(), "Exported database %s", dbi->name());
	}
	
	return true;
}

/*************************************************************************/

bool XMLImport::import_data(std::string const &dir,
		std::string const &database)
{
	if (me->index().exists(database.c_str()))
	{
		errorf("Database %s already exists", database.c_str());
		return false;
	}
	
	XMLReader xr;
	XICallback xic(database.c_str());
	
	if (!xr.load(misc::absolute_path(dir, database + ".xml"), xic))
	{
		if (*xr.error())
			errorf("XML loading error: %s", xr.error());
		else
			errorf("Write error");
		return false;
	}
	
	me->index().add(xic.dbinfo());
	
	me->core().log(InfoMySQL(), "Imported database %s", database.c_str());
	return true;
}

/*************************************************************************/

} /* namespace mysql */

/*************************************************************************/
