/* Database file.
 *
 * Twilight database module (c) 2005 PegSoft
 * Contact us at pegsoft@pegsoft.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see the file COPYING); if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <system/headers.h>

#include <databases/field.h>
#include <misc/misc.h>
#include <modules/core.h>

#include "conf.h"
#include "decl.h"
#include "dbfile.h"
#include "module.h"

using namespace std;

/*************************************************************************/

namespace twilight
{

/*************************************************************************/

DBFile::DBFile(std::string const &filename) : m_filename(filename),
		m_absfilename(misc::absolute_path(me->cd().directory, filename)),
		m_handle(0)
{
	
}

/*************************************************************************/

DBFile::~DBFile()
{
	close();
}

/*************************************************************************/

void DBFile::close()
{
	if (m_handle)
	{
		while ((fclose(m_handle) == EOF) && (errno == EINTR))
			;
		m_handle = 0;
	}
}

/*************************************************************************/

void DBFile::error(char const *msg, char const *filename) const
{
	me->core().log_perror(ErrorTwilightDB(), msg, filename);
}

/*************************************************************************/

bool DBReader::open()
{
	/* Just a little sanity check. :P */
	if (m_handle)
		return false;
	
	if (!(m_handle = fopen(m_absfilename.c_str(), "rb")))
	{
		error("Can't open %s for reading", m_absfilename.c_str());
		return false;
	}
	
	return true;
}

/*************************************************************************/

bool DBReader::read_count(uint32_t *count)
{
	return read_unsigned(count);
}

/*************************************************************************/

bool DBReader::read_dbinfo(uint16_t *version)
{
	return read_unsigned(version);
}

/*************************************************************************/

bool DBReader::read_fieldinfo(uint16_t *fieldid, char const **name,
				uint16_t *type, uint32_t *size, uint16_t *attrs,
				char const **defval)
{
	if (!read_unsigned(fieldid) || !read_string(name))
		return false;
	
	m_fieldname = *name;
	*name = m_fieldname;
	
	return (read_unsigned(type) && read_unsigned(size) &&
			read_unsigned(attrs) && read_string(defval));
}

/*************************************************************************/

bool DBReader::read_fieldinfo(uint16_t *fieldid, char const **name)
{
	uint16_t type, attrs;
	uint32_t size;
	char const *defval;
	
	return read_fieldinfo(fieldid, name, &type, &size, &attrs, &defval);
}

/*************************************************************************/

bool DBReader::read_id(databases::Record::identifier_type *id)
{
	*id = 0;
	
	/* Timestamps are stored on 64-bit */
	
	for (unsigned int i = 0; i < (8 - sizeof(id)); ++i)
	{
		if (fgetc(m_handle) == EOF)
		{
			error("Can't read from %s (corrupted file?)", m_absfilename.c_str());
			return false;
		}
	}
	
	for (int i = (sizeof(id) - 1); i >= 0; --i)
	{
		int c;
		
		if ((c = fgetc(m_handle)) == EOF)
		{
			error("Can't read from %s (corrupted file?)", m_absfilename.c_str());
			return false;
		}
		
		*id <<= 8;
		*id |= (c & 0xFF);
	}
	
	return true;
}

/*************************************************************************/

bool DBReader::read_string(char const **str)
{
	uint32_t len = 0;
	
	if (!read_unsigned(&len))
		return false;
	
	if (!len)
	{
		*str = 0;
		return true;
	}
	
	reserve(len + 1);
	m_str[len] = '\0';
	
	uint32_t read;
	
	if ((read = fread(m_str, 1, len, m_handle)) != len)
	{
		error("Can't read from %s (corrupted file?)", m_absfilename.c_str());
		return false;
	}
	
	*str = m_str;
	return true;
}

/*************************************************************************/

bool DBReader::read_time(time_t *t)
{
	*t = 0;
	
	/* Timestamps are stored on 64-bit */
	
	for (unsigned int i = 0; i < (8 - sizeof(t)); ++i)
	{
		if (fgetc(m_handle) == EOF)
		{
			error("Can't read from %s (corrupted file?)",
					m_absfilename.c_str());
			return false;
		}
	}
	
	for (int i = (sizeof(t) - 1); i >= 0; --i)
	{
		int c;
		
		if ((c = fgetc(m_handle)) == EOF)
		{
			error("Can't read from %s (corrupted file?)",
					m_absfilename.c_str());
			return false;
		}
		
		*t <<= 8;
		*t |= (c & 0xFF);
	}
	
	return true;
}

/*************************************************************************/

bool DBReader::read_valueinfo(uint16_t *fieldid, uint16_t *type)
{
	return (read_unsigned(fieldid) && read_unsigned(type));
}

/*************************************************************************/

bool DBReader::read_version()
{
	uint16_t ver;
	
	if (!read_unsigned(&ver))
		return false;
	
	if (ver > dbfile_version)
	{
		error("Bad version in file %s", m_absfilename.c_str());
		return false;
	}
	
	return true;
}

/*************************************************************************/

void DBReader::reserve(size_t len)
{
	if (m_maxlen >= len)
		return;
	
	m_str.reserve((m_maxlen = len));
}

/*************************************************************************/

bool DBWriter::cleanup()
{
	if (remove(m_oldafn.c_str()) == -1 && errno != ENOENT)
	{
		error("Can't remove %s", m_oldafn.c_str());
		return false;
	}
	
	return true;
}

/*************************************************************************/

bool DBWriter::commit()
{
	/* Delete any existing old file. */
	if (remove(m_oldafn.c_str()) == -1 && errno != ENOENT)
	{
		error("Can't remove %s", m_oldafn.c_str());
		return false;
	}
	
	/* Try to rename the main file to the old file. */
	if ((rename(m_absfilename.c_str(), m_oldafn.c_str()) == -1) &&
			(errno != ENOENT))
	{
		error("Can't rename %s", m_absfilename.c_str());
		return false;
	}
	
	/* Close the file */
	close();
	
	if (rename(m_newafn.c_str(), m_absfilename.c_str()) == -1)
	{
		error("Can't rename %s", m_newafn.c_str());
		return false;
	}
	
	return true;
}

/*************************************************************************/

bool DBWriter::open()
{
	/* Just a little sanity check. :P */
	if (m_handle)
		return false;
	
	if (!(m_handle = fopen(m_newafn.c_str(), "wb")))
	{
		error("Can't open %s for writing", m_newafn.c_str());
		return false;
	}
	
	return true;
}

/*************************************************************************/

bool DBWriter::revert()
{
	if (remove(m_absfilename.c_str()) == -1)
	{
		error("Can't remove %s", m_absfilename.c_str());
		return false;
	}
	
	if (rename(m_oldafn.c_str(), m_absfilename.c_str()) == -1)
	{
		error("Can't rename %s", m_oldafn.c_str());
		return false;
	}
		
	return true;
}

/*************************************************************************/

bool DBWriter::write_count(uint32_t count)
{
	return write_unsigned(count);
}

/*************************************************************************/

bool DBWriter::write_dbinfo(uint16_t version)
{
	return write_unsigned(version);
}

/*************************************************************************/

bool DBWriter::write_endfields()
{
	return (write_fieldinfo(end_fieldid, "res_last", end_type, 0, 0, ""));
}

/*************************************************************************/

bool DBWriter::write_endrecord()
{
	return (write_valueinfo(end_fieldid, end_type));
}

/*************************************************************************/

bool DBWriter::write_fieldinfo(uint16_t fieldid, char const *name,
		uint16_t type, uint32_t size, uint16_t attrs, char const *defval)
{
	return (write_unsigned(fieldid) && write_string(name) &&
			write_unsigned(type) && write_unsigned(size) &&
			write_unsigned(attrs) && write_string(defval));
}

/*************************************************************************/

bool DBWriter::write_fieldinfo(uint16_t fieldid, databases::Field const *field)
{
	return write_fieldinfo(fieldid, field->name(), field->type(),
			field->size(), field->attrs(), field->default_value());
}

/*************************************************************************/

bool DBWriter::write_id(databases::Record::identifier_type id)
{
	/* Store the identifier on 64-bit, in case we need to make it greater
	 * than 32-bit one day (you never know :P) */
	
	for (unsigned int i = 0; i < (8 - sizeof(id)); ++i)
	{
		if (fputc(0, m_handle) == EOF)
		{
			error("Can't write to %s", m_newafn.c_str());
			return false;
		}
	}
	
	for (int i = (sizeof(id) - 1); i >= 0; --i)
	{
		if (fputc(((id >> (i * 8)) & 0xFF), m_handle) == EOF)
		{
			error("Can't write to %s", m_newafn.c_str());
			return false;
		}
	}
	
	return true;
}

/*************************************************************************/

bool DBWriter::write_integer(int32_t n)
{	
	for (int i = 3; i >= 0; --i)
	{
		if (fputc(((n >> (i * 8)) & 0xFF), m_handle) == EOF)
		{
			error("Can't write to %s", m_newafn.c_str());
			return false;
		}
	}
	
	return true;
}

/*************************************************************************/

bool DBWriter::write_string(char const *s)
{
	uint32_t len = (s ? strlen(s) : 0);
	
	/* Write the length of the string first. */
	if (!write_unsigned(len))
		return false;
	
	/* Then write the string, if length is different than zero. */
	if (len && (fwrite(s, 1, len, m_handle) != len))
	{
		error("Can't write to %s", m_newafn.c_str());
		return false;
	}
	
	return true;
}

/*************************************************************************/

bool DBWriter::write_time(time_t t)
{
	/* We store the time on 64-bit, so that it may be changed sometimes
	 * after 2038 ^^ */
	
	for (unsigned int i = 0; i < (8 - sizeof(t)); ++i)
	{
		if (fputc(0, m_handle) == EOF)
		{
			error("Can't write to %s", m_newafn.c_str());
			return false;
		}
	}
	
	for (int i = (sizeof(t) - 1); i >= 0; --i)
	{
		if (fputc(((t >> (i * 8)) & 0xFF), m_handle) == EOF)
		{
			error("Can't write to %s", m_newafn.c_str());
			return false;
		}
	}
	
	return true;
}

/*************************************************************************/

bool DBWriter::write_unsigned(uint32_t n)
{	
	for (int i = 3; i >= 0; --i)
	{
		if (fputc(((n >> (i * 8)) & 0xFF), m_handle) == EOF)
		{
			error("Can't write to %s", m_newafn.c_str());
			return false;
		}
	}
	
	return true;
}

/*************************************************************************/

bool DBWriter::write_valueinfo(uint16_t fieldid, uint16_t type)
{
	return (write_unsigned(fieldid) && write_unsigned(type));
}

/*************************************************************************/

bool DBWriter::write_version()
{
	return (write_unsigned(dbfile_version));
}

/*************************************************************************/

} /* namespace twilight */

/*************************************************************************/
