/* NickServ functions.
 *
 * Epona (c) 2000-2002, 2004 PegSoft
 * Contact us at epona@pegsoft.net
 *
 * This program is free but copyrighted software; see the file COPYING for
 * details.
 *
 * Based on the original code of Services by Andy Church.
 *
 */

#include <system/headers.h>

#include <core/directives.h>
#include <core/epona.h>
#include <databases/database.h>
#include <log/levels.h>
#include <mail/mail.h>
#include <misc/misc.h>
#include <modules/manager.h>

#include "chanserv.h"
#include "epona.h"
#include "nickserv.h"
#include "operserv.h"

#include "services.h"
#include "pseudo.h"

using namespace std;
using namespace core;

using misc::simple_range;

/*************************************************************************/

#define HASH(nick)	((static_cast<unsigned char>(tolower((nick)[0]))&31)<<5 | (static_cast<unsigned char>(tolower((nick)[1]))&31))

#define TO_COLLIDE   0			/* Collide the user with this nick */
#define TO_RELEASE   1			/* Release a collided nick */

/*************************************************************************/

LOGLEVELH(NoticeNickServ, xlog::LogNotice, "Notice::NickServ", "nickserv: ");
LOGLEVELH(NoticeNickServPrivs, NoticeNickServ, "Notice::NickServ::Privileges", "nickserv privileged: ");

/*************************************************************************/

static int group_identified(User *u, NickGroup *nc);

static void collide(NickAlias *na, int from_timeout);
static void release(NickAlias *na, int from_timeout);
static void add_ns_timeout(NickAlias *na, int type, time_t delay);
static void del_ns_timeout(NickAlias *na, int type);

static void do_help(User *u);
static void do_register(User *u);
static void do_group(User *u);
static void do_identify(User *u);
static void do_logout(User *u);
static void do_auth(User *u);
static void do_drop(User *u);
static void do_set(User *u);
static void do_set_gname(User *u, NickGroup *nc, char *param);
static void do_set_password(User *u, NickGroup *nc, char *param);
static void do_set_language(User *u, NickGroup *nc, char *param);
static void do_set_url(User *u, NickGroup *nc, char *param);
static void do_set_email(User *u, NickGroup *nc, char *param);
static void do_set_im(User *u, NickGroup *nc, char *param);
static void do_set_greet(User *u, NickGroup *nc, char *param);
static void do_set_kill(User *u, NickGroup *nc, char *param);
static void do_set_secure(User *u, NickGroup *nc, char *param);
static void do_set_private(User *u, NickGroup *nc, char *param);
static void do_set_host(User *u, NickGroup *nc, char *param);
static void do_set_hide(User *u, NickGroup *nc, char *param);
static void do_set_noexpire(User *u, NickAlias *na, char *param);
static void do_access(User *u);
static void do_info(User *u);
static void do_list(User *u);
static void do_glist(User *u);
static void do_recover(User *u);
static void do_release(User *u);
static void do_ghost(User *u);
static void do_status(User *u);
static void do_host(User *u);
static void do_getpass(User *u);
static void do_sendpass(User *u);
static void do_forbid(User *u);

/* Obsolete commands */
static void do_link(User *u);
static void do_unlink(User *u);
static void do_listlinks(User *u);

/*************************************************************************/

static int access_msgs[ListCmd::cmListCmdTotal];

/*************************************************************************/

namespace
{

/*************************************************************************/

class AccessCmd : public ListCmd
{
	public:
		/// Constructor.
		AccessCmd();
	protected:
		/// Overrides ListCmd::container().
		virtual container_type *container(User *u);
		
		/// Overrides ListCmd::changeable().
		virtual bool changeable(User *u, container_type &ctn,
				char const *sub) const;
		
		/// Overrides ListCmd::add().
		virtual int add(User *u, container_type &ctn, char const *sub);
		
		/// Adds a record to the access list.
		int add_record(User *u, NickGroup *owner, char const *mask);
		
		/// Overrides ListCmd::del_record().
		virtual bool del_record(User *u, container_type &ctn,
				char const *sub, databases::Record *rec);
		
		/// Overrides ListCmd::list_record().
		virtual bool list_record(User *u, container_type &ctn,
				char const *sub, uint16_t num, databases::Record const *rec,
				uint16_t count);
};

/*************************************************************************/

}

/*************************************************************************/

static Command cmds[] = {
    { "HELP",     do_help,     NULL,  -1,                     -1,-1,-1,-1 },
    { "REGISTER", do_register, NULL,  NICK_HELP_REGISTER,     -1,-1,-1,-1 },
    { "GROUP",    do_group,    NULL,  NICK_HELP_GROUP,        -1,-1,-1,-1 },
    { "IDENTIFY", do_identify, NULL,  NICK_HELP_IDENTIFY,     -1,-1,-1,-1 },
    { "SIDENTIFY",do_identify, NULL,  -1,                     -1,-1,-1,-1 },
    { "LOGOUT",   do_logout,   NULL,  -1,
		NICK_HELP_LOGOUT, NICK_SERVADMIN_HELP_LOGOUT,
		NICK_SERVADMIN_HELP_LOGOUT, NICK_SERVADMIN_HELP_LOGOUT },
	{ "AUTH",     do_auth,     NULL,  NICK_HELP_AUTH,         -1,-1,-1,-1 },
    { "DROP",     do_drop,     NULL,  -1,
		NICK_HELP_DROP, NICK_SERVADMIN_HELP_DROP,
		NICK_SERVADMIN_HELP_DROP, NICK_SERVADMIN_HELP_DROP },
	{ "LINK",     do_link,     NULL,  -1,                     -1,-1,-1,-1 },
	{ "UNLINK",   do_unlink,   NULL,  -1,                     -1,-1,-1,-1 },
	{ "LISTLINKS",do_listlinks,NULL,  -1,                     -1,-1,-1,-1 },
    { "ACCESS",   do_access,   NULL,  NICK_HELP_ACCESS,       -1,-1,-1,-1 },
    { "SET",      do_set,      NULL,  NICK_HELP_SET,
		-1, NICK_SERVADMIN_HELP_SET,
		NICK_SERVADMIN_HELP_SET, NICK_SERVADMIN_HELP_SET },
	{ "SET GNAME",    NULL,    NULL,  NICK_HELP_SET_GNAME,    -1,-1,-1,-1 },
    { "SET PASSWORD", NULL,    NULL,  NICK_HELP_SET_PASSWORD, -1,-1,-1,-1 },
    { "SET URL",      NULL,    NULL,  NICK_HELP_SET_URL,      -1,-1,-1,-1 },
    { "SET EMAIL",    NULL,    NULL,  NICK_HELP_SET_EMAIL,    -1,-1,-1,-1 },
    { "SET IM",       NULL,    NULL,  NICK_HELP_SET_IM,       -1,-1,-1,-1 },
    { "SET GREET",    NULL,    NULL,  NICK_HELP_SET_GREET,    -1,-1,-1,-1 },
    { "SET KILL",     NULL,    NULL,  NICK_HELP_SET_KILL,     -1,-1,-1,-1 },
    { "SET SECURE",   NULL,    NULL,  NICK_HELP_SET_SECURE,   -1,-1,-1,-1 },
    { "SET PRIVATE",  NULL,    NULL,  NICK_HELP_SET_PRIVATE,  -1,-1,-1,-1 },
    { "SET HOST",     NULL,    NULL,  NICK_HELP_SET_HOST,     -1,-1,-1,-1 },
    { "SET HIDE",     NULL,    NULL,  NICK_HELP_SET_HIDE,     -1,-1,-1,-1 },
    { "SET NOEXPIRE", NULL,    NULL,  -1, -1,
		NICK_SERVADMIN_HELP_SET_NOEXPIRE,
		NICK_SERVADMIN_HELP_SET_NOEXPIRE,
		NICK_SERVADMIN_HELP_SET_NOEXPIRE },
    { "RECOVER",  do_recover,  NULL,  NICK_HELP_RECOVER,      -1,-1,-1,-1 },
    { "RELEASE",  do_release,  NULL,  NICK_HELP_RELEASE,      -1,-1,-1,-1 },
    { "GHOST",    do_ghost,    NULL,  NICK_HELP_GHOST,        -1,-1,-1,-1 },
    { "INFO",     do_info,     NULL,  NICK_HELP_INFO,
		-1, NICK_HELP_INFO, NICK_SERVADMIN_HELP_INFO,
		NICK_SERVADMIN_HELP_INFO },
    { "LIST",     do_list,     NULL,  -1,
		NICK_HELP_LIST, NICK_SERVADMIN_HELP_LIST,
		NICK_SERVADMIN_HELP_LIST, NICK_SERVADMIN_HELP_LIST },
	{ "GLIST",    do_glist,    NULL,  -1,        
		NICK_HELP_GLIST, NICK_SERVADMIN_HELP_GLIST, 
		NICK_SERVADMIN_HELP_GLIST, NICK_SERVADMIN_HELP_GLIST },
    { "STATUS",   do_status,   NULL,  NICK_HELP_STATUS,       -1,-1,-1,-1 },
    { "HOST",   do_host,       NULL,   -1,
		NICK_HELP_HOST, NICK_SERVADMIN_HELP_HOST,
		NICK_SERVADMIN_HELP_HOST, NICK_SERVADMIN_HELP_HOST },
    { "SENDPASS", do_sendpass, NULL,  NICK_HELP_SENDPASS,     -1,-1,-1,-1 },
    { "GETPASS",  do_getpass,  is_services_admin,  -1,
		-1, NICK_SERVADMIN_HELP_GETPASS,
		NICK_SERVADMIN_HELP_GETPASS, NICK_SERVADMIN_HELP_GETPASS },
    { "FORBID",   do_forbid,   is_services_admin,  -1,        -1,-1,-1,-1 },
    { NULL }
};

/*************************************************************************/
/*************************************************************************/

/* Display total number of registered nicks and info about each; or, if
 * a specific nick is given, display information about that nick (like
 * /msg NickServ INFO <nick>).  If count_only != 0, then only display the
 * number of registered nicks (the nick parameter is ignored).
 */

void listnicks(int count_only, const char *nick)
{
    NickAlias *na;
    char *end;

    if (count_only) {

	printf("%d nicknames registered.\n", aliases.size());

    } else if (nick) {

	time_t t;
	struct tm *tm;
	char buf[512];
	static const char commastr[] = ", ";
	int need_comma = 0;

	if (!(na = aliases.find(nick))) {
	    printf("%s not registered.\n", nick);
	    return;
	} 
	printf("%s is %s\n", nick, (na->last_realname() ? na->last_realname() : "<unknown>"));
	if (na->last_usermask())
		printf("Last seen address: %s\n", na->last_usermask());
	t = na->created();
	tm = localtime(&t);
	strftime(buf, sizeof(buf), getstring3(STRFTIME_DATE_TIME_FORMAT), tm);
	printf("  Time registered: %s\n", buf);
	t = na->last_seen();
	tm = localtime(&t);
	strftime(buf, sizeof(buf), getstring3(STRFTIME_DATE_TIME_FORMAT), tm);
	printf("   Last seen time: %s\n", buf);
	if (na->ng()->url())
	    printf("              URL: %s\n", na->ng()->url());
	if (na->ng()->email())
	    printf("   E-mail address: %s\n", na->ng()->email());
	if (na->ng()->im_type() && na->ng()->im_id())
	    printf("            %s: %s\n", na->ng()->im_type(), na->ng()->im_id());
	if (na->ng()->greet())
		printf("            Greet: %s\n", na->ng()->greet());
#ifdef HAS_CHGHOST
	if (na->ng()->chost())
		printf("  Custom hostname: %s (set by %s)\n", na->ng()->chost(), 
				na->ng()->chost_setby());
#endif
	*buf = 0;
	end = buf;
	if (na->ng()->flags() & NickGroup::gfKill) {
	    end += snprintf(end, sizeof(buf)-(end-buf), "Kill protection");
	    need_comma = 1;
	}
	if (na->ng()->flags() & NickGroup::gfSecure) {
	    end += snprintf(buf, sizeof(buf)-(end-buf), "%sSecurity",
			need_comma ? commastr : "");
	    need_comma = 1;
	}
	if (na->ng()->flags() & NickGroup::gfPrivate) {
	    end += snprintf(buf, sizeof(buf)-(end-buf), "%sPrivate",
			need_comma ? commastr : "");
	    need_comma = 1;
	}
	if (na->status() & NickAlias::asNoExpire) {
	    end += snprintf(buf, sizeof(buf)-(end-buf), "%sNo Expire",
			need_comma ? commastr : "");
	    need_comma = 1;
	}
	printf("          Options: %s\n", *buf ? buf : "None");

	} else {
	
		Aliases::const_niterator it, end;
	
		for (it = aliases.nbegin(), end = aliases.nend(); it != end; ++it)
		{
			na = (NickAlias *)(*it);
			
			printf("    %s %-20s  %s\n", 
				    na->status() & NickAlias::asNoExpire ? "!" : " ", 
				    na->nick(),
				    (na->last_usermask() ? na->last_usermask() : ""));
		}
		
		printf("%d nicknames registered.\n", aliases.size());
		
	} 
}

/*************************************************************************/
/*************************************************************************/

/* NickServ initialization. */

void ns_init(void)
{
}

/*************************************************************************/

bool nickserv_load()
{
	/* Load password module */
	password::Encrypter *enc = 0;
		
	if (!epona->cd().nickserv.password_module.value().empty())
	{
		enc = groups.load_encrypter(epona->cd().nickserv.password_module.value().c_str());
		if (!enc)
			return false;
	}
	
	groups.encrypter(enc);
	
	if (!epona->modules().load(epona->cd().nickserv.database_groups))
	{
		epona->log(ErrorNickServ(), "Can't load database module %s: %s",
				epona->cd().nickserv.database_groups.value().c_str(),
				epona->modules().error().c_str());
		return false;
	}
	
	if (!epona->modules().load(epona->cd().nickserv.database_aliases))
	{
		epona->log(ErrorNickServ(), "Can't load database module %s: %s",
				epona->cd().nickserv.database_aliases.value().c_str(),
				epona->modules().error().c_str());
		return false;
	}
	
	if (!epona->modules().load(epona->cd().nickserv.database_access))
	{
		epona->log(ErrorNickServ(), "Can't load database module %s: %s",
				epona->cd().nickserv.database_access.value().c_str(),
				epona->modules().error().c_str());
		return false;
	}
	
	if (!epona->modules().load(epona->cd().nickserv.database_memos))
	{
		epona->log(ErrorNickServ(), "Can't load database module %s: %s",
				epona->cd().nickserv.database_memos.value().c_str(),
				epona->modules().error().c_str());
		return false;
	}
	
	access_msgs[ListCmd::cmSyntax] = NICK_ACCESS_SYNTAX;
	access_msgs[ListCmd::cmDisabled] = NICK_ACCESS_DISABLED;
	access_msgs[ListCmd::cmEmpty] = NICK_ACCESS_EMPTY;
	access_msgs[ListCmd::cmSyntaxDel] = NICK_ACCESS_DEL_SYNTAX;
	access_msgs[ListCmd::cmNoSuchEntry] = NICK_ACCESS_NO_SUCH_ENTRY;
	access_msgs[ListCmd::cmNoMatch] = NICK_ACCESS_NO_MATCH;
	access_msgs[ListCmd::cmDeletedOne] = NICK_ACCESS_DELETED_ONE;
	access_msgs[ListCmd::cmDeletedSeveral] = NICK_ACCESS_DELETED_SEVERAL;
	access_msgs[ListCmd::cmNotFound] = NICK_ACCESS_NOT_FOUND;
	access_msgs[ListCmd::cmDeleted] = NICK_ACCESS_DELETED;
	access_msgs[ListCmd::cmSyntaxList] = -1; /* Not used for ACCESS */
	access_msgs[ListCmd::cmListHeader] = NICK_ACCESS_LIST_HEADER;
	access_msgs[ListCmd::cmListedOne] = NICK_ACCESS_LISTED_ONE;
	access_msgs[ListCmd::cmListedSeveral] = NICK_ACCESS_LISTED_SEVERAL;
	access_msgs[ListCmd::cmClearPartially] = NICK_ACCESS_CLEAR_PARTIALLY;
	access_msgs[ListCmd::cmClear] = NICK_ACCESS_CLEAR;
	
	return true;
}

/*************************************************************************/

void nickserv_notify(modules::Module &sender, int msg, void *arg)
{
	switch (msg)
	{
		case modules::Module::mnLoaded:
			if (sender.name() == epona->cd().nickserv.database_groups.value())
			{
				databases::Database *db = (databases::Database *)
						sender.factory(databases::Database::IID);
				if (!db)
				{
					epona->fatalf("Can't create database nickserv_groups: "
							"module %s does not expose the Database interface",
							sender.name().c_str());	
				}
				
				/* Add a reference to the module */
				sender.references_add();
				/* Setup the database */
				groups.setup_database(db);
			}
			
			if (sender.name() == epona->cd().nickserv.database_aliases.value())
			{
				databases::Database *db = (databases::Database *)
						sender.factory(databases::Database::IID);
				if (!db)
				{
					epona->fatalf("Can't create database nickserv_aliases: "
							"module %s does not expose the Database interface",
							sender.name().c_str());	
				}
				
				/* Add a reference to the module */
				sender.references_add();
				/* Setup the database */
				aliases.setup_database(db);
			}
			
			if (sender.name() == epona->cd().nickserv.database_access.value())
			{
				databases::Database *db = (databases::Database *)
						sender.factory(databases::Database::IID);
				if (!db)
				{
					epona->fatalf("Can't create database nickserv_access: "
							"module %s does not expose the Database interface",
							sender.name().c_str());	
				}
				
				/* Add a reference to the module */
				sender.references_add();
				/* Setup the database */
				groups.accesses().setup_database(db);
			}
			
			if (sender.name() == epona->cd().nickserv.database_memos.value())
			{
				databases::Database *db = (databases::Database *)
						sender.factory(databases::Database::IID);
				if (!db)
				{
					epona->fatalf("Can't create database nickserv_memos: "
							"module %s does not expose the Database interface",
							sender.name().c_str());	
				}
				
				/* Add a reference to the module */
				sender.references_add();
				/* Setup the database */
				memos.setup_database(db);
			}
		default:
			break;
	}
}

/*************************************************************************/

void nickserv_start()
{
	if (!groups.db() || !groups.db()->start())
		epona->fatal("Can't start database nickserv_groups");
	
	aliases.setup_owner();
	
	if (!aliases.db() || !aliases.db()->start())
		epona->fatal("Can't start database nickserv_aliases");
	
	groups.accesses().setup_owner();
	
	if (!groups.accesses().db() || !groups.accesses().db()->start())
		epona->fatal("Can't start database nickserv_access");
	
	memos.setup_owner();
	
	if (!memos.db() || !memos.db()->start())
		epona->fatal("Can't start database nickserv_memos");
}

/*************************************************************************/

/* Main NickServ routine. */

void nickserv(User *u, char *buf)
{
    char *cmd, *s;

    cmd = strtok(buf, " ");

    if (!cmd) {
		return;
    } else if (stricmp(cmd, "\1PING") == 0) {
		s = strtok(NULL, "\1");
	    
		notice(s_NickServ, u->nick, "\1PING%s%s\1", (s ? " " : ""),
				(s ? s : ""));
    } else {
		run_cmd(s_NickServ, u, cmds, cmd);
    }

}

/*************************************************************************/

/* Check whether a user is on the access list of the nick they're using If 
 * not, send warnings as appropriate.  If so (and not NickGroup::gfSecure), update 
 * last seen info.  
 * Return 1 if the user is valid and recognized, 0 otherwise (note
 * that this means an NickGroup::gfSecure nick will return 0 from here).  
 * If the user's nick is not registered, 0 is returned.
 */

int validate_user(User *u)
{
    NickAlias *na;
    int on_access;

    if (!(na = u->na))
	return 0;

	on_access = na->ng()->on_access(u);
	if (on_access)
		na->status_add(NickAlias::asOnAccess);

    if (!(na->ng()->flags() & NickGroup::gfSecure) && on_access) 
    {
		na->status_add(NickAlias::asRecognized);
		na->post_recognized();
		return 1;
    }

    if (on_access || !(na->ng()->flags() & NickGroup::gfKillImmed)) {
		if (na->ng()->flags() & NickGroup::gfSecure)
	    	notice_lang(s_NickServ, u, NICK_IS_SECURE, s_NickServ);
		else
	    	notice_lang(s_NickServ, u, NICK_IS_REGISTERED, s_NickServ);
    }

    if ((na->ng()->flags() & NickGroup::gfKill) && !on_access) {
		if (na->ng()->flags() & NickGroup::gfKillImmed) {
			notice_lang(s_NickServ, u, FORCENICKCHANGE_NOW);
	    	collide(na, 0);
		} else if (na->ng()->flags() & NickGroup::gfKillQuick) {
	    	notice_lang(s_NickServ, u, FORCENICKCHANGE_IN_20_SECONDS);
	    	add_ns_timeout(na, TO_COLLIDE, 20);
		} else {
	    	notice_lang(s_NickServ, u, FORCENICKCHANGE_IN_1_MINUTE);
	    	add_ns_timeout(na, TO_COLLIDE, 60);
		}
    }

    return 0;
}

/*************************************************************************/

/* Cancel validation flags for a nick (i.e. when the user with that nick
 * signs off or changes nicks).  Also cancels any impending collide. */

void cancel_user(User *u)
{
    NickAlias *na = u->na;
    
    if (na) {
		if (na->status() & NickAlias::asGuested) {
	    	s_nick(u->nick, epona->cd().nickserv.enforcer.username,
	    			epona->cd().nickserv.enforcer.hostname,
	    			"Services Enforcer", "+", 0);
	    	add_ns_timeout(na, TO_RELEASE,
	    			epona->cd().nickserv.release_timeout);
	    	na->status_remove(NickAlias::asTemporary);
	    	na->status_add(NickAlias::asKillHeld);
		} else {
	    	na->status_remove(NickAlias::asTemporary);
		}
		
		del_ns_timeout(na, TO_COLLIDE);
    }
}

/*************************************************************************/

/* Return whether a user has identified for their nickname. */

int nick_identified(User *u)
{
    return u->na && (u->na->status() & NickAlias::asIdentified);
}

/*************************************************************************/

/* Return whether a user is recognized for their nickname. */

int nick_recognized(User *u)
{
    return u->na && (u->na->status() & (NickAlias::asIdentified | NickAlias::asRecognized));
}

/*************************************************************************/

/* Returns whether a user is identified AND in the group nc */

static int group_identified(User *u, NickGroup *ng) 
{
	return nick_identified(u) && u->na->ng() == ng;
}

/*************************************************************************/

#ifdef HAS_CHGHOST

/* Prerequisites: user is identified */

void custom_host_set(User *user)
{
	if (!epona->cd().nickserv.custom_host.enable)
	{
		custom_host_unset(user);
		return;
	}
	
	if (user->na->ng()->flags() & NickGroup::gfCustomHost) 
	{
		if (!user->na->ng()->chost()) 
		{
			custom_host_unset(user);
			user->na->ng()->flags_remove(NickGroup::gfCustomHost);
			return;
		}
		
#ifdef HAS_VHOST
		if (!strcmp(user->vhost, user->na->ng()->chost()))
			return;
#else
		if (!strcmp(user->host, user->na->ng()->chost()))
			return;
#endif
		
		/* Store the ircd-generated vhost in case we
		 * need to restore it later. */
		if (!user->ivhost) {
#ifdef HAS_VHOST
			user->ivhost = user->vhost;
			user->vhost = NULL; /* Important */
#else
			user->ivhost = user->host;
			user->host = NULL; /* Important */
#endif
		}
		
		s_chghost(user->nick, user->na->ng()->chost(), user->timestamp);
		change_user_host(user, user->na->ng()->chost());
		
#ifdef IRC_SEQUANA
		if (!IsHostHidden(user)) {
			char *av[1];
			av[0] = "+" HiddenHostMode;
			set_umode(user, 1, av);
		}
#else
# ifdef HAS_VHOST
		if (!IsHostHidden(user))
			change_user_mode(s_NickServ, user, "+" HiddenHostMode, NULL);
# endif
#endif
	}
}

/*************************************************************************/

void custom_host_unset(User *user)
{
	/* If we changed the vhost before... */
	if (user->ivhost) 
	{
		/* ... restore the ircd-generated vhost */
		s_chghost(user->nick, user->ivhost, user->timestamp);
		change_user_host(user, user->ivhost);
		
#ifdef IRC_SEQUANA
		if (!IsHostHidden(user)) {
			char *av[1];
			av[0] = "+" HiddenHostMode;
			set_umode(user, 1, av);
		}
#endif
		
		free(user->ivhost);
		user->ivhost = NULL;
	}
}

#endif /* HAS_CHGHOST */

/*************************************************************************/
/*********************** NickServ private routines ***********************/
/*************************************************************************/

/* Collide a nick. 
 *
 * NickServ is now able to force a nick change instead of killing the user. 
 * If a nick change is forced, we do not introduce the enforcer nick until
 * the user's nick actually changes. This is watched for and done in
 * cancel_user(). */

static void collide(NickAlias *na, int from_timeout)
{
	if (!from_timeout)
		del_ns_timeout(na, TO_COLLIDE);

	s_svsnick(s_NickServ, na->u);
	na->status_add(NickAlias::asGuested);
}

/*************************************************************************/

/* Release hold on a nick. */

static void release(NickAlias *na, int from_timeout)
{
    if (!from_timeout)
	del_ns_timeout(na, TO_RELEASE);
    send_cmd(na->nick(), "QUIT");
    na->status_remove(NickAlias::asKillHeld);
}

/*************************************************************************/
/*************************************************************************/

static struct my_timeout {
    struct my_timeout *next, *prev;
    NickAlias *na;
    Timeout *to;
    int type;
} *my_timeouts;

/*************************************************************************/

/* Remove a collide/release timeout from our private list. */

static void rem_ns_timeout(NickAlias *na, int type)
{
    struct my_timeout *t, *t2;

    t = my_timeouts;
    while (t) {
		if (t->na == na && t->type == type) {
			t2 = t->next;
	    	if (t->next)
			t->next->prev = t->prev;
	    	if (t->prev)
			t->prev->next = t->next;
	    	else
			my_timeouts = t->next;
	    	free(t);
	    	t = t2;
		} else {
	    	t = t->next;
		}
    }
}

/*************************************************************************/

/* Collide a nick on timeout. */

static void timeout_collide(Timeout *t)
{
    NickAlias *na = static_cast<NickAlias *>(t->data);

	rem_ns_timeout(na, TO_COLLIDE);
    /* If they identified or don't exist anymore, don't kill them. */
    if ((na->status() & NickAlias::asIdentified) || !na->u || na->u->my_signon > t->settime)
	return;
    /* The RELEASE timeout will always add to the beginning of the
     * list, so we won't see it.  Which is fine because it can't be
     * triggered yet anyway. */
    collide(na, 1);
}

/*************************************************************************/

/* Release a nick on timeout. */

static void timeout_release(Timeout *t)
{
    NickAlias *na = static_cast<NickAlias *>(t->data);
    
	rem_ns_timeout(na, TO_RELEASE);
    release(na, 1);
}

/*************************************************************************/

/* Add a collide/release timeout. */

void add_ns_timeout(NickAlias *na, int type, time_t delay)
{
    Timeout *to;
    struct my_timeout *t;
    void (*timeout_routine)(Timeout *);

    if (type == TO_COLLIDE)
		timeout_routine = timeout_collide;
    else if (type == TO_RELEASE)
		timeout_routine = timeout_release;
    else {
		epona->log(ErrorNickServ(),
				"Unknown timeout type %d! na=%p (%s), delay=%ld", type,
				static_cast<void *>(na), na->nick(), (long) delay);
		return;
    }
    
    to = add_timeout(delay, timeout_routine, 0);
    to->data = na;
    
    t = static_cast<my_timeout *>(scalloc(sizeof(struct my_timeout), 1));
    t->na = na;
    t->to = to;
    t->type = type;
    
    t->prev = NULL;
    t->next = my_timeouts;
    my_timeouts = t;
    /* Andy Church should stop coding while being drunk. 
     * Here's the two lines he forgot that produced the timed_update evil bug
     * and a *big* memory leak. 
     */
    if (t->next)
    t->next->prev = t;
}

/*************************************************************************/

/* Delete a collide/release timeout. */

static void del_ns_timeout(NickAlias *na, int type)
{
    struct my_timeout *t, *t2;

    t = my_timeouts;
    while (t) {
		if (t->na == na && t->type == type) {
			t2 = t->next;
	    	if (t->next)
			t->next->prev = t->prev;
	    	if (t->prev)
			t->prev->next = t->next;
	    	else
			my_timeouts = t->next;
	    	del_timeout(t->to);
	    	free(t);
	    	t = t2;
		} else {
	    	t = t->next;
		}
    }
}

/*************************************************************************/

/* Deletes all timeouts belonging to a given nick.
 * This should only be called before nick deletion. 
 */

void clean_ns_timeouts(NickAlias *na)
{
	struct my_timeout *t, *next;
	
	for (t = my_timeouts; t; t = next) {
		next = t->next;
		if (t->na == na) {
			if (epona->debuglevel())
			epona->log(DebugNickServ(), "Deleting timeout type %d from %s",
					t->type, t->na->nick()); 
			/* If the timeout has the TO_RELEASE type, we should release the user */
			if (t->type == TO_RELEASE)
			release(na, 1); 
			if (t->next)
			t->next->prev = t->prev;
	    	if (t->prev)
			t->prev->next = t->next;
	    	else
			my_timeouts = t->next;
	    	del_timeout(t->to);
	    	free(t);
		}	
	}
}

/*************************************************************************/
/*********************** NickServ command routines ***********************/
/*************************************************************************/

/* Return a help message. */

static void do_help(User *u)
{
    char *cmd = strtok(NULL, "");

    if (!cmd) {
    	notice_help(s_NickServ, u, NICK_HELP);
		if (epona->cd().nickserv.expire && 
				epona->cd().nickserv.expire.delay >= 86400)
	    	notice_help(s_NickServ, u, NICK_HELP_EXPIRES,
	    			epona->cd().nickserv.expire.delay / 86400);
		if (is_services_oper(u))
	    	notice_help(s_NickServ, u, NICK_SERVADMIN_HELP);
    } else if (stricmp(cmd, "SET LANGUAGE") == 0) {
		int i;
		notice_help(s_NickServ, u, NICK_HELP_SET_LANGUAGE);
		for (i = 0; i < NUM_LANGS && langlist[i] >= 0; i++)
			notice_user(s_NickServ, u, "    %2d) %s", i + 1,
					langnames[langlist[i]]);
    } else {
		help_cmd(s_NickServ, u, cmds, cmd);
    }
}

/*************************************************************************/

/* Register a nick. */

static void do_register(User *u)
{
	NickAlias *na;
	size_t prefixlen = strlen(epona->cd().services.guest_nick);
	size_t nicklen = strlen(u->nick);
	char *pass = strtok(NULL, " ");
	char *email = strtok(NULL, " ");
	
	if (!groups.writable() || !aliases.writable()) 
	{
		notice_lang(s_NickServ, u, NICK_REGISTRATION_DISABLED);
		return;
	}
	
	/* Prevent guest nicks registration */
	if (nicklen <= prefixlen+5 && nicklen >= prefixlen+1 &&
			stristr(u->nick, epona->cd().services.guest_nick) == u->nick &&
			strspn(u->nick+prefixlen, "1234567890ABCDEFabcdef") ==
							nicklen-prefixlen) 
	{
	    notice_lang(s_NickServ, u, NICK_CANNOT_BE_REGISTERED, u->nick);
	    return;
	}
	
	if (!pass || (epona->cd().nickserv.require_email && !email)) {
		syntax_error(s_NickServ, u, "REGISTER", NICK_REGISTER_SYNTAX_EMAIL);
	} else if (!is_oper(u) &&
			(time(0) < (u->lastnickreg + epona->cd().nickserv.reg_delay))) {
		notice_lang(s_NickServ, u, NICK_REG_PLEASE_WAIT, 
				(time_t) epona->cd().nickserv.reg_delay);
	} else if (u->na) {	/* i.e. there's already such a nick regged */
	    notice_lang(s_NickServ, u, NICK_ALREADY_REGISTERED, u->nick);
	
	} else if (stricmp(u->nick, pass) == 0 ||
			(epona->cd().strict_passwords && strlen(pass) < 5)) {
		notice_lang(s_NickServ, u, MORE_OBSCURE_PASSWORD);
	} else if (email && !mail::Mail::valid_address(email, true)) {
		notice_lang(s_NickServ, u, MAIL_X_INVALID, email);
	} else {
		NickGroup *ng = (NickGroup *)groups.create();
		
		if (!ng)
		{
			notice_lang(s_NickServ, u, NICK_REGISTRATION_FAILED);
			return;
		}
		
		if (strlen(pass) > NickGroup::pass_rlen)
		{
			pass[NickGroup::pass_rlen] = 0;
			notice_lang(s_NickServ, u, PASSWORD_TRUNCATED,
					NickGroup::pass_rlen);
		}
		
		if (!ng->change_password(pass))
		{
			memset(pass, 0, strlen(pass));
			notice_lang(s_NickServ, u, NICK_REGISTRATION_FAILED);
			delete ng;
			
			return;
		}
		
		if (ng->enc())
			memset(pass, 0, strlen(pass));
		
		if (email) 
			ng->email(email);		
		ng->language(epona->cd().nickserv.default_language);
		ng->maxmemos(epona->cd().memoserv.max_memos);
		
		na = (NickAlias *)aliases.create();
		
		if (na)
		{
			na->nick(u->nick);
			
			if (!epona->cd().nickserv.auth ||
					!epona->cd().nickserv.auth.restrict.identify())
				na->status_add(NickAlias::asIdentified | NickAlias::asRecognized);
			
			if (!groups.add(ng))
			{	
				notice_lang(s_NickServ, u, NICK_REGISTRATION_FAILED);
				
				epona->log(ErrorNickServ(),
						"Groups::add(%s) failed", u->nick);
				
				delete na;
				delete ng;
				
				return;
			}
			
			/* Set na->ng() AFTER the group has been added */
			na->ng(ng);
			
			/* Set this so Aliases::add() doesn't have to search for
			 * the user. */
			na->u = u;
			
			/* Set this so we don't have to update them immediately
			 * after the record has been created. */
			na->last_usermask(u->username, GetHost(u));
			na->last_realname(u->realname);
			na->last_seen(time(0));
			
			if (!aliases.add(na))
			{
				notice_lang(s_NickServ, u, NICK_REGISTRATION_FAILED);
				
				epona->log(ErrorNickServ(),
						"Aliases::add(%s) failed", u->nick);
				
				/* If removing the group fails... too bad, but we
				 * can probably continue anyway... I think. ;) */
				groups.remove(ng);
				delete na;
			}
			
			/* Set the name AFTER both the group and the alias have
			 * been created because the alias needs to have an ID
			 * before the group can use it. */
			na->ng()->change_name(na);
			
			/* Now create the access list entry... Don't fail if
			 * it can't be created though. */
			
			NickGroup::Access *nga = (NickGroup::Access *)
					groups.accesses().create();
			
			nga->owner(ng);
			char *accmask = create_mask(u);
			nga->mask(accmask);
			free(accmask);
			
			if (!groups.accesses().add(nga))
			{
				epona->log(ErrorNickServ(),
						"Groups::Accesses::add(%s, %s) failed", u->nick,
						accmask);
				
				delete nga;
				nga = 0;
			}
			
			/* Success */
			
			epona->log(InfoNickServList(),
					"`%s' registered by %s@%s (e-mail: %s)", u->nick,
					u->username, GetHost(u), (email ? email : "none"));
			
			notice_lang(s_NickServ, u, NICK_REGISTERED, u->nick,
					(nga ? nga->mask() : "(none)"));
			
			if (!na->ng()->enc())
		    	notice_lang(s_NickServ, u, NICK_PASSWORD_IS, na->ng()->pass());
			
			u->lastnickreg = time(0);
			
			if (epona->cd().nickserv.auth && epona->cd().nickserv.auth.required)
				na->ng()->require_auth(u, true);
		} else {
			delete ng;
			
			epona->log(ErrorNickServ(), "Creation of alias %s failed", u->nick);
			notice_lang(s_NickServ, u, NICK_REGISTRATION_FAILED);
		}
    }
}

/*************************************************************************/

/* Register a nick in a specified group. */

static void do_group(User *u)
{
	NickAlias *na, *target;
	char *nick = strtok(NULL, " ");
	char *pass = strtok(NULL, " ");
	
	if (!aliases.writable() || !groups.writable()) 
	{
		notice_lang(s_NickServ, u, NICK_GROUP_DISABLED);
		return;
	}
	
	if (!nick || !pass) {
		syntax_error(s_NickServ, u, "GROUP", NICK_GROUP_SYNTAX);
	} else if (!is_oper(u) &&
			time(0) < (u->lastnickreg + epona->cd().nickserv.reg_delay)) {
		notice_lang(s_NickServ, u, NICK_GROUP_PLEASE_WAIT,
				(time_t) epona->cd().nickserv.reg_delay);
	} else if (u->na && !epona->cd().nickserv.group_change) {
		notice_lang(s_NickServ, u, NICK_GROUP_CHANGE_DISABLED, s_NickServ);
	} else if (u->na && !nick_identified(u)) {
		notice_lang(s_NickServ, u, NICK_IDENTIFY_REQUIRED, s_NickServ);
	} else if (!(target = aliases.find(nick))) {
		notice_lang(s_NickServ, u, NICK_X_NOT_REGISTERED, nick);
	} else if (u->na && target->ng() == u->na->ng()) {
		notice_lang(s_NickServ, u, NICK_GROUP_SAME, target->nick());
	} else if (epona->cd().nickserv.auth.restrict.group() &&
			!(target->ng()->flags() & NickGroup::gfAuth)) {
		notice_lang(s_NickServ, u, NICK_GROUP_NOT_AUTH);
	} else if (epona->cd().nickserv.max_aliases &&
			(target->ng()->aliases().size() >= epona->cd().nickserv.max_aliases) &&
			!nick_is_services_admin(target->ng())) {
		notice_lang(s_NickServ, u, NICK_GROUP_TOO_MANY, target->nick(), 
				s_NickServ, s_NickServ);
	} else if (!target->ng()->check_password(pass)) {
		epona->log(InfoNickServRights(),
				"Failed GROUP for %s!%s@%s (invalid password)", u->nick,
				u->username, GetHost(u));
		notice_lang(s_NickServ, u, PASSWORD_INCORRECT);
		bad_password(u);
	} else {
		/* If the nick is already registered, drop it.
		 * If not, check that it is valid. */
		if (u->na) 
		{
			if (!aliases.remove(u->na))
			{
				notice_lang(s_NickServ, u, NICK_GROUP_FAILED);
				return;
			}
		} else {
			size_t prefixlen = strlen(epona->cd().services.guest_nick);
			size_t nicklen = strlen(u->nick);
			
			if (nicklen <= prefixlen+5 && nicklen >= prefixlen+1 &&
					stristr(u->nick, epona->cd().services.guest_nick) == u->nick &&
					strspn(u->nick+prefixlen, "1234567890ABCDEFabcdef") ==
					nicklen-prefixlen)
			{
				notice_lang(s_NickServ, u, NICK_CANNOT_BE_REGISTERED, u->nick);
				return;
			}
		}
		
		na = (NickAlias *)aliases.create();
		
		if (na) 
		{
			na->nick(u->nick);
			na->status_add(NickAlias::asIdentified | NickAlias::asRecognized);
			
			/* Set the group */
			na->ng(target->ng());
			
			/* Set this here so that Aliases::add doesn't have
			 * to look for the user */
			na->u = u;
			
			/* Set this so we don't have to update them immediately
			 * after the record has been created. */
			na->last_usermask(u->username, GetHost(u));
			na->last_realname(u->realname);
			na->last_seen(time(0));
			
			if (!aliases.add(na))
			{
				notice_lang(s_NickServ, u, NICK_GROUP_FAILED);
				
				epona->log(ErrorNickServ(),
						"Aliases::add(%s) failed", u->nick);
				
				delete na;
			}
			
			epona->log(InfoNickServList(),
					"%s!%s@%s makes %s join group of %s (%s) (e-mail: %s)",
					u->nick, u->username, GetHost(u), u->nick, target->nick(), 
					target->ng()->name(), (target->ng()->email() ? 
					target->ng()->email(): "none"));
			notice_lang(s_NickServ, u, NICK_GROUP_JOINED, target->nick());
			
			u->lastnickreg = time(NULL);
		} else {
			epona->log(ErrorNickServ(), "Creation of aliases %s failed",
					u->nick);
			notice_lang(s_NickServ, u, NICK_GROUP_FAILED);
		}
	}
}

/*************************************************************************/

static void do_identify(User *u)
{
    char *pass = strtok(NULL, " ");
    NickAlias *na;

    if (!pass) {
		syntax_error(s_NickServ, u, "IDENTIFY", NICK_IDENTIFY_SYNTAX);
    } else if (!(na = u->na)) {
		notice_lang(s_NickServ, u, NICK_NOT_REGISTERED);
    } else if (!na->ng()->check_password(pass)) {
		epona->log(InfoNickServRights(), "Failed IDENTIFY for %s!%s@%s",
				u->nick, u->username, GetHost(u));
		notice_lang(s_NickServ, u, PASSWORD_INCORRECT);
		bad_password(u);
    } else {
		if (epona->cd().nickserv.auth &&
				epona->cd().nickserv.auth.restrict.identify() &&
				!(na->ng()->flags() & NickGroup::gfAuth)) 
		{
			if (!epona->cd().nickserv.auth.required ||
					na->ng()->auth_code() || !na->ng()->email())
				notice_lang(s_NickServ, u, NICK_IDENTIFY_NOT_AUTH);
			if (!epona->cd().nickserv.auth.required ||
					!na->ng()->email())
				notice_lang(s_NickServ, u,
						NICK_IDENTIFY_NOT_AUTH_HOWTO, s_NickServ);
		} else
		{
			na->status_add(NickAlias::asIdentified);
			na->post_identified();
		
			epona->log(InfoNickServRights(), "%s!%s@%s identified for nick %s",
					u->nick, u->username, GetHost(u), u->nick);
			notice_lang(s_NickServ, u, NICK_IDENTIFY_SUCCEEDED);
	    	
	    	if (epona->cd().nickserv.require_email && !na->ng()->email()) 
	    	{
				notice_lang(s_NickServ, u, NICK_IDENTIFY_EMAIL_REQUIRED);
				notice_lang(s_NickServ, u, NICK_IDENTIFY_EMAIL_HOWTO, s_NickServ);
			}
		}
		
		if (epona->cd().nickserv.auth && epona->cd().nickserv.auth.required)
			na->ng()->require_auth(u, false);
    }
}

/*************************************************************************/

static void do_logout(User *u)
{
    char *nick = strtok(NULL, " ");
    char *param = strtok(NULL, " ");
    User *u2;

    if (!is_services_admin(u) && nick) {
		syntax_error(s_NickServ, u, "LOGOUT", NICK_LOGOUT_SYNTAX);
    } else if (!(u2 = (nick ? finduser(nick) : u))) {
		notice_lang(s_NickServ, u, NICK_X_NOT_IN_USE, nick);
	} else if (!u2->na) {
		if (nick)
	    	notice_lang(s_NickServ, u, NICK_X_NOT_REGISTERED, nick);
		else
	    	notice_lang(s_NickServ, u, NICK_NOT_REGISTERED);
    } else if (!nick && !nick_identified(u)) {
		notice_lang(s_NickServ, u, NICK_IDENTIFY_REQUIRED, s_NickServ);
	} else {
		u2->na->post_deidentified();
		
		if (nick && param && !stricmp(param, "REVALIDATE")) {
			cancel_user(u2);
			validate_user(u2);
		} else {
			u2->na->status_remove(NickAlias::asIdentified | NickAlias::asRecognized);	
		}

		epona->log(InfoNickServRights(), "%s!%s@%s logged out nickname %s",
				u->nick, u->username, GetHost(u), u2->nick);
	
		if (nick)
			notice_lang(s_NickServ, u, NICK_LOGOUT_X_SUCCEEDED, nick);
		else
	    	notice_lang(s_NickServ, u, NICK_LOGOUT_SUCCEEDED);
    }
}

/*************************************************************************/

static void do_auth(User *u)
{
	if (!epona->cd().nickserv.auth)
	{
		notice_lang(s_NickServ, u, NICK_AUTH_UNAVAILABLE);
		return;
	}
	
	if (!groups.writable())
	{
		notice_lang(s_NickServ, u, NICK_AUTH_DISABLED);
		return;
	}
	
	if (!u->na)
	{
		notice_lang(s_NickServ, u, NICK_NOT_REGISTERED);
		return;
	}
	
	char const *param = strtok(NULL, " ");
	
	if (!param)
	{
		bool identified = nick_identified(u);
		
		if (!epona->cd().nickserv.auth.restrict.identify() && !identified) {
			notice_lang(s_NickServ, u, NICK_IDENTIFY_REQUIRED, s_NickServ);
		} else if (u->na->ng()->flags() & NickGroup::gfAuth) {
			notice_lang(s_NickServ, u, NICK_AUTH_START_AUTHENTIFIED);
		} else if (u->na->ng()->auth_code()) {
			notice_lang(s_NickServ, u, NICK_AUTH_START_ALREADY);
		} else if (!u->na->ng()->email()) {
			notice_lang(s_NickServ, u,
					NICK_AUTH_START_EMAIL_MISSING, s_NickServ);
		} else
		{
			if (!u->na->ng()->begin_auth())
			{
				notice_lang(s_NickServ, u, NICK_AUTH_START_FAILED);
				return;
			}
			
			if (!u->na->ng()->mail_auth(u->na->nick(), u->host, false))
			{
				u->na->ng()->end_auth(false);
				
				notice_lang(s_NickServ, u, NICK_AUTH_START_FAILED);
				return;
			}
			
			epona->log(InfoNickServ(), "%s!%s@%s started an authentication "
					"process (group %s; e-mail: %s)", u->nick, u->username,
					GetHost(u), u->na->ng()->name(), u->na->ng()->email());
			
			if (identified)
				notice_lang(s_NickServ, u, NICK_AUTH_START_DONE,
						u->na->ng()->email());
			else
				notice_lang(s_NickServ, u, NICK_AUTH_START_DONE_PRIVATE);
		}
	} else if (!stricmp(param, "CANCEL"))
	{
		if (!epona->cd().nickserv.auth.restrict.identify() &&
				!nick_identified(u)) {
			notice_lang(s_NickServ, u, NICK_IDENTIFY_REQUIRED, s_NickServ);
		} else if (!u->na->ng()->auth_code()) {
			notice_lang(s_NickServ, u, NICK_AUTH_NOT_IN_PROGRESS);
		} else 
		{
			u->na->ng()->end_auth(false);
			
			epona->log(InfoNickServ(), "%s!%s@%s cancelled its authentication "
					"process (group %s; e-mail: %s)", u->nick, u->username,
					GetHost(u), u->na->ng()->name(), u->na->ng()->email());
			notice_lang(s_NickServ, u, NICK_AUTH_CANCEL_DONE);
		}
	} else if (isxdigit(*param) || !stricmp(param, "PASS"))
	{
		char *newpass = 0;
		
		if (!stricmp(param, "PASS"))
		{
			if (!u->na->ng()->enc())
			{
				notice_lang(s_NickServ, u, NICK_AUTH_PASS_UNAVAILABLE);
				return;
			}
			
			param = strtok(NULL, " ");
			newpass = strtok(NULL, " ");
			
			if (!param || !newpass)
			{
				syntax_error(s_NickServ, u, "AUTH", NICK_AUTH_SYNTAX);
				return;
			}
		}
		
		if (!epona->cd().nickserv.auth.restrict.identify() && !newpass &&
				!nick_identified(u)) {
			notice_lang(s_NickServ, u, NICK_IDENTIFY_REQUIRED, s_NickServ);
		} else if (!u->na->ng()->auth_code()) {
			notice_lang(s_NickServ, u, NICK_AUTH_NOT_IN_PROGRESS);
		} else if (strtoul(param, 0, 16) != u->na->ng()->auth_code()) 
		{
			notice_lang(s_NickServ, u, NICK_AUTH_END_BAD_CODE, param);
			if (u->na->ng()->bad_auths_inc() >=
					epona->cd().nickserv.auth.attempts)
			{
				u->na->ng()->end_auth(false);
			
				epona->log(InfoNickServ(), "%s!%s@%s did too many attempts "
						"for authentication (group %s; e-mail: %s)", u->nick,
						u->username, GetHost(u), u->na->ng()->name(),
						u->na->ng()->email());
				notice_lang(s_NickServ, u, NICK_AUTH_END_TOO_MANY);
			}
		} else
		{
			bool wasauth = (u->na->ng()->flags() & NickGroup::gfAuth);
			
			if (!u->na->ng()->end_auth(true))
			{
				notice_lang(s_NickServ, u, NICK_AUTH_END_FAILED);
				return;
			}
			
			epona->log(InfoNickServ(), "%s!%s@%s completed authentication "
					"(group %s; e-mail: %s)", u->nick, u->username, GetHost(u),
					u->na->ng()->name(), u->na->ng()->email());
			if (!wasauth)
				notice_lang(s_NickServ, u, NICK_AUTH_END_DONE);
			
			if (newpass)
				do_set_password(u, u->na->ng(), newpass);
		}
	} else
	{
		syntax_error(s_NickServ, u, "AUTH", NICK_AUTH_SYNTAX);
	}
}

/*************************************************************************/

static void do_drop(User *u)
{
    char *nick = strtok(NULL, " ");
    NickAlias *na;
    int is_servadmin = is_services_admin(u);
    int is_mine;		/* Does the nick being dropped belong to the user that is dropping? */

    if (!aliases.writable() || !groups.writable()) 
    {
		notice_lang(s_NickServ, u, NICK_DROP_DISABLED);
		return;
    }

    if (!(na = (nick ? aliases.find(nick) : u->na))) 
    {
		if (nick)
	    	notice_lang(s_NickServ, u, NICK_X_NOT_REGISTERED, nick);
		else
	    	notice_lang(s_NickServ, u, NICK_NOT_REGISTERED);
	    return;
	}
	
	is_mine = (u->na && (u->na->ng() == na->ng())); 
	
	if (is_mine && !nick_identified(u)) {
		notice_lang(s_NickServ, u, NICK_IDENTIFY_REQUIRED, s_NickServ);
	} else if (!is_mine && !is_servadmin) {
		notice_lang(s_NickServ, u, ACCESS_DENIED);
    } else if (epona->cd().nickserv.protect_admins && !is_mine &&
    		nick_is_services_admin(na->ng()) && !is_services_root(u)) {
		notice_lang(s_NickServ, u, PERMISSION_DENIED);
    } else {
	    epona->log(InfoNickServList(),
	    		"%s!%s@%s is dropping nickname %s (group %s) (e-mail: %s)", 
	    		u->nick, u->username, GetHost(u), na->nick(),
	    		na->ng()->name(),
	    		(na->ng()->email() ? na->ng()->email() : "none"));
	    
	    if (!aliases.remove(na))
	    {
			notice_lang(s_NickServ, u, NICK_DROP_FAILED);
			epona->log(InfoNickServList(), "Removal of nickname %s failed",
					na->nick());
			return;
		}
	
		if (!is_mine) {
			if (epona->cd().operserv.wall & Epona::Directives::OperServ::Wall::waDrop)
				wallops(s_NickServ, "\2%s\2 used DROP on \2%s\2", u->nick, nick);
	    	notice_lang(s_NickServ, u, NICK_X_DROPPED, nick);
		} else {
			if (nick) 
				notice_lang(s_NickServ, u, NICK_X_DROPPED, nick);
			else 
				notice_lang(s_NickServ, u, NICK_DROPPED);
		}
    }
}

/*************************************************************************/

/* A special version of notice lang made for SET, that allows for a different
 * reply when a services admin uses the command on another user.
 */
static void notice_lang_set(char const *source, User *u, int msg, int samsg,
		NickGroup *target, char const *p1 = 0, char const *p2 = 0)
{
	if (u->na->ng() == target)
		notice_lang(source, u, msg, p1, p2);
	else
		notice_lang(source, u, samsg, target->name(), p1, p2);
}

/*************************************************************************/

static void do_set(User *u) 
{
	char *cmd    = strtok(NULL, " ");    
	char *param  = strtok(NULL, " ");
    
    NickAlias *na;
    int is_servadmin = is_services_admin(u);
    int set_nick = 0;

    if (!aliases.writable() || !groups.writable()) 
    {
		notice_lang(s_NickServ, u, NICK_SET_DISABLED);
		return;
    }

    if (is_servadmin && cmd && (*cmd == '=')) {
		if (!(na = aliases.find(cmd+1))) {
			notice_lang(s_NickServ, u, NICK_X_NOT_REGISTERED, cmd + 1);
			return;
		}
		
		cmd = param;
		param = strtok(NULL, " ");
		set_nick = 1;
    } else {
		na = u->na;
	}

    if (!param && (!cmd || (stricmp(cmd,"URL")!=0 && stricmp(cmd,"EMAIL")!=0 && stricmp(cmd,"GREET")!=0 && stricmp(cmd,"IM")!=0))){
		if (is_servadmin) {
	    	syntax_error(s_NickServ, u, "SET", NICK_SET_SERVADMIN_SYNTAX);
		} else {
	    	syntax_error(s_NickServ, u, "SET", NICK_SET_SYNTAX);
		}
	} else if (!na) {
		notice_lang(s_NickServ, u, NICK_NOT_REGISTERED);
    } else if (!is_servadmin && !nick_identified(u)) {
		notice_lang(s_NickServ, u, NICK_IDENTIFY_REQUIRED, s_NickServ);
	} else if (stricmp(cmd, "GNAME") == 0) {
		do_set_gname(u, na->ng(), param);
    } else if (stricmp(cmd, "DISPLAY") == 0) {
		notice_lang(s_NickServ, u, OBSOLETE_COMMAND, "SET GNAME");
	} else if (stricmp(cmd, "PASSWORD") == 0) {
		do_set_password(u, na->ng(), param);
    } else if (stricmp(cmd, "LANGUAGE") == 0) {
		do_set_language(u, na->ng(), param);
    } else if (stricmp(cmd, "URL") == 0) {
		do_set_url(u, na->ng(), param);
    } else if (stricmp(cmd, "EMAIL") == 0) {
		do_set_email(u, na->ng(), param);
	} else if (stricmp(cmd, "ICQ") == 0) {
		notice_lang(s_NickServ, u, OBSOLETE_COMMAND, "SET IM");
	} else if (stricmp(cmd, "IM") == 0) {
		do_set_im(u, na->ng(), param);
	} else if (stricmp(cmd, "GREET") == 0) {
		do_set_greet(u, na->ng(), param);
    } else if (stricmp(cmd, "KILL") == 0) {
		do_set_kill(u, na->ng(), param);
    } else if (stricmp(cmd, "SECURE") == 0) {
		do_set_secure(u, na->ng(), param);
    } else if (stricmp(cmd, "PRIVATE") == 0) {
		do_set_private(u, na->ng(), param);
	} else if (stricmp(cmd, "HOST") == 0) {
		do_set_host(u, na->ng(), param);
    } else if (stricmp(cmd, "HIDE") == 0) {
		do_set_hide(u, na->ng(), param);
    } else if (stricmp(cmd, "NOEXPIRE") == 0) {
		do_set_noexpire(u, na, param);
    } else {
	    notice_lang(s_NickServ, u, NICK_SET_UNKNOWN_OPTION, cmd);
    }
}

/*************************************************************************/

static void do_set_gname(User *u, NickGroup *ng, char *param)
{
	NickAlias *na = 0;
	
	/* First check whether param is a valid nick of the group */
	VectorNA::const_iterator it, end;
	
	for (it = ng->aliases().begin(), end = ng->aliases().end();
			it != end; ++it)
	{
		na = (NickAlias *)(*it);
		if (!stricmp(na->nick(), param))
			break;
	}
	
	if (it == end) 
	{
		notice_lang(s_NickServ, u, NICK_SET_GNAME_INVALID);
		return;
	}
	
	/* Don't move the notice after the change ! It won't work. */
	notice_lang_set(s_NickServ, u, NICK_SET_GNAME_CHANGED, 
			NICK_SET_SERVADMIN_GNAME_CHANGED, ng, na->nick());
	ng->change_name(na);
}

/*************************************************************************/

static void do_set_password(User *u, NickGroup *nc, char *param)
{
    size_t len = strlen(param);

    if (epona->cd().nickserv.protect_admins && u->na->ng() != nc &&
    		nick_is_services_admin(nc) && !is_services_root(u)) {
		notice_lang(s_NickServ, u, PERMISSION_DENIED);
		return;
    } else if (stricmp(nc->name(), param) == 0 ||
    		(epona->cd().strict_passwords && len < 5)) {
		notice_lang(s_NickServ, u, MORE_OBSCURE_PASSWORD);
		return;
	}

	if (len > NickGroup::pass_rlen)
	{
		param[NickGroup::pass_rlen] = 0;
		notice_lang(s_NickServ, u, PASSWORD_TRUNCATED, NickGroup::pass_rlen);
	}
	
	if (!nc->change_password(param))
	{
		memset(param, 0, len);
		notice_lang(s_NickServ, u, NICK_SET_PASSWORD_FAILED);
		return;
	}
	
	if (nc->enc())
	{
		memset(param, 0, len);
		notice_lang_set(s_NickServ, u, NICK_SET_PASSWORD_CHANGED, 
				NICK_SET_SERVADMIN_PASSWORD_CHANGED, nc);
	} else {
		notice_lang_set(s_NickServ, u, NICK_SET_PASSWORD_CHANGED_TO, 
				NICK_SET_SERVADMIN_PASSWORD_CHANGED_TO, nc, nc->pass());
	}

	if (u->na && u->na->ng() != nc && is_services_admin(u)) {
		epona->log(NoticeNickServPrivs(),
				"%s!%s@%s used SET PASSWORD as services admin on %s (e-mail: %s)",
				u->nick, u->username, GetHost(u), 
				nc->name(), (nc->email() ? nc->email() : "none"));
		if (epona->cd().operserv.wall & Epona::Directives::OperServ::Wall::waSetpass)
	    	wallops(s_NickServ, "\2%s\2 used SET PASSWORD as Services admin on \2%s\2", u->nick, nc->name());
	} else {
		epona->log(InfoNickServSettings(),
				"%s!%s@%s (e-mail: %s) changed its password", u->nick,
				u->username, GetHost(u), (nc->email() ? nc->email() : "none"));
	}
}

/*************************************************************************/

static void do_set_language(User *u, NickGroup *nc, char *param)
{
    int langnum;

    if (param[strspn(param, "0123456789")] != 0) {  /* i.e. not a number */
		syntax_error(s_NickServ, u, "SET LANGUAGE", NICK_SET_LANGUAGE_SYNTAX);
		return;
    }
    langnum = atoi(param)-1;
    if (langnum < 0 || langnum >= NUM_LANGS || langlist[langnum] < 0) {
		notice_lang(s_NickServ, u, NICK_SET_LANGUAGE_UNKNOWN, langnum+1, s_NickServ);
		return;
    }
    nc->language(langlist[langnum]);
    notice_lang_set(s_NickServ, u, NICK_SET_LANGUAGE_CHANGED,
    		NICK_SET_SERVADMIN_LANGUAGE_CHANGED, nc);
}

/*************************************************************************/

static void do_set_url(User *u, NickGroup *ng, char *param)
{
	ng->url(param);
	if (param) 
		notice_lang_set(s_NickServ, u, NICK_SET_URL_CHANGED, 
				NICK_SET_SERVADMIN_URL_CHANGED, ng, ng->url());
	else
		notice_lang_set(s_NickServ, u, NICK_SET_URL_UNSET, 
				NICK_SET_SERVADMIN_URL_UNSET, ng);
}

/*************************************************************************/

static void do_set_email(User *u, NickGroup *ng, char *param)
{
	if (param && strlen(param) > NickGroup::email_len)
		param[NickGroup::email_len] = 0;
	
	if (!param && epona->cd().nickserv.require_email) {
		notice_lang(s_NickServ, u, NICK_SET_EMAIL_UNSET_IMPOSSIBLE);
		return;
	} else if (param && !mail::Mail::valid_address(param, true)) {
		notice_lang(s_NickServ, u, MAIL_X_INVALID, param);
		return;
	}
	
	if (!ng->reset_auth())
	{
		notice_lang(s_NickServ, u, NICK_SET_DISABLED);
		return;
	}
	
	epona->log(InfoNickServSettings(),
			"%s!%s@%s (e-mail: %s) changed its e-mail to %s", u->nick, 
			u->username, GetHost(u), (ng->email() ? ng->email() : "none"),
			(param ? param : "none"));
	
	ng->email(param);
	
	if (param)
		notice_lang_set(s_NickServ, u, NICK_SET_EMAIL_CHANGED, 
				NICK_SET_SERVADMIN_EMAIL_CHANGED, ng, ng->email());
	else
		notice_lang_set(s_NickServ, u, NICK_SET_EMAIL_UNSET,
				NICK_SET_SERVADMIN_EMAIL_UNSET, ng);
	
	if (u->na && u->na->ng() == ng &&
			epona->cd().nickserv.auth && epona->cd().nickserv.auth.required)
		ng->require_auth(u, true);
	else
		ng->end_auth(false); /* To cancel the current authentication code */
}

/*************************************************************************/

static void do_set_im(User *u, NickGroup *ng, char *param)
{
	if (param) 
	{
		char *type = param, *id = strtok(NULL, " ");
		
		if (!type || !id) 
		{
			syntax_error(s_NickServ, u, "SET IM", NICK_SET_IM_SYNTAX);
		}
		else if (stricmp(type, "AIM") && stricmp(type, "ICQ") &&
				stricmp(type, "MSN") && stricmp(type, "Yahoo") &&
				stricmp(type, "Yahoo!") && stricmp(type, "Y!M"))
		{
			notice_lang(s_NickServ, u, NICK_SET_IM_UNSUPPORTED_TYPE, type,
					"AIM, ICQ, MSN, Yahoo!");
		} else {
			ng->change_im(type, id);	
			notice_lang_set(s_NickServ, u, NICK_SET_IM_CHANGED, 
					NICK_SET_SERVADMIN_IM_CHANGED, ng, id, type);
		}
	} else {
		ng->change_im(0, 0);
		notice_lang_set(s_NickServ, u, NICK_SET_IM_UNSET, 
				NICK_SET_SERVADMIN_IM_UNSET, ng);
	}
}

/*************************************************************************/

static void do_set_greet(User *u, NickGroup *ng, char *param)
{
    if (param) {
    	char buf[BUFSIZE];
    	char *end = strtok(NULL, "");
    	
    	snprintf(buf, sizeof(buf), "%s%s%s", param, (end ? " " : ""), (end ? end : ""));
    	
		ng->greet(buf);
		notice_lang_set(s_NickServ, u, NICK_SET_GREET_CHANGED,
				 NICK_SET_SERVADMIN_GREET_CHANGED, ng, ng->greet());
    } else {
		ng->greet(0);
		notice_lang_set(s_NickServ, u, NICK_SET_GREET_UNSET,
				NICK_SET_SERVADMIN_GREET_UNSET, ng);
    }
}

/*************************************************************************/

static void do_set_kill(User *u, NickGroup *nc, char *param)
{
    if (stricmp(param, "ON") == 0) {
		nc->flags_add(NickGroup::gfKill);
		nc->flags_remove(NickGroup::gfKillQuick | NickGroup::gfKillImmed);
		notice_lang_set(s_NickServ, u, NICK_SET_KILL_ON, 
				NICK_SET_SERVADMIN_KILL_ON, nc);
    } else if (stricmp(param, "QUICK") == 0) {
		nc->flags_add(NickGroup::gfKill | NickGroup::gfKillQuick);
		nc->flags_remove(NickGroup::gfKillImmed);
		notice_lang_set(s_NickServ, u, NICK_SET_KILL_QUICK, 
				NICK_SET_SERVADMIN_KILL_QUICK, nc);
    } else if (stricmp(param, "IMMED") == 0) {
		if (epona->cd().nickserv.kill_immed) {
	    	nc->flags_add(NickGroup::gfKill | NickGroup::gfKillImmed);
	    	nc->flags_remove(NickGroup::gfKillQuick);
	    	notice_lang_set(s_NickServ, u, NICK_SET_KILL_IMMED, 
	    			NICK_SET_SERVADMIN_KILL_IMMED, nc);
		} else {
	    	notice_lang(s_NickServ, u, NICK_SET_KILL_IMMED_DISABLED);
		}
    } else if (stricmp(param, "OFF") == 0) {
		nc->flags_remove(NickGroup::gfKill | NickGroup::gfKillQuick | NickGroup::gfKillImmed);
		notice_lang_set(s_NickServ, u, NICK_SET_KILL_OFF, 
				NICK_SET_SERVADMIN_KILL_OFF, nc);
    } else {
		syntax_error(s_NickServ, u, "SET KILL",
				epona->cd().nickserv.kill_immed ? NICK_SET_KILL_IMMED_SYNTAX :
				NICK_SET_KILL_SYNTAX);
    }
}

/*************************************************************************/

static void do_set_secure(User *u, NickGroup *nc, char *param)
{
    if (stricmp(param, "ON") == 0) {
		nc->flags_add(NickGroup::gfSecure);
		notice_lang_set(s_NickServ, u, NICK_SET_SECURE_ON,
				NICK_SET_SERVADMIN_SECURE_ON, nc);
    } else if (stricmp(param, "OFF") == 0) {
		nc->flags_remove(NickGroup::gfSecure);
		notice_lang_set(s_NickServ, u, NICK_SET_SECURE_OFF,
				NICK_SET_SERVADMIN_SECURE_OFF, nc);
    } else {
		syntax_error(s_NickServ, u, "SET SECURE", NICK_SET_SECURE_SYNTAX);
    }
}

/*************************************************************************/

static void do_set_private(User *u, NickGroup *nc, char *param)
{
    if (stricmp(param, "ON") == 0) {
		nc->flags_add(NickGroup::gfPrivate);
		notice_lang_set(s_NickServ, u, NICK_SET_PRIVATE_ON,
				NICK_SET_SERVADMIN_PRIVATE_ON, nc);
    } else if (stricmp(param, "OFF") == 0) {
		nc->flags_remove(NickGroup::gfPrivate);
		notice_lang_set(s_NickServ, u, NICK_SET_PRIVATE_OFF,
				NICK_SET_SERVADMIN_PRIVATE_OFF, nc);
    } else {
		syntax_error(s_NickServ, u, "SET PRIVATE", NICK_SET_PRIVATE_SYNTAX);
    }
}

/*************************************************************************/

static void do_set_host(User *u, NickGroup *nc, char *param)
{
#ifdef HAS_CHGHOST
	if (!epona->cd().nickserv.custom_host.enable)
	{
#endif
		notice_lang(s_NickServ, u, NICK_SET_OPTION_DISABLED, "HOST");
#ifdef HAS_CHGHOST
	} else if (stricmp(param, "ON") == 0) {
		if (!nc->chost()) {
			notice_lang(s_NickServ, u, NICK_SET_HOST_UNSET);
		} else {
			nc->flags_add(NickGroup::gfCustomHost);
			notice_lang_set(s_NickServ, u, NICK_SET_HOST_ON,
					NICK_SET_SERVADMIN_HOST_ON, nc);
		}
    } else if (stricmp(param, "OFF") == 0) {
		nc->flags_remove(NickGroup::gfCustomHost);
		notice_lang_set(s_NickServ, u, NICK_SET_HOST_OFF,
				NICK_SET_SERVADMIN_HOST_OFF, nc);
    } else {
		syntax_error(s_NickServ, u, "SET HOST", NICK_SET_HOST_SYNTAX);
    }
#endif
}

/*************************************************************************/

static void do_set_hide(User *u, NickGroup *nc, char *param)
{
    int flag, onmsg, offmsg, saonmsg, saoffmsg;

    if (stricmp(param, "EMAIL") == 0) {
		flag = NickGroup::gfHideEmail;
		onmsg = NICK_SET_HIDE_EMAIL_ON;
		offmsg = NICK_SET_HIDE_EMAIL_OFF;
		saonmsg = NICK_SET_SERVADMIN_HIDE_EMAIL_ON;
		saoffmsg = NICK_SET_SERVADMIN_HIDE_EMAIL_OFF;
    } else if (stricmp(param, "USERMASK") == 0) {
		flag = NickGroup::gfHideMask;
		onmsg = NICK_SET_HIDE_MASK_ON;
		offmsg = NICK_SET_HIDE_MASK_OFF;
		saonmsg = NICK_SET_SERVADMIN_HIDE_MASK_ON;
		saoffmsg = NICK_SET_SERVADMIN_HIDE_MASK_OFF;
    } else if (stricmp(param, "QUIT") == 0) {
		flag = NickGroup::gfHideQuit;
		onmsg = NICK_SET_HIDE_QUIT_ON;
		offmsg = NICK_SET_HIDE_QUIT_OFF;
		saonmsg = NICK_SET_SERVADMIN_HIDE_QUIT_ON;
		saoffmsg = NICK_SET_SERVADMIN_HIDE_QUIT_OFF;
    } else {
		syntax_error(s_NickServ, u, "SET HIDE", NICK_SET_HIDE_SYNTAX);
		return;
    }
    
    param = strtok(NULL, " ");
    if (!param) {
		syntax_error(s_NickServ, u, "SET HIDE", NICK_SET_HIDE_SYNTAX);
    } else if (stricmp(param, "ON") == 0) {
		nc->flags_add(flag);
		notice_lang_set(s_NickServ, u, onmsg, saonmsg, nc, s_NickServ);
    } else if (stricmp(param, "OFF") == 0) {
		nc->flags_remove(flag);
		notice_lang_set(s_NickServ, u, offmsg, saoffmsg, nc, s_NickServ);
    } else {
		syntax_error(s_NickServ, u, "SET HIDE", NICK_SET_HIDE_SYNTAX);
    }
}

/*************************************************************************/

static void do_set_noexpire(User *u, NickAlias *na, char *param)
{
    if (!is_services_admin(u)) {
		notice_lang(s_NickServ, u, PERMISSION_DENIED);
		return;
    }
    if (!param) {
		syntax_error(s_NickServ, u, "SET NOEXPIRE", NICK_SET_NOEXPIRE_SYNTAX);
		return;
    }
    if (stricmp(param, "ON") == 0) {
		na->status_add(NickAlias::asNoExpire);
		notice_lang(s_NickServ, u, NICK_SET_NOEXPIRE_ON, na->nick());
    } else if (stricmp(param, "OFF") == 0) {
		na->status_remove(NickAlias::asNoExpire);
		notice_lang(s_NickServ, u, NICK_SET_NOEXPIRE_OFF, na->nick());
    } else {
		syntax_error(s_NickServ, u, "SET NOEXPIRE", NICK_SET_NOEXPIRE_SYNTAX);
    }
}

/*************************************************************************/

static void do_link(User *u)
{
	notice_lang(s_NickServ, u, OBSOLETE_COMMAND, "GROUP");	
}

/*************************************************************************/

static void do_unlink(User *u)
{
	notice_lang(s_NickServ, u, OBSOLETE_COMMAND, "DROP");	
}

/*************************************************************************/

static void do_listlinks(User *u)
{
	notice_lang(s_NickServ, u, OBSOLETE_COMMAND, "GLIST");
}

/*************************************************************************/

AccessCmd::AccessCmd() : ListCmd("ACCESS", s_NickServ, access_msgs, true)
{ 

}

/*************************************************************************/

int AccessCmd::add(User *u, container_type &ctn, char const *sub)
{
	char *mask = strtok(NULL, " ");
	
	if (!mask)
	{
		syntax_error(m_service, u, m_name, NICK_ACCESS_ADD_SYNTAX);
		return crStop;
	}
	
	/* Validates the mask */
	
	if (strlen(mask) > NickGroup::Access::mask_len)
	{
		notice_lang(m_service, u, NICK_ACCESS_MASK_TOO_LONG,
				NickGroup::Access::mask_len);
		return crStop;
	}
	
	if (!strchr(mask, '@'))
	{
		notice_lang(m_service, u, NICK_ACCESS_MASK_INVALID, mask);
		return crStop;
	}
	
	if (strchr(mask, '!'))
	{
		notice_lang(m_service, u, NICK_ACCESS_MASK_NO_NICK);
		return crStop;
	}
	
	/* Add the record */
	
	if (add_record(u, u->na->ng(), mask) < 0)
		return crStop;

	return crStop;
}

/*************************************************************************/

int AccessCmd::add_record(User *u, NickGroup *owner, char const *mask)
{
	/* Check whether we can add more entries. */
	if (owner->accesses().size() >= epona->cd().nickserv.max_access)
	{
		NickGroup::accesses_type::const_iterator it, end;
		
		for (it = owner->accesses().begin(), end = owner->accesses().end();
				it != end; ++it)
		{
			if ((*it) == 0)
				break;
		}
		
		if (it == end)
		{
			if (u)
				notice_lang(m_service, u, NICK_ACCESS_REACHED_LIMIT,
						epona->cd().nickserv.max_access.value());
			return -1;
		}
	}
	
	NickGroup::Access *nga;
	
	nga = (NickGroup::Access *)groups.accesses().create();
	
	if (!nga)
	{
		if (u)
			notice_lang(m_service, u, NICK_ACCESS_CREATION_FAILED);
		return -1;
	}
	
	nga->owner(owner);
	nga->mask(mask);
	
	/* Try to add the access entry to the list */
	
	int error;
	
	if (!groups.accesses().add(nga, &error))
	{
		if (u)
		{
			if (error == Groups::Accesses::ceName ||
					error == databases::Collection::ceList)
				notice_lang(m_service, u, NICK_ACCESS_ALREADY_PRESENT,
				 		nga->mask());
			else
				notice_lang(m_service, u, m_msgs[cmDisabled]);
		}
				
		delete nga;
		return -1;
	} else {
		if (u)
			notice_lang(m_service, u, NICK_ACCESS_ADDED, nga->mask());
	}
	
	return 0;
}

/*************************************************************************/

bool AccessCmd::changeable(User *u, container_type &ctn,
		char const *sub) const
{
	return groups.accesses().writable();
}

/*************************************************************************/

AccessCmd::container_type *AccessCmd::container(User *u)
{
	if (!u->na)
	{
		notice_lang(m_service, u, NICK_NOT_REGISTERED);
		return 0;
	}
	
    if (!nick_identified(u)) 
    {
		notice_lang(m_service, u, NICK_IDENTIFY_REQUIRED, m_service.str());
		return 0;
	}
	
	return &u->na->ng()->accesses();
}

/*************************************************************************/

bool AccessCmd::del_record(User *u, container_type &ctn, char const *sub,
		databases::Record *rec)
{
	return groups.accesses().remove(rec);
}

/*************************************************************************/

bool AccessCmd::list_record(User *u, container_type &ctn, char const *sub,
		uint16_t num, databases::Record const *rec, uint16_t count)
{
	NickGroup::Access *nga = (NickGroup::Access *)rec;
	
	if (count == 0)
		notice_lang(m_service, u, NICK_ACCESS_LIST_COLHEAD);
	
	notice_lang(m_service, u, NICK_ACCESS_LIST_FORMAT, num, nga->mask());
	return true;
}

/*************************************************************************/

static void do_access(User *u)
{
	static AccessCmd access_cmd;
	access_cmd.execute(u);
}

/*************************************************************************/

/* Show hidden info to nick owners and sadmins when the "ALL" parameter is
 * supplied. If a nick is online, the "Last seen address" changes to "Is 
 * online from".
 * Syntax: INFO <nick> {ALL}
 * -TheShadow (13 Mar 1999)
 */

static void do_info(User *u)
{
    char *nick = strtok(NULL, " ");
    char *param = strtok(NULL, " ");
    
    NickAlias *na;
    int is_servadmin = is_services_admin(u);

    if (!nick) {
    	syntax_error(s_NickServ, u, "INFO", NICK_INFO_SYNTAX);
    } else if (!(na = aliases.find(nick))) {
		notice_lang(s_NickServ, u, NICK_X_NOT_REGISTERED, nick);
    } else {
    	time_t t;
		struct tm *tm;
		char buf[BUFSIZE], *end;
		const char *commastr = getstring(u->na, COMMA_SPACE);
		int need_comma = 0;
		int nick_online = 0;
		int show_hidden = 0;

		/* Is the real owner of the nick we're looking up online? -TheShadow */
		if (na->status() & (NickAlias::asRecognized | NickAlias::asIdentified))
	    	nick_online = 1;

    	/* Only show hidden fields to owner and sadmins and only when the ALL
	 	 * parameter is used. -TheShadow */
    	if (param && stricmp(param, "ALL") == 0 && u->na &&
    			((nick_identified(u) && (na->ng() == u->na->ng())) ||
    			  is_servadmin))
            show_hidden = 1;

		notice_lang(s_NickServ, u, NICK_INFO_REALNAME, na->nick(), 
				(na->last_realname() ? na->last_realname() :
				 getstring(u->na, UNKNOWN)));
		
		if (nick_is_services_admin(na->ng())) /* This will also include the services root(s) */
			notice_lang(s_NickServ, u, NICK_INFO_SERVICES_ADMIN, na->nick());
		else if (nick_is_services_oper(na->ng()))
			notice_lang(s_NickServ, u, NICK_INFO_SERVICES_OPER, na->nick());
		

		if (nick_online) {
	    	if (na->last_usermask() && (show_hidden ||
	    			!(na->ng()->flags() & NickGroup::gfHideMask)))
			notice_lang(s_NickServ, u, NICK_INFO_ADDRESS_ONLINE, na->last_usermask());
	    	else
			notice_lang(s_NickServ, u, NICK_INFO_ADDRESS_ONLINE_NOHOST, na->nick());
		} else {
	    	if (na->last_usermask() &&
	    			(show_hidden || !(na->ng()->flags() & NickGroup::gfHideMask)))
			notice_lang(s_NickServ, u, NICK_INFO_ADDRESS, na->last_usermask());
		}	    
	    
	    t = na->created();
		tm = localtime(&t);
		strftime_lang(buf, sizeof(buf), u, STRFTIME_DATE_TIME_FORMAT, tm);
		notice_lang(s_NickServ, u, NICK_INFO_TIME_REGGED, buf);
		
		if (show_hidden) {
			t = na->ng()->created();
			tm = localtime(&t);
			strftime_lang(buf, sizeof(buf), u, STRFTIME_DATE_TIME_FORMAT, tm);
			notice_lang(s_NickServ, u, NICK_INFO_GROUP_CREATED, buf);
		}
		
		if (!nick_online) {
			t = na->last_seen();
			tm = localtime(&t);
            strftime_lang(buf, sizeof(buf), u, STRFTIME_DATE_TIME_FORMAT, tm);
            notice_lang(s_NickServ, u, NICK_INFO_LAST_SEEN, buf);
        }
	
		if (na->last_quit() && (show_hidden || !(na->ng()->flags() & NickGroup::gfHideQuit)))
	    	notice_lang(s_NickServ, u, NICK_INFO_LAST_QUIT, na->last_quit());
	    
	    if (na->ng()->url())
	    	notice_lang(s_NickServ, u, NICK_INFO_URL, na->ng()->url());
		if (na->ng()->email() && (show_hidden || 
				!(na->ng()->flags() & NickGroup::gfHideEmail)))
		{
	    	notice_lang(s_NickServ, u, 
	    			((show_hidden && (na->ng()->flags() & NickGroup::gfAuth)) ?
	    			NICK_INFO_EMAIL_AUTH : NICK_INFO_EMAIL), 
	    			na->ng()->email());
	    }
	    if (na->ng()->im_type() && na->ng()->im_id())
	    	notice_lang(s_NickServ, u, NICK_INFO_IM, na->ng()->im_type(),
	    			na->ng()->im_id());
	
		if (show_hidden) {
			if (na->ng()->greet())
			notice_lang(s_NickServ, u, NICK_INFO_GREET, na->ng()->greet());

#ifdef HAS_CHGHOST
			if (na->ng()->chost()) {
				if (is_servadmin) {
					t = na->ng()->chost_changed();
					tm = localtime(&t);
					strftime_lang(buf, sizeof(buf), u, STRFTIME_DATE_TIME_FORMAT, tm);
					notice_lang(s_NickServ, u, NICK_INFO_CUSTOM_HOST_WHO,
							na->ng()->chost(), na->ng()->chost_setby(), buf);
				} else {
					notice_lang(s_NickServ, u, NICK_INFO_CUSTOM_HOST,
							na->ng()->chost());
				}
			}
#endif
		
			*buf = 0;
			end = buf;
			
			if (na->ng()->flags() & NickGroup::gfKill) {
	    		end += snprintf(end, sizeof(buf)-(end-buf), "%s", 
	    			getstring(u->na, NICK_INFO_OPT_KILL));
	    		need_comma = 1;
			}
			if (na->ng()->flags() & NickGroup::gfSecure) {
	    		end += snprintf(end, sizeof(buf)-(end-buf), "%s%s", 
	    			need_comma ? commastr : "", 
	    			getstring(u->na, NICK_INFO_OPT_SECURE));
	    		need_comma = 1;
			}
			if (na->ng()->flags() & NickGroup::gfPrivate) {
	    		end += snprintf(end, sizeof(buf)-(end-buf), "%s%s",
					need_comma ? commastr : "",
					getstring(u->na, NICK_INFO_OPT_PRIVATE));
	    		need_comma = 1;
			}
			if (na->ng()->flags() & NickGroup::gfCustomHost) {
	    		end += snprintf(end, sizeof(buf)-(end-buf), "%s%s",
					need_comma ? commastr : "",
					getstring(u->na, NICK_INFO_OPT_CUSTOM_HOST));
	    		need_comma = 1;
			}
		
			notice_lang(s_NickServ, u, NICK_INFO_OPTIONS, *buf ? buf : getstring(u->na, NICK_INFO_OPT_NONE));
		
			if ((na->status() & NickAlias::asPendingExpire) &&
					epona->cd().nickserv.expire)
	   			notice_lang(s_NickServ, u, NICK_INFO_PENDING_EXPIRE);
	   		
	   		if (na->status() & NickAlias::asNoExpire &&
					epona->cd().nickserv.expire)
	   			notice_lang(s_NickServ, u, NICK_INFO_NO_EXPIRE);
		}
	    
		if (!show_hidden && ((u->na && (na->ng() == u->na->ng()) && nick_identified(u)) || is_servadmin))
			notice_lang(s_NickServ, u, NICK_INFO_FOR_MORE, s_NickServ, na->nick());
    }
}

/*************************************************************************/

static void do_list(User *u)
{
	if (epona->cd().nickserv.list_opers && !is_oper(u)) 
	{
		notice_lang(s_NickServ, u, PERMISSION_DENIED);
		return;
	}
	
	uint32_t status = 0;
	bool nolimit = false;
	char *pattern = strtok(NULL, " "), *range = 0, *opt;
	
	if (pattern && isdigit(*pattern))
	{
		range = pattern;
		pattern = 0;
	}
	
	while ((pattern || range) && (opt = strtok(NULL, " ")))
	{
		if (isdigit(*opt) && !range && !status)
		{
			range = opt;
			continue;
		}
		
		if (is_oper(u) && !stricmp(opt, "NOLIMIT") /*&& (pattern || range) implied */)
		{
			nolimit = true;
			continue;
		}
		
		if (is_services_admin(u) && !stricmp(opt, "NOEXPIRE"))
		{
			status |= NickAlias::asNoExpire;
			continue;
		} 
		
		syntax_error(s_NickServ, u, "LIST",
				(is_oper(u) ? NICK_LIST_SERVADMIN_SYNTAX : NICK_LIST_SYNTAX));
		return;
	}
	
	unsigned long first = 0, last = 0, index = 0;
	int count = 0;
	
	if (!misc::simple_range(range, &first, &last)) 
	{
		notice_lang(s_NickServ, u, NICK_LIST_INVALID_RANGE, range);
		return;
	}
	
	bool header = false;
	char buf[BUFSIZE];
	
	NickGroup *myng = ((u->na && nick_identified(u)) ? u->na->ng() : 0);
	
	NickAlias *na;
	Aliases::const_niterator it, end;
	
	for (it = aliases.nbegin(), end = aliases.nend(); it != end; ++it)
	{
		na = (NickAlias *)(*it);
		
		if (status && !(na->status() & status))
			continue;
		
		if ((na->ng()->flags() & NickGroup::gfPrivate) &&
				!is_services_admin(u) && (na->ng() != myng))
			continue;
		
		if (pattern && strcmp(pattern, "*"))
		{
			snprintf(buf, sizeof(buf), "%s!%s", na->nick(),
					na->last_usermask() ? na->last_usermask() : "*@*");
			if (stricmp(na->nick(), pattern) &&
					!misc::irc_match(pattern, buf))
				continue;
		}
		
		if (++index < first || index > last)
			continue;
		
		if (!header)
		{
			notice_lang(s_NickServ, u,
					(pattern ? NICK_LIST_HEADER : 
					NICK_LIST_HEADER_NO_PATTERN), pattern);
			header = true;
		}
		
		if ((++count < epona->cd().nickserv.list_max) || nolimit)
		{
			char noexpire_char = ' ';
			if (is_services_admin(u) && (na->status() & NickAlias::asNoExpire))
				noexpire_char = '!';
			
			if ((na->ng()->flags() & NickGroup::gfHideMask) &&
					!is_services_admin(u) && (na->ng() != myng)) {
				snprintf(buf, sizeof(buf), "%-20s  [Hidden]", na->nick());
			} else {
				snprintf(buf, sizeof(buf), "%-20s  %s", na->nick(), 
						(na->last_usermask() ? na->last_usermask() : ""));
			}
			notice_user(s_NickServ, u, "   %c%s", noexpire_char, buf);
		}
	}
	
	if (!header && pattern)
	{
		notice_lang(s_NickServ, u, NICK_LIST_NO_NICKS_MATCH, pattern);
		return;
	}
	
	if (!nolimit && count > epona->cd().nickserv.list_max)
		count = epona->cd().nickserv.list_max;
	
	if (range && count)
	{
		if (first == last)
			notice_lang(s_NickServ, u,
					NICK_LIST_RESULTS_RANGE_ONE, first);
		else
			notice_lang(s_NickServ, u, NICK_LIST_RESULTS_RANGE,
					first, (first + count - 1));
	} else {
		notice_lang(s_NickServ, u, NICK_LIST_RESULTS, count, index);
	}
}

/*************************************************************************/

static void do_glist(User *u)
{
    char *nick = strtok(NULL, " ");
    
    char buf[BUFSIZE];
    time_t t;
    struct tm *tm;
	NickAlias *na, *na2;

    if ((nick ? !is_services_admin(u) : !nick_identified(u))) {
		notice_lang(s_NickServ, u, ACCESS_DENIED);
    } else if ((!nick ? !(na = u->na) : !(na = aliases.find(nick)))) {
		notice_lang(s_NickServ, u, (!nick ? NICK_NOT_REGISTERED : NICK_X_NOT_REGISTERED), nick);
    } else {
    	notice_lang(s_NickServ, u, (nick ? NICK_GLIST_HEADER_X :
    			NICK_GLIST_HEADER), na->ng()->name());
    	
    	VectorNA::const_iterator it, end;
    	
    	for (it = na->ng()->aliases().begin(), end = na->ng()->aliases().end();
    			it != end; ++it)
    	{
			na2 = (NickAlias *)(*it);
			
			if (!epona->cd().nickserv.expire || (na2->status() & NickAlias::asNoExpire)) {
				notice_lang(s_NickServ, u, NICK_GLIST_ENTRY_NOEXPIRE, na2->nick());
			} else if (na2->status() & NickAlias::asPendingExpire) {
				notice_lang(s_NickServ, u, NICK_GLIST_ENTRY_PENDING, na2->nick());
			} else {
				t = na2->last_seen() + epona->cd().nickserv.expire.delay;
				tm = localtime(&t);
				strftime_lang(buf, sizeof(buf), u, STRFTIME_DATE_TIME_FORMAT, tm);
				notice_lang(s_NickServ, u, NICK_GLIST_ENTRY, na2->nick(), buf);
			}
		}	
		notice_lang(s_NickServ, u, NICK_GLIST_FOOTER,
				na->ng()->aliases().size());
    }
}

/*************************************************************************/

static void do_recover(User *u)
{
    char *nick = strtok(NULL, " ");
    char *pass = strtok(NULL, " ");
    NickAlias *na;
    User *u2;

    if (!nick) {
		syntax_error(s_NickServ, u, "RECOVER", NICK_RECOVER_SYNTAX);
    } else if (!(u2 = finduser(nick))) {
		notice_lang(s_NickServ, u, NICK_X_NOT_IN_USE, nick);
    } else if (!(na = u2->na)) {
		notice_lang(s_NickServ, u, NICK_X_NOT_REGISTERED, nick);
	} else if (stricmp(nick, u->nick) == 0) {
		notice_lang(s_NickServ, u, NICK_NO_RECOVER_SELF);
    } else if (pass) {
		if (na->ng()->check_password(pass)) {
			notice_lang(s_NickServ, u2, FORCENICKCHANGE_NOW);
	    	collide(na, 0);
	    	notice_lang(s_NickServ, u, NICK_RECOVERED, s_NickServ, nick);
		} else {
	    	notice_lang(s_NickServ, u, ACCESS_DENIED);
	    	
	    	epona->log(InfoNickServRights(),
					"RECOVER: invalid password for %s by %s!%s@%s",
					nick, u->nick, u->username, GetHost(u));
			bad_password(u);
		}
    } else {
		if (group_identified(u, na->ng()) ||
				(!(na->ng()->flags() & NickGroup::gfSecure) && na->ng()->on_access(u))) 
		{
	    	notice_lang(s_NickServ, u2, FORCENICKCHANGE_NOW);
	    	collide(na, 0);
	    	notice_lang(s_NickServ, u, NICK_RECOVERED, s_NickServ, nick);
		} else {
	    	notice_lang(s_NickServ, u, ACCESS_DENIED);
		}
    }
}

/*************************************************************************/

static void do_release(User *u)
{
    char *nick = strtok(NULL, " ");
    char *pass = strtok(NULL, " ");
    NickAlias *na;

    if (!nick) {
		syntax_error(s_NickServ, u, "RELEASE", NICK_RELEASE_SYNTAX);
    } else if (!(na = aliases.find(nick))) {
		notice_lang(s_NickServ, u, NICK_X_NOT_REGISTERED, nick);
	} else if (!(na->status() & NickAlias::asKillHeld)) {
		notice_lang(s_NickServ, u, NICK_RELEASE_NOT_HELD, nick);
    } else if (pass) {
		if (na->ng()->check_password(pass)) {
	    	release(na, 0);
	    	notice_lang(s_NickServ, u, NICK_RELEASED);
		} else {
	    	notice_lang(s_NickServ, u, ACCESS_DENIED);
	    	
			epona->log(InfoNickServRights(),
					"RELEASE: invalid password for %s by %s!%s@%s",
					nick, u->nick, u->username, GetHost(u));
			bad_password(u);
		}
    } else {
		if (group_identified(u, na->ng()) || (!(na->ng()->flags() & NickGroup::gfSecure) && na->ng()->on_access(u))) {
	    	release(na, 0);
	    	notice_lang(s_NickServ, u, NICK_RELEASED);
		} else {
	    	notice_lang(s_NickServ, u, ACCESS_DENIED);
		}
    }
}

/*************************************************************************/

static void do_ghost(User *u)
{
    char *nick = strtok(NULL, " ");
    char *pass = strtok(NULL, " ");
    NickAlias *na;
    User *u2;

    if (!nick) {
		syntax_error(s_NickServ, u, "GHOST", NICK_GHOST_SYNTAX);
    } else if (!(u2 = finduser(nick))) {
		notice_lang(s_NickServ, u, NICK_X_NOT_IN_USE, nick);
    } else if (!(na = u2->na)) {
		notice_lang(s_NickServ, u, NICK_X_NOT_REGISTERED, nick);
	} else if (stricmp(nick, u->nick) == 0) {
		notice_lang(s_NickServ, u, NICK_NO_GHOST_SELF);
    } else if (pass) {
		if (na->ng()->check_password(pass)) {
	    	char buf[NICKMAX+32];
	    	snprintf(buf, sizeof(buf), "GHOST command used by %s", u->nick);
#ifdef IRC_BAHAMUT
			send_cmd(NULL, "SVSKILL %s :%s", nick, buf);
#else
	    	kill_user(s_NickServ, nick, buf);
#endif
	    	notice_lang(s_NickServ, u, NICK_GHOST_KILLED, nick);
		} else {
	    	notice_lang(s_NickServ, u, ACCESS_DENIED);
	    	epona->log(InfoNickServRights(), 
					"GHOST: invalid password for %s by %s!%s@%s", nick,
					u->nick, u->username, GetHost(u));
			bad_password(u);
		}
    } else {
		if (group_identified(u, na->ng()) || (!(na->ng()->flags() & NickGroup::gfSecure) && na->ng()->on_access(u))) {
	    	char buf[NICKMAX+32];
	    	snprintf(buf, sizeof(buf), "GHOST command used by %s", u->nick);
#ifdef IRC_BAHAMUT
			send_cmd(NULL, "SVSKILL %s :%s", nick, buf);
#else
	    	kill_user(s_NickServ, nick, buf);
#endif
	    	notice_lang(s_NickServ, u, NICK_GHOST_KILLED, nick);
		} else {
	    	notice_lang(s_NickServ, u, ACCESS_DENIED);
		}
    }
}

/*************************************************************************/

static void do_status(User *u)
{
    char *nick;
    User *u2;
    int i = 0;

    while ((nick = strtok(NULL, " ")) && (i++ < 16)) {
		if (!(u2 = finduser(nick)))
	    	notice_user(s_NickServ, u, "STATUS %s 0", nick);
		else if (nick_identified(u2))
	    	notice_user(s_NickServ, u, "STATUS %s 3", nick);
		else if (nick_recognized(u2))
	    	notice_user(s_NickServ, u, "STATUS %s 2", nick);
		else
	    	notice_user(s_NickServ, u, "STATUS %s 1", nick);
    }
}

/*************************************************************************/

static void do_host(User *u)
{	
#ifdef HAS_CHGHOST
	if (!epona->cd().nickserv.custom_host.enable) 
	{
#endif
		notice_lang(s_NickServ, u, NICK_HOST_UNAVAILABLE);
		return;
#ifdef HAS_CHGHOST
	} 
	
	char *cmd = strtok(NULL, " ");
	
	if (!cmd) {
		syntax_error(s_NickServ, u, "HOST", NICK_HOST_SYNTAX);
	} else if (!is_services_oper(u) && (epona->cd().nickserv.custom_host.restrict ||
				stricmp(cmd, "SET"))) {
		notice_lang(s_NickServ, u, (epona->cd().nickserv.custom_host.restrict ?
				ACCESS_DENIED : PERMISSION_DENIED)); 
	} else if (!stricmp(cmd, "SET")) 
	{	
		if (!groups.writable())
		{
			notice_lang(s_NickServ, u, NICK_HOST_SET_DISABLED);
			return;
		}
			
		NickAlias *na;
		time_t now = time(NULL);
		size_t len;
		
		char *nick = strtok(NULL, " ");
		char *host = strtok(NULL, " ");
		
		if (nick && (na = aliases.find(nick))) {
			;
		} else if (host) {
			notice_lang(s_NickServ, u, NICK_X_NOT_REGISTERED, nick);
			return;
		} else {
			if (!u->na) {
				notice_lang(s_NickServ, u, NICK_NOT_REGISTERED);
				return;
			}
			na = u->na;
			host = nick;
		}
		
		if (host) 
		{
			bool strict = ((epona->cd().nickserv.custom_host.enforce_strict ||
					!is_oper(u)) ? true : false);
			
			if (strict && (now - na->ng()->created()) < 
						epona->cd().nickserv.custom_host.age) {
				notice_lang(s_NickServ, u, NICK_HOST_TOO_YOUNG);
			} else if ((len = strlen(host)) > HOSTMAX) {
				notice_lang(s_NickServ, u, NICK_HOST_TOO_LONG, HOSTMAX);
			} else if ((*host == '.') || (*host == '-') || (*(host+len-1) == '.') ||
					(*(host+len-1) == '-') || !strchr(host, '.') || 
					strspn(host, "0123456789abcdefghijklmnopqrstuvwxyz"
					"ABCDEFGHIJKLMNOPQRSTUVWXYZ.-") != len) {
				notice_lang(s_NickServ, u, NICK_HOST_INVALID);
			} else {
				if (!na->ng()->chost() || strcmp(host, na->ng()->chost())) {
					if (strict && na->ng()->chost_changed() && 
							((now - na->ng()->chost_changed()) < 
							epona->cd().nickserv.custom_host.delay)) 
					{
						notice_lang(s_NickServ, u, NICK_HOST_TOO_FAST);
						return;
					}
					
					if (!na->ng()->change_chost(host, now,
							(u->na ? u->na->nick() : u->nick)))
					{
						notice_lang(s_NickServ, u, NICK_HOST_SET_FAILED);
						return;
					}
					
					/* Normally happens only when a custom hostname
				 	 * was previously set. */
					if (na->ng()->flags() & NickGroup::gfCustomHost) 
					{
						NickAlias *na2;
						VectorNA::const_iterator it, end;
						
						for (it = na->ng()->aliases().begin(),
								end = na->ng()->aliases().end();
								it != end; ++it)
						{
							na2 = (NickAlias *)(*it);
							if ((na2->u) && nick_identified(na2->u))
								custom_host_set(na2->u);
						}	
					}
			
					if (u->na && (u->na->ng() == na->ng())) {
						notice_lang(s_NickServ, u, NICK_HOST_CHANGED, host);
					} else {
						notice_lang(s_NickServ, u, NICK_HOST_CHANGED_X, na->nick(), host);
					}
				} else {
					if (u->na && (u->na->ng() == na->ng())) {
						notice_lang(s_NickServ, u, NICK_HOST_UNCHANGED, host);
					} else {
						notice_lang(s_NickServ, u, NICK_HOST_UNCHANGED_X, na->nick(), host);
					}
				}
			}
		} else {
			/* Always let the host be unset, BUT consider it
			 * as a change. */
			
			if (na->ng()->chost()) 
			{
				/* Remove it first... so we don't get in a state where
				 * the flag is still set but the host couldn't be
				 * changed. */
				if ((na->ng()->flags_remove(NickGroup::gfCustomHost) &
						NickGroup::gfCustomHost) ||
						!na->ng()->change_chost(0, now, 0))
				{
					notice_lang(s_NickServ, u, NICK_HOST_SET_FAILED);
					return;
				}
			}
			
			if (u->na && (u->na->ng() == na->ng())) {
				notice_lang(s_NickServ, u, NICK_HOST_UNSET);
			} else {
				notice_lang(s_NickServ, u, NICK_HOST_UNSET_X, na->nick());
			}
		} 
	} else if (!stricmp(cmd, "LIST")) {
		char *mask = strtok(NULL, " ");
		char *range = strtok(NULL, " ");
		
		if (!range && mask && isdigit(*mask)) 
		{
			range = mask;
			mask = NULL;
		}
		
		unsigned long first, last;
		
		if (simple_range(range, &first, &last)) {
			if (!mask)
				notice_lang(s_NickServ, u, NICK_HOST_LIST_HEADER);
			else
				notice_lang(s_NickServ, u, NICK_HOST_LIST_HEADER_MATCH, mask);
			
			unsigned long index = 0;
			int count = 0;
			
			Groups::const_iterator it, end;
			
			for (it = groups.begin(), end = groups.end(); it != end; ++it) 
			{
				NickGroup *ng = (NickGroup *)(*it);
				
				if (!ng->chost())
					continue;
				
				if (mask && !match_wild_nocase(mask, ng->chost()))
					continue;
				
				if (++index < first || index > last)
					continue;
				
				if (++count < epona->cd().nickserv.list_max) 
				{
					char buf[BUFSIZE];
					time_t t = ng->chost_changed();
					struct tm *tm = localtime(&t);
					strftime_lang(buf, sizeof(buf), u,
							STRFTIME_SHORT_DATE_FORMAT, tm);
					
					notice_lang(s_NickServ, u, NICK_HOST_LIST_ENTRY,
							ng->name(), ng->chost(), ng->chost_setby(),
							buf);
				}
			}
			
			count = ((count > epona->cd().nickserv.list_max) ?
					epona->cd().nickserv.list_max : count);
			
			if (range && count) {
				if (first == last)
					notice_lang(s_NickServ, u,
							NICK_HOST_LIST_RESULTS_RANGE_ONE, first);
				else
					notice_lang(s_NickServ, u, NICK_HOST_LIST_RESULTS_RANGE,
							first, (first + count - 1)); 
			} else {
				notice_lang(s_NickServ, u, NICK_HOST_LIST_RESULTS,
						count, index);
			}
		} else {
			/* range is guaranteed to be non-NULL in that case. */
			notice_lang(s_NickServ, u, NICK_HOST_INVALID_RANGE, range);
		}
	} else {
		syntax_error(s_NickServ, u, "HOST", NICK_HOST_SYNTAX);
	}
#endif
}

/*************************************************************************/

static void do_getpass(User *u)
{
	char *nick = strtok(NULL, " ");
	NickAlias *na;
	
	/* Assumes that permission checking has already been done. */
	
	if (!nick) {
		syntax_error(s_NickServ, u, "GETPASS", NICK_GETPASS_SYNTAX);
	} else if (!(na = aliases.find(nick))) {
		notice_lang(s_NickServ, u, NICK_X_NOT_REGISTERED, nick);
	} else if (epona->cd().nickserv.protect_admins && 
			nick_is_services_admin(na->ng()) && !is_services_root(u)) {
		notice_lang(s_NickServ, u, PERMISSION_DENIED);
	} else {
		char passbuf[13];
		char const *pass = passbuf;
		
		if (na->ng()->enc() != 0)
		{
			if (!groups.writable())
			{
				notice_lang(s_NickServ, u, NICK_GETPASS_DISABLED);
				return;
			}
			
			if (!na->ng()->generate_password(passbuf, sizeof(passbuf)))
			{
				notice_lang(s_NickServ, u, NICK_GETPASS_UNAVAILABLE);
				return;
			}
		} else {
			pass = na->ng()->pass();
		}
		
		epona->log(NoticeNickServPrivs(), "%s!%s@%s used GETPASS on %s",
				u->nick, u->username, GetHost(u), na->nick());
		if (epona->cd().operserv.wall &
				Epona::Directives::OperServ::Wall::waGetpass)
			wallops(s_NickServ, "\2%s\2 used GETPASS on \2%s\2",
					u->nick, na->nick());
		notice_lang(s_NickServ, u, NICK_GETPASS_PASSWORD_IS, na->nick(), pass);
		
		if (na->ng()->enc() != 0)
			memset(passbuf, 0, sizeof(passbuf));
	}
}

/*************************************************************************/

static void do_sendpass(User *u)
{
	if (!mail::Mail::configured())
	{
		notice_lang(s_NickServ, u, NICK_SENDPASS_UNCONFIGURED);
		return;
	}
	
	char *nick = strtok(NULL, " ");
	NickAlias *na;
	
	if (!nick) {
		syntax_error(s_NickServ, u, "SENDPASS", NICK_SENDPASS_SYNTAX);
	} else if (epona->cd().nickserv.sendpass_opers && !is_oper(u)) {
		notice_lang(s_NickServ, u, PERMISSION_DENIED);
	} else if (!(na = aliases.find(nick))) {
		notice_lang(s_NickServ, u, NICK_X_NOT_REGISTERED, nick); 
	} else if (na->ng()->enc() && !groups.writable()) {
		notice_lang(s_NickServ, u, NICK_SENDPASS_DISABLED);
	} else {
		if (na->ng()->enc() != 0)
		{	
			if (!na->ng()->begin_auth())
			{
				notice_lang(s_NickServ, u, NICK_SENDPASS_UNAVAILABLE);
				return;
			}
		}
		
		char subject[BUFSIZE];
		snprintf(subject, sizeof(subject),
				getstring(na, NICK_SENDPASS_SUBJECT), na->nick());
		
		mail::Mail mail;
		
		mail.subject(subject);
		mail.to_name(na->ng()->name());
		mail.to_address(na->ng()->email());
		mail.origin(u->host);
		
		mail.text_add(getstring(na, NICK_SENDPASS_HEAD));
		mail.text_add("");
		mail.text_addf(getstring(na, NICK_SENDPASS_LINE_1), na->nick());
		mail.text_add("");
		if (!na->ng()->enc())
			mail.text_addf(getstring(na, NICK_SENDPASS_LINE_2),
					na->ng()->pass());
		else
			mail.text_addf(getstring(na, NICK_SENDPASS_LINE_2_ENCRYPTION),
					s_NickServ, na->ng()->auth_code());
		mail.text_add("");
		mail.text_add(getstring(na, NICK_SENDPASS_LINE_3));
		mail.text_add("");
		mail.text_add(getstring(na, NICK_SENDPASS_LINE_4));
		mail.text_add("");
		mail.text_addf(getstring(na, NICK_SENDPASS_LINE_5),
				(const char *)epona->cd().services.network_name);
		
		uint32_t mflags = 0;
		
		if (is_oper(u))
			mflags |= mail::Mail::mfNoQuota;
		if (na->ng()->flags() & NickGroup::gfAuth)
			mflags |= mail::Mail::mfAuthenticated;
		
		uint32_t err;
		std::string errdesc;
		
		if (!mail.send(mflags, &err, &errdesc))
		{
			notice_lang(s_NickServ, u, ((err & mail::Mail::meTemporary) ?
					NICK_SENDPASS_FAILURE_TEMPORARY : NICK_SENDPASS_FAILURE),
					nick);
			
			if (err & mail::Mail::meQuota)
				epona->log(InfoNickServ(),
						"SENDPASS could not send an e-mail to %s: %s",
						na->ng()->email(), errdesc.c_str());
			else
				epona->log(ErrorNickServ(),
						"SENDPASS could not send an e-mail to %s: %s",
						na->ng()->email(), errdesc.c_str());
			return;
		}
		
		epona->log(InfoNickServSettings(),
				"%s!%s@%s used SENDPASS on %s", u->nick, u->username,
				GetHost(u), nick);
		notice_lang(s_NickServ, u, NICK_SENDPASS_OK, nick);
	}
}

/*************************************************************************/

static void do_forbid(User *u)
{
    notice_lang(s_NickServ, u, REMOVED_COMMAND, s_OperServ, "SQLINE");
}

/*************************************************************************/
